#ifndef UTILITIES_H
#define UTILITIES_H

#include "lesion.h"

typedef struct SMentry {
  String	code ;
  String	graphemes ;
  String	phonemes ;
  Real		frequency ;
} SMentry ;

typedef struct SMlist {
  int		size ;
  int		numEntries ;
  SMentry	*entries ;
} SMlist ;


extern	void	Pause			ARGS((String msg)) ;
extern	Real	sq			();
extern	Real	now			ARGS(( )) ;
extern	double	poz			ARGS((double z)) ;
extern	Real	gaussianNoise		ARGS((double sd)) ;
extern	int	groupIdxFromName	ARGS((Net net, String name)) ;
extern	void	corruptWeights		ARGS((Group preGroup, Group postGroup,
					      double sd)) ;
extern	void	randomizeWeights	ARGS((Group preGroup, Group postGroup,
					      double randRange)) ;
extern	void	scaleWeights		ARGS((Group preGroup, Group postGroup,
					      double scale)) ;
extern	void	restoreWeights		ARGS((Group preGroup, Group postGroup)) ;
extern	Real	proximity		ARGS((Real* v1, Real* v2, int n)) ;
extern	Real	correlation		ARGS((Real* v1, Real* v2, int n)) ;
extern	Real	divergence		ARGS((Real* act, Real* targ, int n)) ;
extern	Real	gpe			ARGS((Real* act, Real* targ, int n)) ;
extern	int	netCountUnits		ARGS((Net net)) ;
extern	void	netSaveOutputs		ARGS((Net net, Real *v)) ;
extern	void	netSaveTotalInputs	ARGS((Net net, Real *v)) ;
extern	void	netRestoreOutputs	ARGS((Net net, Real *v)) ;
extern	void	netRestoreTotalInputs	ARGS((Net net, Real *v)) ;
extern	void	saveOutputs		ARGS((Group group, Real *v)) ;
extern	void	saveOutputsAt		ARGS((int t, Group group, Real *v)) ;
extern	void	restoreOutputs		ARGS((Group group, Real *v)) ;
extern	Real	groupOutputProximity	ARGS((Group group, Real* v)) ;
extern	Real	groupError		ARGS((Group group)) ;
extern	Real	sumSquaredError		ARGS((Net net)) ;
extern	Real	crossEntropyError	ARGS((Net net)) ;
extern	void	*getExampleSetData	ARGS((ExampleSet exampleSet)) ;
extern	void	putExampleSetData	ARGS((ExampleSet exampleSet, void* data)) ;
extern	void	initEpoch 		ARGS((Net net, int onset)) ;
extern	int	lastTargetSlice		ARGS((Net net)) ;
extern	void	lesion			ARGS((Group preGroup, Group postGroup,
					      double prob, double minWt,
					      int print)) ;
extern	Real	maxDiffFromTarget	ARGS((Net net, double tolerance)) ;
extern	Real	aveOutputActivity	ARGS((Net net, double thresh)) ;
extern	Real	aveOutputStress		ARGS((Net net)) ;
extern	Real	groupAveOutputStress	ARGS((Group group)) ;
extern	void	groupIncEnergy		ARGS((Group group, void* data)) ;
extern	Real	groupEnergy		ARGS((Group group)) ;
extern	Real	netEnergy		ARGS((Net net)) ;
extern	Real	outputProximity		ARGS((Net net)) ;
extern	Boolean	outputCorrect		ARGS((Net net, 
					      double tolerance)) ;
extern	Boolean	outputCorrectWTA	ARGS((Net net, 
					      double tolerance)) ;
extern	Boolean	groupOutputCorrect	ARGS((Group group,
					      double tolerance)) ;
extern	Boolean	groupOutputCorrectWTA	ARGS((Group group,
					      double tolerance)) ;
extern	Boolean	attentionCorrectWTA	ARGS((Group Attention, Group Done,
					      double tolerance)) ;
extern	Boolean	outputClean		ARGS((Net net, 
					      double tolerance)) ;
extern	Boolean	groupOutputClean	ARGS((Group group,
					      double tolerance)) ;
extern	Boolean	groupOutputCleanWTA	ARGS((Group group,
					      double tolerance)) ;
extern	Boolean	groupOutputCleanOff	ARGS((Group group,
					      double tolerance)) ;
extern	Boolean	attentionCleanWTA	ARGS((Group Attention, Group Done,
					      double tolerance)) ;
extern	Real	sumOutput		ARGS((Group group)) ;
extern	Real	maxOutput		ARGS((Group group)) ;
extern	Real	maxOutputAt		ARGS((Group group, int time)) ;
extern	int	maxOutputPosition	ARGS((Group group)) ;
extern	int	maxTargetPosition	ARGS((Group group)) ;
extern	void	cleanupWTA		ARGS((Group group)) ;
extern	void	cleanupThreshold	ARGS((Group group, double thresh)) ;
extern	void	SMnormalizeOutput	ARGS((Group group)) ;
extern  void	SMnormalizeInputDeriv	ARGS((Group group)) ;
extern	Real	SMnormalizedRespProb	ARGS((Group group, 
					      double criterion, double sd)) ;
extern	Boolean	SMoutputClean		ARGS((Group group, double tolerance)) ;
extern	Boolean	SMnsylPronounceCorrect	ARGS((Net net, String phonemes)) ;

extern	void	invertOutput		ARGS((Unit unit, void* data)) ;
extern	void	corruptOutput		ARGS((Unit unit, void* data)) ;
extern	void	printVector		ARGS((Real* v, int n, String str)) ;

extern	void	Fixate			ARGS((Group Letters,
					      Group Attention,
					      int   saccade,
					      int   attention,
					      int   time)) ;

extern	Real	featLMSE		ARGS((Net net)) ;
extern	Boolean	outputCorrectReal	ARGS((Net net, 
					      double binTol, double realTol)) ;


#define MIN_UNIT_OUTPUT	0.000001
#define MAX_UNIT_OUTPUT	0.999999

#define MAX_UNIT_INPUT	15.0

/*
#define WITH_PROBABILITY(p)	((p)==1.0 || ((p)>0.0 && IRandReal()<(p)))
#define uniformNoise(lo,hi)	(IRandReal()*((hi)-(lo))+(lo))
#define randomInteger(hi)	(((int)(IRandReal()*(hi)))%(hi))
*/
extern	double	drand48			ARGS(()) ;
#define WITH_PROBABILITY(p)	((p)==1.0 || ((p)>0.0 && ((Real)drand48())<(p)))
#define uniformNoise(lo,hi)	(((Real)drand48())*((hi)-(lo))+(lo))
#define randomInteger(hi)	(((int)(((Real)drand48())*(hi)))%(hi))

#define MIN(x, y)		((x) < (y) ? (x) : (y))
#define MAX(x, y)		((x) > (y) ? (x) : (y))
#define CLIP(x, min, max)	((x) > (max) ? (max) : ((x) < (min) ? (min) : (x)))

#define IS_BIDIRECTIONAL(link)		((link)->type & BIDIRECTIONAL)


#define EXP(x) exp((double)(x))
#define LOG(x) log((double)(x))

#define sigmoid(x)	(1.0/(1.0+EXP(-(x))))
#define sigmoidDeriv(x)	((x)*(1.0 - (x)))

#define sigmoidInv(y)	((y)<MIN_UNIT_OUTPUT ? (-MAX_UNIT_INPUT) :	\
			 ((y)>MAX_UNIT_OUTPUT ? MAX_UNIT_INPUT : 		\
			  LOG((y)/(1.0-(y)))))

#define expUnit(x)      ((x) > 0.0 ? 1.0-EXP(-(x)) : 0.0)
/* NOTE: assumes constant derivative of 1.0 for negative input */
#define expUnitDeriv(x) ((x) > 0.0 ? EXP(-(x))     : 1.0)
/* NOTE: only non-negative inverses */
#define expUnitInv(y)   (-LOG(1.0-(y)))	


/* NOTE: this returns stress in base 2 */
#define unitStress(y)	((y) < MIN_UNIT_OUTPUT || (y) > MAX_UNIT_OUTPUT ? 1.0 : \
			 (((y)*LOG(y)+(1.0-(y))*LOG(1.0-(y)))/-LOG(0.5))+1.0)

#define CROSS_ENTROPY	0
#define SQUARED_ERROR	1

#define SIMPLE_INFO(output, target)		\
  (LOG((target)/(output))*(target) 		\
   + LOG((1.0 - (target))/(1.0 - (output)))*(1.0 - (target)))

#define INFO(output, target)			\
  ((target) == 0.0 ? -LOG(1.0 - MIN(output,MAX_UNIT_OUTPUT)) \
   : (target) == 1.0 ? -LOG(MAX(output,MIN_UNIT_OUTPUT)) : SIMPLE_INFO(output, target))

#define INFO_DERIV(output, target)				\
  ((target) == 0.0 ? 1.0 / (1.0 - MIN(output,MAX_UNIT_OUTPUT)) :	\
   ((target) == 1.0 ? -1.0 / MAX(output,MIN_UNIT_OUTPUT) :		\
    ((1.0-(target))/(1.0 - MIN(output,MAX_UNIT_OUTPUT))) - ((target) / MAX(output,MIN_UNIT_OUTPUT))))

#define DIVERGENCE(y, d)		((d) == 0.0 ? 0.0 : -(d)*log(MIN_UNIT_OUTPUT+y))
#define DIVERGENCE_DERIV(y, d)		((d) == 0.0 ? 0.0 : -(d)/(MIN_UNIT_OUTPUT+y))

/****************************************************************************/

/*
 * For link length and position-based lesioning
 */
#define MxPosition(unit)	((unit)->activityColumn)
#define MyPosition(unit)	((unit)->activityRow)


#define SQ_DIST(x1,y1,x2,y2)	(sq(x1-x2)+sq(y1-y2))

#define DISTANCE(x1,y1,x2,y2)	sqrt(SQ_DIST(x1,y1,x2,y2))

#define GAUSS_2D(x1,y1,x2,y2,sd)					\
     EXP(-0.5*SQ_DIST(x1,y1,x2,y2)/sq(sd))

#define GAUSS_DIST(x1,y1,x2,y2,sd)					\
     (1.0/GAUSS_2D(x1,y1,x2,y2,sd))

#define MINR_DIST(x1,y1,x2,y2,r)					\
	pow(pow(fabs((x1)-(x2)),(double)r)+pow(fabs((y1)-(y2)),(double)r),(double)(1.0/(r)))

extern void	setLinkLength() ;
extern Real	getLinkLength() ;

extern int	unitXPosition() ;
extern int	unitYPosition() ;

#endif				/* UTILITIES_H */
