
/**********************************************************************
 * $Id: bp.c,v 1.9 91/09/26 11:05:14 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	  Copyright 1990,1991 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/example.h>
#include <xerion/utilsDCP.h>
#include "barak.h"

#define DB(x) x ;

static	Boolean	_presentedAssociate ;
static	Real	_decisionOldBaseline = -1.0 ;

struct TRACE    inRelaxation ;
struct TRACE    primingInRelaxation ;
struct TRACE	forwardEachTime ;
struct TRACE    afterExample ;

static Proc	updateNetActivity      ARGS((Net	net)) ;
static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initGroup    		ARGS((Group	 group)) ;
static void	deinitGroup    		ARGS((Group	 group)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExample	    	ARGS((Example    example)) ;
static void	deinitExample		ARGS((Example    example)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;

/**********************************************************************
 * 				Initialization
 **********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateGroupHook(initGroup) ;
  setDestroyGroupHook(deinitGroup) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleHook(initExample) ;
  setDestroyExampleHook(deinitExample) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}
/***********************************************************************/
static void	initNet (net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  /* non-zero defaults */
  MinitialUnitOutput(net)	= 0.2 ;
  MinitializeUnitsProb(net)	= 1.0 ;
  MinputStrength(net)		= 1.0 ;
  MinputGain(net)		= 1.0 ;
  MmaxTimePerSlice(net)		= 1.0 ;
  Mdt(net)			= 1.0 ;
  MdecisionDtFactor(net)	= 1.0 ;
  MuseStoppingCriteria(net)	= TRUE ;
  MlengthScaleDerivs(net)	= 1 ;
  MgraceTimeFirstSlice(net)	= -1.0 ; /* will default to PerSlice values */
  MminTimeFirstSlice(net)	= -1.0 ;
  MmaxTimeFirstSlice(net)	= -1.0 ;
  MgraceTimeLastSlice(net)	= -1.0 ; /* will default to PerSlice values */
  MminTimeLastSlice(net)	= -1.0 ;
  MmaxTimeLastSlice(net)	= -1.0 ;
  MassocMinProb(net)		= 1.0 ;
  MminDecrements(net)		= 2 ;
  MuseMinimize(net)		= TRUE ;
  MexampleInStrength(net)	= 1.0 ;
}
/**********************************************************************/
static void	deinitNet (net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}
/***********************************************************************/
static void	initGroup(group)
  Group	group ;
{
  group->extension = (GroupExtension)calloc(1, sizeof(GroupExtensionRec)) ;

  MinputClamping(group)		= -1 ;
  MminBias(group)		= -MAXFLOAT ;
  MmaxBias(group)		=  MAXFLOAT ;
  MminIncomingWeight(group)	= -MAXFLOAT ;
  MmaxIncomingWeight(group)	=  MAXFLOAT ;
  MminOutgoingWeight(group)	= -MAXFLOAT ;
  MmaxOutgoingWeight(group)	=  MAXFLOAT ;
  MgroupXPosition(group)	= -1 ;
  MgroupYPosition(group)	= -1 ;
}

static void	deinitGroup(group)
  Group	group ;
{
  if (group->extension != NULL)
    free(group->extension) ;
}
/***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;

  MerrorFactor(unit) = 1.0 ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}
/***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}
/***********************************************************************/
static void	initExample(example)
  Example  example ;
{
  int i ;
  example->extension = 
    (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;

  example->extension->numAssocs = 0 ;
  for (i = 0 ; i < MAX_ASSOCS ; i++)
    example->extension->assoc[i] = NO_ASSOC ;
}

static void	deinitExample(example)
  Example  example ;
{
  if (example->extension != NULL)
    free(example->extension) ;
}
/***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleSetExtension)calloc(1, sizeof(ExampleSetExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}
/**********************************************************************
 * 				Utilities
 **********************************************************************/
static void	clampInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Net	net = unit->net ;
  Mask	clamping = (MinputClamping(unit->group) >= 0 ?	/* init value = -1 */
		    MinputClamping(unit->group)      :	/* override */
		    MinputClamping(net)) ;		/* default */
  int   s = (int)data ;
  Real	y ;
  Real	in_y   = Mi_slice(unit)[s] ;
  Real	init_y = MinitialUnitOutput(net) ;
  Real	inStr  = MinputStrength(net) * MexampleInStrength(net) ;

  if (!MinitializeUnits(net) || s > 0) init_y = My(unit) ;
  Mext(unit) = 0.0 ;
  if (isNaN(in_y)) {
    Mstatus(unit) &= ~CLAMPED ;
    y = init_y ;
  } else {
    if (in_y < 0.0) {	/* soft-clamped input */
      in_y  += 2.0 ;	/* convention is that [0,1] is scaled to [-2,-1] in file */
      clamping |= SOFT ;
    }
    if (MinputNoise(net) > 0.0) {
      if      (in_y == 0.0) in_y  = fabs(gaussianNoise(MinputNoise(net))) ;
      else if (in_y == 1.0) in_y -= fabs(gaussianNoise(MinputNoise(net))) ;
      else {
	in_y += gaussianNoise(MinputNoise(net)) ;
	in_y  = CLIP(in_y, 0.0, 1.0) ;
      }
    }
    y = in_y ;
    if (clamping&SOFT) {
      /* make soft clamping relative to NEUTRAL (instead of current output) */
      /* y = init_y + inStr*(in_y - init_y) ;  */
      y = MinitialUnitOutput(net) + inStr*(in_y - MinitialUnitOutput(net)) ;
      Mext(unit) = sigmoidInv(y) ;
      if (MtimeAverageInputs(net)) Mx(unit) = sigmoidInv(init_y) ;
      if (! (clamping&ONESHOT)) y = init_y ;
    }
    /* this is reversed because clamping=0 is normal clamping */
    if (clamping) Mstatus(unit) &= ~CLAMPED ;
    else          Mstatus(unit) |=  CLAMPED ;
  }
  My(unit) = My_slice(unit)[s] = y ;
}
/**********************************************************************/
static void	setTarget(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  Md(unit) = Md_slice(unit)[(int)slice] ;
}
/**********************************************************************/
static void	forceTarget(unit, tolerance)
  Unit		unit ;
  void		*tolerance ;
{
  Real	d 	= Md(unit) ;
  Real	tol	= *((Real *)tolerance) ;

  if (!isNaN(d) && fabs(My(unit) - d) > tol) 
    My(unit) = fabs(d - tol) ;
}
/**********************************************************************/
static void	saveOutputAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  My_t(unit)[(int)tick] = My(unit) ;
}
/**********************************************************************/
static void	restoreOutputAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  My(unit) = My_t(unit)[(int)tick] ;
}
/**********************************************************************/
static void	saveTargetAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  Md_t(unit)[(int)tick] = Md(unit) ;
}
/**********************************************************************/
static void	restoreTargetAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  Md(unit) = Md_t(unit)[(int)tick] ;
}
/**********************************************************************/
static void	saveNaNTargetAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  setNaN(Md_t(unit)[(int)tick]) ;
}
/**********************************************************************/
static void	saveSliceOutput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  My_slice(unit)[(int)slice] = My(unit) ;
}
/**********************************************************************/
static void	restoreSliceOutput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  My(unit) = My_slice(unit)[(int)slice] ;
}
/**********************************************************************/
static void	saveInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Moldx(unit) = Mx(unit) ;
}
/**********************************************************************/
static void	zeroInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mx(unit) = 0.0 ;
}
/**********************************************************************/
static void	addExternalInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mx(unit) += Mext(unit) ;
}
/**********************************************************************/
static void	zeroDerivs(unit, data)
  Unit		unit ;
  void		*data ;
{
  MdEdy(unit) = MdEdx(unit) = McdEdy(unit) = 0.0 ;
}
/**********************************************************************/
static void	printOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.2f%s", My(unit), (Md(unit)==1.0?"+":Md(unit)==0.0?"-":" ")) ;
}

/**********************************************************************/
static void	displayNet(net)
  Net	net ;
{
  int	idx, delayCount = MdelayCount(net), tmp = 0 ;

  netForAllOtherUnits(net, BIAS, (UnitDataProc)saveSliceOutput, (void *)Mslice(net)) ;
  IDoTrace(&inRelaxation) ;
  if (MrunPrimingInRelaxationTrace(net)) IDoTrace(&primingInRelaxation) ;
  /* this nonsense is just to slow the display and fool the compiler */
  for (idx = 0, tmp = net->currentTime ; idx < delayCount*10000 ; idx++) net->currentTime++ ;
  net->currentTime = tmp ;
}
/**********************************************************************
 * 				Error
 **********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = My(unit), target = Md(unit) ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  if (MnonExtremalZeroTarget(net) > 0.0 && target == 0.0) {
    if (output <= MnonExtremalZeroTarget(net)) return 0.0 ;
    else target = MnonExtremalZeroTarget(net) ;
  }
  if (MnonExtremalOneTarget(net) > 0.0 && target == 1.0) {
    if (output >= MnonExtremalOneTarget(net)) return 0.0 ;
    else target = MnonExtremalOneTarget(net) ;
  }

  if (fabs(output-target) > McorrectTolerance(net)) MoutputsIncorrect(net)++ ;

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    error = INFO(output, target) ; 
/*
    error = CE(output, target) ;
*/
    break ;
  case SQUARED_ERROR:
    error = (output - target) * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  error *= MerrorFactor(unit) ;
  return error ;
}
/**********************************************************************/
static Real	unitErrorDeriv(unit, tick)
  const Unit	unit ;
  int		tick ;
{
  Net	net = unit->net ;
  Real	error, output = My(unit), target = Md_t(unit)[tick] ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  if (MnonExtremalZeroTarget(net) > 0.0 && target == 0.0) {
    if (output <= MnonExtremalZeroTarget(net)) return 0.0 ;
    else target = MnonExtremalZeroTarget(net) ;
  }
  if (MnonExtremalOneTarget(net) > 0.0 && target == 1.0) {
    if (output >= MnonExtremalOneTarget(net)) return 0.0 ;
    else target = MnonExtremalOneTarget(net) ;
  }
  
  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    error = INFO_DERIV(output, target) ; 
/*
    error = CE_DERIV(output, target) ;
*/
    break ;
  case SQUARED_ERROR:
    error = 2.0 * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  error *= MerrorFactor(unit) ;
  return error ;
}
/**********************************************************************/
static void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real diff ;

  *(Real *)data += unitError(unit) ;
  if (isNaN(unit->target)) return ;
  diff = fabs(unit->output - unit->target) ;
  if (diff > MworstOutput(unit->net)) MworstOutput(unit->net) = diff ;
}

Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  groupForAllUnits(group, (UnitDataProc)incrementError, &error) ;
  return error ;
}

/**********************************************************************/
static void	groupIncrementUnitCost(group, data)
  Group		group ;
  void		*data ;
{
  int	i ;
  Real	mean = 0.0 ;
  Real	prior = MunitPrior(group) ;
  Unit	*unit ;

  if (MunitCost(group) == 0.0) return ;
  for (i = 0, unit = group->unit ; i < group->numUnits ; i++, unit++) 
    mean += My(*unit) ;
  MmeanOutput(group) = mean /= (Real)(group->numUnits) ;
  if (mean > prior)
    *(Real *)data += MunitCost(group) * CE(mean, prior) ;
}
/**********************************************************************/
static void	groupInjectUnitCost(group, data)
  Group		group ;
  void		*data ;
{
  int	i ;
  Real	mean = 0.0, dCost ;
  Real	prior = MunitPrior(group) ;
  Unit	*unit ;

  if (MunitCost(group) == 0.0) return ;
  for (i = 0, unit = group->unit ; i < group->numUnits ; i++, unit++) 
    mean += My(*unit) ;
  mean /= (Real)(group->numUnits) ;
  if (mean > prior) {
    dCost = MunitCost(group) * CE_DERIV(mean, prior) ;
    for (i = 0, unit = group->unit ; i < group->numUnits ; i++, unit++) 
      McdEdy(*unit) += dCost ;
  }
}

/**********************************************************************
 * 				Forward
 **********************************************************************/
/**********************************************************************/
static void	timeAverageInput(unit, dt)
  Unit		unit ;
  void		*dt ;
{
  if (!(Mstatus(unit)&CLAMPED))
    Mx(unit) = Moldx(unit) + *(Real *)dt * (Mx(unit) - Moldx(unit)) ;
}
/**********************************************************************/
static	void	netAccumulateInput(net)
  Net		net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*wt, **post_x, **pre_y ;

  if (MtimeAverageInputs(net))
    netForAllUnits(net, ALL, (UnitDataProc)saveInput, NULL) ;
  netForAllUnits(net, ALL, (UnitDataProc)zeroInput, NULL) ;
  netForAllUnits(net, INPUT, (UnitDataProc)addExternalInput, NULL) ;

  for (idx=0, pre_y=MpreOutputs(net), post_x=MpostInputs(net), wt=Mwts(net) ;
       idx < numCompiledLinks ; 
       idx++)
    *(*post_x++) += *(*pre_y++) * (*wt++) ;

  if (MtimeAverageInputs(net))
    netForAllUnits(net, ALL, (UnitDataProc)timeAverageInput, (void *)&Mcdt(net)) ;
}
/**********************************************************************/
int	_numOutputs ;
static void	computeOutput(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  Net	net 	= unit->net ;
  Real	noise	= MunitNoise(net) ;
  Real	x, sigx, change ;

  if(isClampedUnit(unit) || isLesionedUnit(unit)) return ;
  x = Mx(unit) * MinputGain(net) ;
  if (noise > 0.0) x += gaussianNoise(noise) ;
  Msigx_t(unit)[(int)tick] = sigx = sigmoid(x) ;

  if (MoutputChangeUseAll(net) ||
      (MoutputChangeByGroup(net) && MuseOutputChange(unit->group)) ||
      (!MoutputChangeByGroup(net) && unit->group->type&OUTPUT)) {
    change = fabs(sigx-My(unit)) ;
    if (MoutputChangeUseAve(net)) {
      MoutputChange(net) += change ;   _numOutputs++ ;
    } else if (change > MoutputChange(net)) {
      MoutputChange(net)  = change ;
    }
    if (MinitialPhonemeCriterion(net)) {
      /* this is just to determine if unit is in onset, vowel, or coda */
      int i ; 
      for (i = 0 ; i < unit->group->numUnits && unit != unit->group->unit[i] ; i++) ;
      if (i == unit->group->numUnits) IErrorAbort("did not find unit %s", unit->name) ;

      if (i <= NSYL_LAST_ONSET) {
	if (MoutputChangeUseAve(net))        MonsetChange(net) += change ;
	else if (change > MonsetChange(net)) MonsetChange(net)  = change ;
      } else if (i <= NSYL_LAST_VOWEL) {
	if (MoutputChangeUseAve(net))        MvowelChange(net) += change ;
	else if (change > MvowelChange(net)) MvowelChange(net)  = change ;
      } else {
	if (MoutputChangeUseAve(net))	     McodaChange(net) += change ;
	else if (change > McodaChange(net))  McodaChange(net)  = change ;
      }
    }
  }
  if (MtimeAverageInputs(net)) My(unit) = sigx ;
  else			       My(unit) += Mcdt(net) * (sigx - My(unit)) ;
}
/**********************************************************************/
static	void	netForward (net, tick, timeThisSlice, graceTimeThisSlice)
  Net	net ;
  int	tick ;
  Real	timeThisSlice, graceTimeThisSlice ;
{
  int	t = tick % MmaxTicks(net) ;

  _numOutputs = 0 ;
  MworstOutput(net)		= 0.0 ;
  MinstantaneousError(net)	= 0.0 ;
  MoutputChange(net)		= 0.0 ;
  if (MinitialPhonemeCriterion(net)) MonsetChange(net) = MvowelChange(net) = McodaChange(net) = 0.0 ;
  MoutputsIncorrect(net)	= 0 ;
  netAccumulateInput (net) ;
  netForAllOtherUnits(net, BIAS,   (UnitDataProc)computeOutput,  (void *)t) ;
  netForAllUnits     (net, ALL,    (UnitDataProc)saveOutputAt,   (void *)t) ;
  netForAllUnits     (net, OUTPUT, (UnitDataProc)incrementError, (void *)&MinstantaneousError(net)) ;
  if (timeThisSlice > graceTimeThisSlice) {
    netForAllUnits   (net, OUTPUT, (UnitDataProc)saveTargetAt,   (void *)t) ;
    net->error += MinstantaneousError(net) ;
  } else
    netForAllUnits   (net, OUTPUT, (UnitDataProc)saveNaNTargetAt,(void *)t) ;
}
/**********************************************************************
 * 				Backward
 **********************************************************************/
static void	injectError(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  if (isLesionedUnit(unit)) return ;
  McdEdy(unit) += unitErrorDeriv(unit, (int)tick) ;
}
/**********************************************************************/
static void	invertError(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  if (isLesionedUnit(unit)) return ;
  MdEdy(unit) += Mcdt(unit->net) * (McdEdy(unit) - MdEdy(unit)) ;
  MdEdx(unit)  = MdEdy(unit) * sigmoidDeriv(Msigx_t(unit)[(int)tick]) ; 
  McdEdy(unit) = 0.0 ;
}
/**********************************************************************/
static void	invertErrorTAI(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	cdEdx ;

  if (isLesionedUnit(unit)) return ;
  MdEdy(unit)  = McdEdy(unit) ;
  cdEdx        = MdEdy(unit) * sigmoidDeriv(My(unit)) ; 
  MdEdx(unit) += Mcdt(unit->net) * (cdEdx - MdEdx(unit)) ;
  McdEdy(unit) = 0.0 ;
}

/**********************************************************************/
static	void	netDistributeError(net)
  Net	net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real 	*wt, *dv, **pre_cdEdy, **post_dEdx, **pre_y ;

  /* factor of dt has been pulled out of dv and applied by netAdjustWeightDerivs */
  for (idx=0, pre_cdEdy=MpreOutputDerivs(net), pre_y=MpreOutputs(net), post_dEdx=MpostInputDerivs(net), wt=Mwts(net), dv=Mdvs(net) ;
       idx < numCompiledLinks ; 
       idx++) {
    Real dEdx        = *(*post_dEdx++) ;
    *(*pre_cdEdy++) += dEdx *  (*wt++) ;
/*
    if (*wt++ > 0.0) *(*pre_cdEdy++) += dEdx ;
    else             *(*pre_cdEdy++) -= dEdx ;
*/
    (*dv++)        += dEdx * *(*pre_y++) ;
  }
}
/***********************************************************************/
static void	netBackward (net, tick, inject)
  Net		net ;
  int		tick ;
  Boolean	inject ;
{
  int	thisTick = tick     % MmaxTicks(net) ;
  int	lastTick = (tick-1) % MmaxTicks(net) ;

  /* NOTE: not worth restoring Md_t[tick] and sigIn_t[thisTick] */
  if (inject)
    netForAllUnits      (net, OUTPUT, (UnitDataProc)injectError,   (void *)thisTick) ;
  if (MtimeAverageInputs(net)) 
    netForAllOtherUnits (net, BIAS,   (UnitDataProc)invertErrorTAI, NULL) ;
  else 
    netForAllOtherUnits (net, BIAS,   (UnitDataProc)invertError,   (void *)thisTick) ;
  netForAllOtherUnits (net, BIAS,   (UnitDataProc)restoreOutputAt, (void *)lastTick) ;
  netDistributeError    (net) ;
}
/***********************************************************************
 * 				Iterate
 ***********************************************************************/
static void	initOutput(unit, tick) 
  Unit		unit ;
  void		*tick ;
{
  if (isLesionedUnit(unit)) return ;
  My(unit)              = MinitialUnitOutput(unit->net) ;
  My_t(unit)[(int)tick] = MinitialUnitOutput(unit->net) ;
}
/***********************************************************************/
static	void	netPresentExample(net, tick)
  Net	net ;
  int	tick ;
{
  int t = tick % MmaxTicks(net) ;
  int s = Mslice(net) ;

  netForAllUnits(net, INPUT,  (UnitDataProc)clampInput,   (void *)s) ;
  netForAllUnits(net, INPUT,  (UnitDataProc)saveOutputAt, (void *)t) ;
  netForAllUnits(net, OUTPUT, (UnitDataProc)setTarget,    (void *)s) ;
}
/**********************************************************************/
static Real	exponentialNoise(tau)
  Real	tau ;
{
  int n = 0 ;
  if (tau <= 0.0) {
    IErrorAbort("cannot call exponentialNoise with tau < 0 (%f)", tau) ;
    return 0 ;
  }
  while (! WITH_PROBABILITY(tau)) n++ ;
  return (Real)n ;
}

/**********************************************************************/
void	netIterate(net, training)
  Net		net ;
  Boolean	training ;
{
  int	forwTick, backTick, ticksSinceBP, outputPos, numSlicesCorrect=0  ;
  Real	forwTime, backTime ;
  Real	timeWindow, bpWindow, timeSinceBP, dt = Mdt(net) ;

  MnumSlices(net) = lastTargetSlice(net) ;
  MfeatError(net) = 0.0 ;
  /* correction for (relatively) large dt - good value is 0.4 */
  Mcdt(net) = (MdtCorrection(net) > 0.0 ? Mdt(net)*(1.0-Mdt(net)*MdtCorrection(net)) : Mdt(net)) ;
  if (training) {
    netCompileTicks(net) ;
    timeWindow = (MtimeWindow(net) > 0.0 ? MtimeWindow(net) : 
		  (MmaxTimeFirstSlice(net) >= 0.0 ? MmaxTimeFirstSlice(net) : MmaxTimePerSlice(net)) + 
		  (MnumSlices(net)-1)*MmaxTimePerSlice(net)) ;
    bpWindow   = (MbpWindow(net)   > 0.0 ? MbpWindow(net)   : timeWindow) ;
  }
  if (MinitializeUnits(net)) {
    /* HIDDEN AND OUTPUT UNITS ARE SPLIT BECAUSE OUTPUT UNITS MAY ALSO BE INPUT UNITS */
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS, (UnitDataProc)initOutput, 0) ;
    netForAllUnits     (net, OUTPUT,            (UnitDataProc)initOutput, 0) ;

    if (MtimeAverageInputs(net)) {
      netForAllOtherUnits(net, INPUT|OUTPUT|BIAS, (UnitDataProc)invertOutput, NULL) ;
      netForAllUnits     (net, OUTPUT,            (UnitDataProc)invertOutput, NULL) ;
    }
  } else
    netForAllOtherUnits(net, BIAS, (UnitDataProc)saveOutputAt, 0) ;
  netForAllUnits(net, BIAS, (UnitDataProc)saveOutputAt, 0) ;
  /*
   * For each input slice...
   */
  for (Mslice(net) = 0, forwTick = 1, forwTime = dt, ticksSinceBP = 1, timeSinceBP = dt ;
       Mslice(net) < MnumSlices(net) ;
       Mslice(net)++) {
    int		ticksThisSlice = 1 ;
    Real	timeThisSlice = dt, timeCorrect = 0.0 ;
    Real	graceTimeThisSlice = MgraceTimePerSlice(net) ;	/* defaults */
    Real	minTimeThisSlice   = MminTimePerSlice(net) ;
    Real	maxTimeThisSlice   = MmaxTimePerSlice(net) ;
    Boolean	sliceDone = FALSE, correct, thisCorrect ;
    Real	worstOutputLastTick = 0.0 ;
    int		numWorstOutputDecrements = 0 ;
    Boolean	worstOutputTrough = FALSE ;

    /* allow variable (random) duration of first and last slice */
    if (Mslice(net) == 0) {
      if (MgraceTimeFirstSlice(net) >= 0.0) graceTimeThisSlice = MgraceTimeFirstSlice(net) ;
      if (MminTimeFirstSlice(net)   >= 0.0) minTimeThisSlice   = MminTimeFirstSlice(net) ;
      if (MmaxTimeFirstSlice(net)   >= 0.0) maxTimeThisSlice   = MmaxTimeFirstSlice(net) ;
      if (MtimeFirstSliceMin(net) < MtimeFirstSliceMax(net)) 
	/* NOTE: this overrides minTimeFirstSlice and maxTimeFirstSlice */
	minTimeThisSlice = maxTimeThisSlice = 
	  /* exponential distribution if associate, otherwise uniform */
	  (_presentedAssociate & MtimeFirstSliceTau(net) > 0.0 ?
	   MtimeFirstSliceMin(net) + dt*exponentialNoise(dt*MtimeFirstSliceTau(net)) :
	   uniformNoise(MtimeFirstSliceMin(net), MtimeFirstSliceMax(net))) ;
    } else if (Mslice(net) == MnumSlices(net)-1) {
      if (MgraceTimeLastSlice(net) >= 0.0) graceTimeThisSlice = MgraceTimeLastSlice(net) ;
      if (MminTimeLastSlice(net)   >= 0.0) minTimeThisSlice   = MminTimeLastSlice(net) ;
      if (MmaxTimeLastSlice(net)   >= 0.0) maxTimeThisSlice   = MmaxTimeLastSlice(net) ;
      if (MtimeLastSliceMin(net) < MtimeLastSliceMax(net)) 
	/* NOTE: this overrides minTimeLastSlice and maxTimeLastSlice */
	minTimeThisSlice = maxTimeThisSlice = 
	  /* exponential distribution if tau is set, otherwise uniform */
	  (MtimeLastSliceTau(net) > 0.0 ?
	   MtimeLastSliceMin(net) + dt*exponentialNoise(dt*MtimeLastSliceTau(net)) :
	   uniformNoise(MtimeLastSliceMin(net), MtimeLastSliceMax(net))) ;
    }
    MdecisionResponse(net) = DECISION_NORESP ;
    if (MdecisionResponseCriterion(net) > 0.0) {
      int     numTicks = ceil(maxTimeThisSlice/dt) ;
      if (numTicks > MdecisionMaxTicks(net)) {
	if (MdecisionBaseline(net)) free(MdecisionBaseline(net)) ;
	MdecisionBaseline(net)    = (Real *)calloc(numTicks, sizeof(Real)) ;
	MdecisionMaxTicks(net)    = numTicks ;
	MdecisionNumAveraged(net) = 0 ;
      }
      if (MdecisionNumAveraged(net) == 0 || MdecisionInitialBaseline(net) != _decisionOldBaseline) {
	int i ; for (i = 0 ; i < MdecisionMaxTicks(net) ; i++) MdecisionBaseline(net)[i] = MdecisionInitialBaseline(net) ;
	_decisionOldBaseline = MdecisionInitialBaseline(net) ;
      }
      MdecisionYesOutput(net) = MdecisionNoOutput(net) = 0.0 ;
    }
    netPresentExample(net, forwTick-1) ;
    /*
     * For each time tick...
     */
    for ( ; !sliceDone ; 
	 forwTick++,       forwTime=forwTick*dt, 
	 ticksThisSlice++, timeThisSlice=ticksThisSlice*dt, 
	 ticksSinceBP++,   timeSinceBP=ticksSinceBP*dt) {
      /*
       * Run forward pass
       */
      if (MinputNoiseEachTick(net) && MinputNoise(net) > 0.0)
	netForAllUnits(net, INPUT, (UnitDataProc)clampInput, (void *)Mslice(net)) ;

      netForward(net, forwTick, timeThisSlice, graceTimeThisSlice) ;

      if (MoutputChangeUseAve(net))         MoutputChange(net)  /= ((Real)_numOutputs) ;
      if (MoutputStressInRelaxation(net))   MoutputStress(net)   = aveOutputStress(net) ;
      if (MoutputActivityInRelaxation(net)) MoutputActivity(net) = aveOutputActivity(net,MoutputActivityThresh(net)) ;

      if (MinitialPhonemeCriterion(net)) {
	if (MoutputChangeUseAve(net)) {
	  MonsetChange(net) /= 23.0 ;
	  MvowelChange(net) /= 14.0 ;
	  McodaChange(net)  /= 24.0 ;
	}
	if (MonsetTime(net) == 0.0 && MonsetChange(net) <= MoutputChangeTolerance(net))
	  MonsetTime(net) = forwTime ;
	if (MvowelTime(net) == 0.0 && MvowelChange(net) <= MoutputChangeTolerance(net))
	  MvowelTime(net) = forwTime ;
	if (McodaTime(net)  == 0.0 && McodaChange(net) <= MoutputChangeTolerance(net))
	  McodaTime(net)  = forwTime ;
	if (MallTime(net)  == 0.0 && MoutputChange(net) <= MoutputChangeTolerance(net))
	  MallTime(net)  = forwTime ;
      }

      if (MfeatErrorInRelaxation(net)) {
	MfeatError(net)     = featLMSE(net) ;
	MfeatPhoneme(net)   = featLMSEphoneme ;
      }
      if (!Mrunning(net)) displayNet(net) ;
      correct = thisCorrect = (training ?
			       outputCorrect(net, McorrectTolerance(net)) :
			       outputCorrectReal(net, McorrectTolerance(net), 0.1)) ;
      if (correct) {
	timeCorrect += dt ;
	if (timeCorrect < MminTimeCorrect(net)) correct = FALSE ;
      } else timeCorrect = 0.0 ;

      /***************************************************/
      /* this is only relevant to learning trajectories */
      if (MtroughTolerance(net) > 0.0 || MfeatErrorInRelaxation(net)) {
	
	if (MfeatErrorInRelaxation(net)) MworstOutput(net) = MfeatError(net) ;

	if (MworstOutput(net) < worstOutputLastTick) 
	  numWorstOutputDecrements++ ;
	else if ((MfeatErrorInRelaxation(net) ||
		  MworstOutput(net) < MtroughTolerance(net)) &&
		 numWorstOutputDecrements >= MminDecrements(net)) {	/* down... up */
	  worstOutputTrough = TRUE ;
	  if (timeThisSlice > graceTimeThisSlice) {
	    /* undo targets and error for this tick set in netForward */
	    net->error -= MinstantaneousError(net) ;
	    netForAllUnits (net, OUTPUT, (UnitDataProc)saveNaNTargetAt,
			    (void *)(forwTick%MmaxTicks(net))) ;
	  }
	} else
	  numWorstOutputDecrements = 0 ;
	worstOutputLastTick = MworstOutput(net) ;
      }
      /***************************************************
       * THIS IS FOR YES/NO DECISIONS 
       ***************************************************/
      if (MuseStoppingCriteria(net) &&
	  MdecisionResponseCriterion(net) > 0.0 && 
	  timeThisSlice >= minTimeThisSlice) {	/* only start integrating at minTimeSlice */
	int	tick  = (int)(timeThisSlice/dt) - 1 ;
	Real	this  = ( MoutputStressInRelaxation(net)   ? MoutputStress(net) :
			 (MoutputActivityInRelaxation(net) ? MoutputActivity(net) : 0.0)) ;
	Real	base  = MdecisionBaseline(net)[tick] ;
	Real	yesIn, noIn ;
	if (MdecisionMinInput(net) < MdecisionMaxInput(net)) {
	  yesIn = (this - MdecisionMinInput(net))/(MdecisionMaxInput(net) - MdecisionMinInput(net)) ;
	  noIn  = (base  - MdecisionMinInput(net))/(MdecisionMaxInput(net) - MdecisionMinInput(net)) ;
	} else {
	  yesIn = this ;  noIn = base ;
	}
	MdecisionYesOutput(net) += DECISION_UPDATE(net, yesIn, MdecisionYesOutput(net), MdecisionNoOutput(net)) ;
	MdecisionNoOutput(net)  += DECISION_UPDATE(net, noIn,  MdecisionNoOutput(net),  MdecisionYesOutput(net)) ;
	if (MdecisionResponse(net) == DECISION_NORESP) 
	  if (MdecisionYesOutput(net) >= MdecisionResponseCriterion(net)) {
	    MdecisionResponse(net) = DECISION_YES ;
	    MdecisionRT(net) = timeThisSlice ; 
	  } else if (MdecisionNoOutput(net) >= MdecisionResponseCriterion(net)) {
	    MdecisionResponse(net) = DECISION_NO ;
	    MdecisionRT(net) = timeThisSlice ; /* starts at dt */
	  }
	if (MdecisionUpdateBaseline(net)) 
	  MdecisionBaseline(net)[tick] += (this-base)/(1.0+(Real)MdecisionNumAveraged(net)) ;
      }
      /***************************************************/

      if (!training && MenergyCriterion(net) != 0.0) Menergy(net) = netEnergy(net) ;
      sliceDone = ((timeThisSlice >= maxTimeThisSlice)	||
		   (timeThisSlice >= minTimeThisSlice &&
		    MuseStoppingCriteria(net) &&
		    (correct || worstOutputTrough || 
		     (!training && 
		      /* don't stop if updating baseline, just record response and RT (above) */
		      ((MdecisionResponse(net) != DECISION_NORESP && !MdecisionUpdateBaseline(net)) ||
		       (MenergyCriterion(net) != 0.0 && (Menergy(net) <= MenergyCriterion(net))) ||
		       ((MinitialPhonemeCriterion(net) && MonsetTime(net)>0.0 && MvowelTime(net)>0.0 &&
			 McodaTime(net)>0.0 && MallTime(net)>0.0) ||
			(!MinitialPhonemeCriterion(net) &&
			 MoutputChange(net) <= MoutputChangeTolerance(net)))))))) ;
      if (sliceDone) numSlicesCorrect += correct ;

      if (Mdebug(net)) {
	fprintf(dout, "%2d %4.2f %1d%s%3.1f %5.3f %5.1f %5.2f %5.2f %c ",
		forwTick, forwTime, Mslice(net), 
		(thisCorrect?"*":(worstOutputTrough?"^":" ")), 
		timeThisSlice, MoutputChange(net), net->error,
		MinstantaneousError(net), MworstOutput(net),
		((MfeatErrorInRelaxation(net) && worstOutputTrough) ?
		 MfeatPhoneme(net) : ' ')) ;

	netForAllUnits(net, OUTPUT, (UnitDataProc)printOutput, NULL) ;
	fprintf(dout, "\n") ;
      }
      netForAllOtherGroups(net, INPUT|OUTPUT, (GroupDataProc)groupIncrementUnitCost, (void *)&net->cost) ;
      /*
       * Run backward pass
       */
      if (training && 
	  (timeSinceBP >= bpWindow || 
	   sliceDone && Mslice(net) == MnumSlices(net)-1 ||
	   MrequireSliceCorrect(net) && !correct)) {
	netForAllUnits(net, ALL, (UnitDataProc)zeroDerivs, NULL) ;
	netForAllOtherGroups(net, INPUT|OUTPUT, (GroupDataProc)groupInjectUnitCost, NULL) ;
	for (backTime = 0,            backTick = forwTick ;
	     backTime < timeWindow && backTick > 0 ;
	     backTime += dt,          backTick--) 
	  netBackward(net, backTick, backTime < bpWindow) ;
	timeSinceBP = 0.0 ;
	ticksSinceBP = 0 ;
 	netForAllOtherUnits (net, BIAS, (UnitDataProc)restoreOutputAt, (void *)(forwTick%MmaxTicks(net))) ;
      }
    } /* ticks */

    MaveTime(net) += timeThisSlice-dt ;	/* extra increment when sliceDone */
    if (MdecisionResponseCriterion(net) > 0.0) {
      if (MdecisionResponse(net) == DECISION_NORESP) {
	MdecisionRT(net) = maxTimeThisSlice ;
	/* forced choice at maxTime */
	MdecisionResponse(net) = (MdecisionYesOutput(net) >= MdecisionYesOutput(net) ? DECISION_YES : DECISION_NO) ;
      } else if (!MuseStoppingCriteria(net))
	MdecisionRT(net) = timeThisSlice-dt ;	/* extra increment when sliceDone */
      if (MdecisionUpdateBaseline(net)) MdecisionNumAveraged(net)++ ;
    }
    netForAllOtherUnits(net, BIAS, (UnitDataProc)saveSliceOutput, (void *)Mslice(net)) ;
    IDoTrace(&forwardEachTime) ;
    if (MrequireSliceCorrect(net) && !correct) break ;
  } /* slice */
  MaveSlicesCorrect(net) += numSlicesCorrect ;
  if (numSlicesCorrect == MnumSlices(net))
    MaveExamplesCorrect(net)++ ;
}
/***********************************************************************
 * 			Calculate Error Derivs
 ***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  _presentedAssociate = FALSE ;
  if (MbatchSize(net) < exampleSet->numExamples) {
    int last      = MexampleNum(net) ;
    int numAssocs = exampleSet->example[last].extension->numAssocs ;
    if (MassocProb(net) > 0.0 && numAssocs > 0 && WITH_PROBABILITY(MassocProb(net))) {
      /* select associated example */
      if (numAssocs == 1)
	MexampleNum(net) = exampleSet->example[last].extension->assoc[0] ;
      else
	do {
	  MexampleNum(net) = exampleSet->example[last].extension->assoc[randomInteger(numAssocs)] ;
	} while (MexampleNum(net) == NO_ASSOC || !WITH_PROBABILITY(exampleSet->example[MexampleNum(net)].frequency)) ;
      _presentedAssociate = TRUE ;
      MinitializeUnits(net) = FALSE ;
    } else
      /* select random example according to its frequency */
      do {
	MexampleNum(net) = randomInteger(exampleSet->numExamples);
	MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
      } while (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net))) ;
      MinitializeUnits(net) = WITH_PROBABILITY(MinitializeUnitsProb(net)) ;
  } else {
    MexampleNum(net) = next ;
    MinitializeUnits(net) = TRUE ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      return FALSE ;       /* skip this example */
  }
  exampleSet->example[MexampleNum(net)].extension->presentations++ ;
  MgetNumber(exampleSet, MexampleNum(net)) ;
  MexampleInStrength(net) = exampleSet->example[MexampleNum(net)].inStrength ;
  return TRUE ;
}

/**********************************************************************/
static	void	netAdjustWeightDerivs(net)
  Net	net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	dt = Mcdt(net) ;
  Real 	*dv;

  if (MmaxWeightDeriv(net) > 0.0) {
    Real absDv, maxDv = MmaxWeightDeriv(net) ;
    for (idx = 0, dv=Mdvs(net) ; idx < numCompiledLinks ; idx++, dv++) {
      absDv = fabs(*dv *= dt) ;
      if (absDv > maxDv) *dv *= maxDv/absDv ;	/* sets to +/- maxDv */
    }
  } else {
    for (idx = 0, dv=Mdvs(net) ; idx < numCompiledLinks ; idx++, dv++)
      *dv *= dt ;
  }
}

/**********************************************************************/
static void	lengthScaleDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Link	*linkPtr  = unit->incomingLink ;
  Link	*finalPtr = linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link	link   = *(linkPtr++) ;
    Real	length = getLinkLength(link) ;
    if (length > 0.0) link->deriv /= length ;
  }
}
/**********************************************************************/
static void	netAdaptWeightCost(net)
  Net		net ;
{
  MaveError(net) = (MaveError(net)<0.0 ? net->error :
		    net->error + MaveErrorFactor(net)*(MaveError(net)-net->error)) ;
  if (net->error < MlastError(net)	||
      net->error < MdesiredError(net)	||
      MlastError(net) < 0.0) 
    MweightCost(net) += MweightCostIncrement(net) ;
  else if (net->error < MaveError(net))
    MweightCost(net) -= MweightCostIncrement(net) ;
  else
    MweightCost(net) *= MweightCostDecrement(net) ;
  if (MweightCost(net) < 0.0) MweightCost(net) = 0.0 ;
  MlastError(net) = net->error ;
}

/**********************************************************************/
static void	unitDecayWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	wtCost ;
  Boolean lengthScaleCost = (! MlengthScaleDerivs(net)) ;

  if (MnoWeightCost(unit->group)) return ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    if (MnoWeightCost(link->preUnit->group)) continue ;
    wtCost  = link->weight * weightCost ;
    if (lengthScaleCost) wtCost *= getLinkLength(link) ;
    net->cost   += wtCost * link->weight ;
    link->deriv += 2.0 * wtCost ;
  }
}
/**********************************************************************/
static void	unitMeltWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	Wo	   	= MstandardWeight(net) ;
  Real	wt, ratio ;
  Boolean lengthScaleCost = (! MlengthScaleDerivs(net)) ;

  if (MnoWeightCost(unit->group)) return ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    if (MnoWeightCost(link->preUnit->group)) continue ;
    wt	 = link->weight ;
    ratio   = (wt * wt) / (Wo * Wo) ;
    if (lengthScaleCost) weightCost *= getLinkLength(link) ;
    net->cost   += weightCost * ratio / (1.0 + ratio) ;
    link->deriv += weightCost * 2.0 * wt / ((1.0 + ratio) * (1.0 + ratio)) ;
  }
}
/**********************************************************************/
static 	void	netUpdateWeightCost(net)
  Net	net ;
{
  if (MweightCost(net) == 0.0) return ;
  if (MstandardWeight(net) == 0.0) 
    netForAllUnits(net, ALL, (UnitDataProc)unitDecayWts, NULL) ;
  else
    netForAllUnits(net, ALL, (UnitDataProc)unitMeltWts, NULL) ;
}
/**********************************************************************/
static void	zeroLinkDeriv(link)
  Link	link ;
{
  link->deriv = 0.0 ;
}
/**********************************************************************/
static void	unitFreezeWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, FROZEN, (LinkDataProc)zeroLinkDeriv, NULL) ;
}
/**********************************************************************/
static 	void	netFreezeWeights(net)
  Net	net ;
{
  netForAllUnits(net, ALL, (UnitDataProc)unitFreezeWts, NULL) ;
}
/***********************************************************************/
static	void	clearStats(net)
  Net	net ;
{
  if (net->mz->repetitionCount==0 || net->mz->nIterations==0 ||
      (net->mz->nIterations-1)%net->mz->repetitionCount==0) {
    net->error 		   	= 0.0 ;
    net->cost 		   	= 0.0 ;
    MaveTime(net)	   	= 0.0 ;
    if (MinitialPhonemeCriterion(net)) MonsetTime(net) = MvowelTime(net) = McodaTime(net) = MallTime(net) = 0.0 ;
    MaveSlicesCorrect(net)	= 0.0 ;
    MaveExamplesCorrect(net)    = 0.0 ;
  }
}
/***********************************************************************/
static	void	normStats(net)
  Net	net ;
{
  Real	scale = 1.0 ;

  if (net->mz->repetitionCount==0 || 
      net->mz->nIterations%net->mz->repetitionCount==0) {
    if (net->mz->nIterations && net->mz->repetitionCount)
      scale = (Real)net->mz->repetitionCount ;
    net->error		     	*= Mdt(net) / scale ;
    net->cost		     	/= scale ;
    scale *= net->batchSize ;
    MaveTime(net)	     	/= scale ;
    if (MinitialPhonemeCriterion(net)) {
      MonsetTime(net) /= scale ;
      MvowelTime(net) /= scale ;
      McodaTime(net)  /= scale ;
      MallTime(net)   /= scale ;
    }
    MaveSlicesCorrect(net) 	/= scale ;
    MaveExamplesCorrect(net)    /= scale ;
  }
}
/**********************************************************************/
static Proc	updateNetActivity (net)
  Net	net ;
{
  Mrunning(net) = FALSE ;
  if (!Mcompiled(net)) netCompile(net) ;
  getCompiledWeights(net) ;

  net->error               	= 0.0 ;
  MexampleFrequency(net)	= 1.0 ;
  MaveTime(net)	 	   	= 0 ;
  if (MinitialPhonemeCriterion(net)) MonsetTime(net) = MvowelTime(net) = McodaTime(net) = MallTime(net) = 0.0 ;
  MaveSlicesCorrect(net)	= 0 ;
  MaveExamplesCorrect(net)	= 0 ;
  MinitializeUnits(net) 	= (MinitializeUnitsProb(net) == 1.0) ;

  netIterate(net, FALSE) ;

  net->error		       *= Mdt(net) ;
  Mslice(net)			= MnumSlices(net) - 1 ;
  IDoTrace(&afterExample) ;
}

/**********************************************************************/
static Real	_epsilon ;
static Real	_alpha ;
/**********************************************************************/
void		linkZeroDeltaWeight(link, data)
  Link		link ;
  void		*data ;
{
  link->deltaWeight = 0.0 ;
}
/**********************************************************************/
static void	unitZeroDeltaWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkZeroDeltaWeight, NULL) ;
}
/**********************************************************************/
static void	unitChangeWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;

  if (_alpha > 0.0)
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link link		= incoming[idx] ;
      link->deltaWeight = _alpha * link->deltaWeight - _epsilon * link->deriv ;
      link->weight     +=  link->deltaWeight ;
    }
  else
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link link		= incoming[idx] ;
      link->deltaWeight = -_epsilon * link->deriv ;
      link->weight     += link->deltaWeight ;
    }
}

static Real	_g2 ;
static Real	_w2 ;
static Real	_gw ;
/**********************************************************************/
void		linkCosAngle(link, data)
  Link		link ;
  void		*data ;
{
  _g2 += link->deriv       * link->deriv ;
  _w2 += link->deltaWeight * link->deltaWeight ;
  _gw += link->deriv       * link->deltaWeight ;
}
/**********************************************************************/
static void	unitCosAngle(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkCosAngle, NULL) ;
}
/***********************************************************************/
static void	calculateError(net, exampleSet, calculateDerivs)
  Net		net ;
  ExampleSet	exampleSet ;
  Boolean	calculateDerivs ;
{
  int 	example ;

  Mrunning(net)  = TRUE ;
  if (!Mcompiled(net)) netCompile(net) ;
  clearStats(net) ;
  getCompiledWeights(net) ;

  if (calculateDerivs && net->mz->nIterations == 0) 
    for (example = 0 ; example < exampleSet->numExamples ; ++example) 
      exampleSet->example[example].extension->presentations = 0 ;

  for (example = 0 ; example < MbatchSize(net) ; ++example) {
    if (!getExample(net, exampleSet, example)) continue ;
    netIterate(net, calculateDerivs) ;
    if (calculateDerivs) IDoTrace(&afterExample) ;
  }
  if (example <= 0)
    IErrorAbort("calculateError: no examples processed") ;
  if (calculateDerivs) {
    netAdjustWeightDerivs(net) ;
    putCompiledDerivs(net) ;
    if (MlengthScaleDerivs(net))
      netForAllUnits(net, ALL, (UnitDataProc)lengthScaleDerivs, net) ;
    if (MweightCostIncrement(net) > 0.0) 
      netAdaptWeightCost(net) ;
    netUpdateWeightCost(net) ;
    netFreezeWeights(net) ;
  }
  normStats(net) ;
  Mrunning(net)  = FALSE ;
}
/***********************************************************************
 *			netMinimize
 ***********************************************************************/
static void	netMinimize(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int	example ;
  Real	cosAngle ;

  fprintf(dout, "WARNING: not returning to minimize\n") ;

  if (net->currentEpoch == 0) {
    fprintf(dout, "Initializing momentum direction.\n") ;
    netForAllUnits(net, ALL, (UnitDataProc)unitZeroDeltaWeights, NULL) ;
  }

  for ( ; net->mz->maxIterations==0 || net->mz->nIterations <= net->mz->maxIterations ; 
       net->mz->nIterations++, net->currentEpoch++) {

    initEpoch(net, MalphaOnset(net)) ;
    _epsilon = net->mz->epsilon ;
    _alpha   = net->mz->alpha ;

    calculateError(net, exampleSet, TRUE) ;

    /* going to be time to print - want to do this BEFORE weight changes */
    if (net->mz->repetitionCount==0 || net->currentEpoch%net->mz->repetitionCount == 0) {
      _g2 = _w2 = _gw = 0.0 ;
      netForAllUnits(net, ALL, (UnitDataProc)unitCosAngle, NULL) ;
      cosAngle = (_g2*_w2 <= 0.0 ? 0.0 :_gw/sqrt(_g2*_w2)) ;
    }
    /* all but last iteration */
    if (net->mz->maxIterations==0 || net->mz->nIterations < net->mz->maxIterations) {
      IDoTrace(&minimizeEndIteration) ;
      /* NOTE: only does gradient/momentum descent */
      netForAllUnits(net, ALL, (UnitDataProc)unitChangeWeights, NULL) ;
      if (net->mz->repetitionCount==0 || net->currentEpoch%net->mz->repetitionCount == 0) 
	IDoTrace(&minimizeEndRepetition) ;
    }
    /* time to print */
    if (net->mz->repetitionCount==0 || net->currentEpoch%net->mz->repetitionCount == 0)
      fprintf(dout, "epoch %d err= %f cos= %f\n", net->currentEpoch, net->error, cosAngle) ;
  } /* epoch */
  Mrunning(net) = FALSE ;
}

/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  if (MuseMinimize(net)) {
    initEpoch(net, MalphaOnset(net)) ;
    calculateError(net, exampleSet, TRUE) ;
  } else 
    netMinimize(net, exampleSet) ;
}
/*********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
/*
  calculateError(net, exampleSet, FALSE) ;
*/
}
/*********************************************************************
 * 				Link Length
 *********************************************************************/
void		setLinkLength(link, length)
  Link		link ;
  Real		length ;
{
  Mlength(link) = length ;
}

Real		getLinkLength(link)
  Link		link ;
{
  return Mlength(link) ;
}

int		unitXPosition(unit)
  Unit	unit ;
{
  return (MgroupXPosition(unit->group) >= 0 ? 
	  MgroupXPosition(unit->group) : MxPosition(unit)) ;
}

int		unitYPosition(unit)
  Unit	unit ;
{
  return (MgroupYPosition(unit->group) >= 0 ? 
	  MgroupYPosition(unit->group) : MyPosition(unit)) ;
}
/**********************************************************************/
/**********************************************************************/
static void	linkReadLensWeight(link, data)
  Link		link ;
  void		*data ;
{
  Real	wt ;
  if (fscanf((FILE *)data, "%f", &wt))
    link->weight = wt ;
  else 
    IErrorAbort("Unable to read weight for link %s: got %f instead\n", link->name, wt) ;
}

static void	unitReadLensWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkReadLensWeight, data) ;
}  

int	command_loadLensWeights(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  int	numWeights ;
  FILE	*fp ;

  IUsage("filename");
  if (GiveHelp(tokc)) {
    ISynopsis("Loads weights written by Lens");
    IHelp
      (IHelpArgs,
       "Loads weights written by Lens.",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  fp = IOpenFileOrAbort(tokv[1], "r", NULL) ;
  fscanf(fp, "%d\n\n", &numWeights) ;
  fprintf(dout, "Reading %d weights....\n", numWeights) ;
  netForAllUnits(currentNet, ALL, (UnitDataProc)unitReadLensWeights, (void *)fp) ;
  ICloseFile(fp, NULL) ;
  return 1 ;
}
/**********************************************************************/
