
/**********************************************************************
 * $Id: bp.h,v 1.7 91/09/26 11:05:16 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#ifndef BARAK_H
#define BARAK_H

#ifndef linux
#include <nan.h>
#endif

#ifndef isNaN
#define HP_NAN		(MAXFLOAT-1.0)
#define isNaN(x)	((x)==HP_NAN)
#define setNaN(x)	((x)=HP_NAN)
#endif

/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  Boolean	running ;
  int	alphaOnset ;		/* netParam: */
  int	delayCount ;		/* netParam: */

  Boolean timeAverageInputs ;	/* netParam: */

  Real	nonExtremalOneTarget ;	/* netParam: */
  Real	nonExtremalZeroTarget ;	/* netParam: */
  Real	zeroErrorRadius ;	/* netParam: */
  Real	correctTolerance ;	/* netParam: */
  Real	maxWeightDeriv ;	/* netParam: */

  int	errorFunction ;		/* netParam: 0=CE 1=SSE */
  Real	initializeUnitsProb ;	/* netParam: */
  Boolean initializeUnits ;
  Real	initialUnitOutput ;	/* netParam: */
  Real	unitNoise ;		/* netParam: */
  Real	inputNoise ;		/* netParam: */
  Real	inputGain ;		/* netParam: */
  Boolean inputNoiseEachTick ;	/* netParam: */
  Real	inputStrength ;		/* netParam: */
  Real	teacherForcing ;	/* netParam: */
  int	inputClamping ;		/* netParam: */
  int	exampleNum ;
  Real	exampleFrequency ;
  int	sampleExamples ;	/* netParam: */
  Real	assocProb ;		/* netParam: */
  Real	exampleInStrength ;	/* netParam: */

  Boolean lengthScaleDerivs ;	/* netParam: */

  /* adaptive weight decay/elimination */
  Real	standardWeight ;	/* netParam: */
  Real	weightCostIncrement ;	/* netParam: */
  Real	weightCostDecrement ;	/* netParam: */
  Real	lastError ;
  Real	aveError ;
  Real	aveErrorFactor ;	/* netParam: */
  Real	desiredError ;		/* netParam: */

  Real	instantaneousError ;
  Real	worstOutput ;
  Real	troughTolerance ;	/* netParam: */
  int	minDecrements ;		/* netParam: */
  char	featPhoneme ;
  Real	featError ;
  Boolean featErrorInRelaxation ;/* netParam: */

  Real	energy ;	
  Real	energyCriterion ;		/* netParam: */
  Real	outputChange ;	
  Real	outputChangeTolerance ;		/* netParam: */
  Boolean outputChangeByGroup ;		/* netParam: */
  Boolean outputChangeUseAll ;		/* netParam: */
  Boolean outputChangeUseAve ;		/* netParam: */
  Boolean outputStressInRelaxation ;	/* netParam: */
  Real	outputStress ;

  Boolean outputActivityInRelaxation ;	/* netParam: */
  Real	outputActivityThresh ;		/* netParam: */
  Real	outputActivity ;

  Boolean initialPhonemeCriterion ;	/* netParam: */
  Real onsetTime ;
  Real vowelTime ;
  Real codaTime ;
  Real allTime ;
  Real onsetChange ;
  Real vowelChange ;
  Real codaChange ;

  /* Usher-McClelland leaky integrator decision process */
  Real	decisionDtFactor ;		/* netParam: */
  Real	decisionSelfExcitation ;	/* netParam: */
  Real	decisionMutualInhibition ;	/* netParam: */
  Real	decisionNoise ;			/* netParam: */
  Real	decisionMinInput ;		/* netParam: */
  Real	decisionMaxInput ;		/* netParam: */
  Real	decisionYesOutput ;
  Real	decisionNoOutput ;
  Real	decisionResponseCriterion ;	/* netParam: */
  Real	decisionRT ;			/* netParam: */
  int	decisionResponse ;
  Real	decisionInitialBaseline ;	/* netParam: */
  Real	decisionBaselineIncrement ;	/* netParam: */
  Boolean decisionUpdateBaseline ;	/* netParam: */
  int	decisionNumAveraged ;
  int	decisionMaxTicks ;
  Real	*decisionBaseline ;		/* counter: decisionMaxTicks */

  Boolean useStoppingCriteria ;		/* netParam: */
  Boolean runPrimingInRelaxationTrace ;	/* netParam: */

  Real	aveTime ;	
  Real	aveSlicesCorrect ;	
  Real	aveExamplesCorrect ;	
  int	outputsIncorrect ;

  Boolean	compiled ;
  int	numCompiledLinks ;
  Real* *preOutputs ;		/* invis: counter: numCompiledLinks */
  Real* *preOutputDerivs ;	/* invis: counter: numCompiledLinks */
  Real*	*postInputs ;		/* invis: counter: numCompiledLinks */
  Real*	*postInputDerivs ;	/* invis: counter: numCompiledLinks */
  Real*	*weights ;		/* invis: counter: numCompiledLinks */
  Real	*wts ;			/* invis: counter: numCompiledLinks */
  Real*	*derivs ;		/* invis: counter: numCompiledLinks */
  Real	*dvs ;			/* invis: counter: numCompiledLinks */

  int	maxTicks ;		
  int	numSlices ;
  Real	graceTimePerSlice ;	/* netParam: */
  Real	minTimeCorrect ;	/* netParam: */
  Real	minTimePerSlice ;	/* netParam: */
  Real	maxTimePerSlice ;	/* netParam: */
  Real	graceTimeFirstSlice ;	/* netParam: */
  Real	minTimeFirstSlice ;	/* netParam: */
  Real	maxTimeFirstSlice ;	/* netParam: */
  Real	timeFirstSliceMin ;	/* netParam: */
  Real	timeFirstSliceMax ;	/* netParam: */
  Real	timeFirstSliceTau ;	/* netParam: */
  Boolean assocMinProb ;	/* netParam: */

  Real	graceTimeLastSlice ;	/* netParam: */
  Real	minTimeLastSlice ;	/* netParam: */
  Real	maxTimeLastSlice ;	/* netParam: */
  Real	timeLastSliceMin ;	/* netParam: */
  Real	timeLastSliceMax ;	/* netParam: */
  Real	timeLastSliceTau ;	/* netParam: */

  Real	timeWindow  ;		/* netParam: */
  Real	bpWindow  ;		/* netParam: */
  Real  dt ;		        /* netParam: */
  Real  dtCorrection ;	        /* netParam: */
  Real  cdt ;			/* corrected dt */
  Boolean	requireSliceCorrect ;	/* netParam: */
  Boolean	useMinimize ;	/* netParam: */
} NetExtensionRec ;

typedef struct GroupExtensionRec {	/* BIND */
  int	inputClamping ;
  int	noWeightCost ;
  Real	minBias ;
  Real	maxBias ;
  Real	minIncomingWeight ;
  Real	maxIncomingWeight ;
  Real	minOutgoingWeight ;
  Real	maxOutgoingWeight ;
  int	groupXPosition ;
  int	groupYPosition ;
  Real	unitCost ;
  Real	unitPrior ;
  Real	meanOutput ;
  Boolean	useOutputChange ;
} GroupExtensionRec ;

#define LESIONED 1
#define CLAMPED  2

typedef struct UnitExtensionRec {	/* BIND */
  Mask		status ;
  Boolean	equateIncomingWeights ;
  Boolean	equateOutgoingWeights ;
  Real		old_x ;		/* old totalInput (when time averaging inputs) */
  Real		ext_x ;		/* external input (for input units) */
  Real*		y_t ;		/* output */
  Real*		d_t ;		/* target */
  Real*		sigx_t ;	/* sigmoid(x) for each tick */
  Real		cdEdy ;		/* current dEdy for time averaging in backward pass */
  Real		errorFactor ;	/* scales derivatives */
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Mask		status ;
  Real		lesionedWeight ;
  Real		length ;
} LinkExtensionRec ;

#define NO_ASSOC	-1
#define MAX_ASSOCS	32

typedef struct ExampleExtensionRec {	/* BIND */
  int	numAssocs ;
  int	assoc[32] ;
  int	presentations ;
} ExampleExtensionRec ;

typedef struct ExampleSetExtensionRec {	/* BIND */
  void	*data ;
} ExampleSetExtensionRec ;

#define MbatchSize(net)		((net)->batchSize)
#define	Mslice(net)		((net)->currentTime)

#define Mdebug(net)		((net)->extension->debug)
#define Mrunning(net)		((net)->extension->running)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MtimeAverageInputs(net)	((net)->extension->timeAverageInputs)
#define MnonExtremalOneTarget(net)  ((net)->extension->nonExtremalOneTarget)
#define MnonExtremalZeroTarget(net) ((net)->extension->nonExtremalZeroTarget)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define McorrectTolerance(net)	((net)->extension->correctTolerance)
#define MmaxWeightDeriv(net)	((net)->extension->maxWeightDeriv)

#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MinitializeUnitsProb(net) ((net)->extension->initializeUnitsProb)
#define MinitializeUnits(net)	((net)->extension->initializeUnits)
#define MinitialUnitOutput(net)	((net)->extension->initialUnitOutput)
#define MunitNoise(net)		((net)->extension->unitNoise)
#define MinputGain(net)		((net)->extension->inputGain)
#define MinputNoise(net)	((net)->extension->inputNoise)
#define MinputNoiseEachTick(net) ((net)->extension->inputNoiseEachTick)
#define MinputStrength(net)	((net)->extension->inputStrength)
#define MteacherForcing(net)	((net)->extension->teacherForcing)
#define MinputClamping(net)	((net)->extension->inputClamping)

#define MerrorFunction(net)	((net)->extension->errorFunction)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MexampleInStrength(net)	((net)->extension->exampleInStrength)
#define MsampleExamples(net) 	((net)->extension->sampleExamples)
#define MassocProb(net) 	((net)->extension->assocProb)

#define MlengthScaleDerivs(net)	((net)->extension->lengthScaleDerivs)

#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MweightCostIncrement(net) ((net)->extension->weightCostIncrement)
#define MweightCostDecrement(net) ((net)->extension->weightCostDecrement)
#define MlastError(net)		((net)->extension->lastError)
#define MaveError(net)		((net)->extension->aveError)
#define MaveErrorFactor(net)	((net)->extension->aveErrorFactor)
#define MdesiredError(net)	((net)->extension->desiredError)
#define MinstantaneousError(net) ((net)->extension->instantaneousError)
#define MworstOutput(net)	((net)->extension->worstOutput)
#define MtroughTolerance(net)	((net)->extension->troughTolerance)
#define MminDecrements(net)	((net)->extension->minDecrements)
#define MfeatPhoneme(net)	((net)->extension->featPhoneme)
#define MfeatError(net)		((net)->extension->featError)
#define MfeatLastError(net)	((net)->extension->featLastError)
#define MfeatErrorInRelaxation(net) ((net)->extension->featErrorInRelaxation)

#define Mcompiled(net)		((net)->extension->compiled)
#define MpostInputs(net)	((net)->extension->postInputs)
#define MpostInputDerivs(net)	((net)->extension->postInputDerivs)
#define MpreOutputs(net)	((net)->extension->preOutputs)
#define MpreOutputDerivs(net)	((net)->extension->preOutputDerivs)
#define Mwts(net)		((net)->extension->wts)
#define Mweights(net)		((net)->extension->weights)
#define Mderivs(net)		((net)->extension->derivs)
#define Mdvs(net)		((net)->extension->dvs)
#define MnumCompiledLinks(net)	((net)->extension->numCompiledLinks)

#define Menergy(net)		((net)->extension->energy)
#define MenergyCriterion(net)	((net)->extension->energyCriterion)
#define MoutputChange(net)	    ((net)->extension->outputChange)
#define MoutputChangeTolerance(net) ((net)->extension->outputChangeTolerance)
#define MoutputChangeByGroup(net)    ((net)->extension->outputChangeByGroup)
#define MoutputChangeUseAve(net)    ((net)->extension->outputChangeUseAve)
#define MoutputChangeUseAll(net)    ((net)->extension->outputChangeUseAll)

#define MoutputStressInRelaxation(net)		((net)->extension->outputStressInRelaxation)
#define MoutputStress(net)			((net)->extension->outputStress)

#define MoutputActivityInRelaxation(net)	((net)->extension->outputActivityInRelaxation)
#define MoutputActivityThresh(net)		((net)->extension->outputActivityThresh)
#define MoutputActivity(net)			((net)->extension->outputActivity)

#define NSYL_LAST_ONSET 22
#define NSYL_LAST_VOWEL 36
#define MinitialPhonemeCriterion(net)	((net)->extension->initialPhonemeCriterion)
#define MonsetTime(net)			((net)->extension->onsetTime)
#define MvowelTime(net)			((net)->extension->vowelTime)
#define McodaTime(net)			((net)->extension->codaTime)
#define MallTime(net)			((net)->extension->allTime)
#define MonsetChange(net)		((net)->extension->onsetChange)
#define MvowelChange(net)		((net)->extension->vowelChange)
#define McodaChange(net)		((net)->extension->codaChange)

#define MdecisionDtFactor(net)		((net)->extension->decisionDtFactor)
#define MdecisionSelfExcitation(net)	((net)->extension->decisionSelfExcitation)
#define MdecisionMutualInhibition(net)	((net)->extension->decisionMutualInhibition)
#define MdecisionNoise(net)		((net)->extension->decisionNoise)
#define MdecisionMinInput(net)		((net)->extension->decisionMinInput)
#define MdecisionMaxInput(net)		((net)->extension->decisionMaxInput)
#define MdecisionYesOutput(net)		((net)->extension->decisionYesOutput)
#define MdecisionNoOutput(net)		((net)->extension->decisionNoOutput)
#define MdecisionResponseCriterion(net)	((net)->extension->decisionResponseCriterion)
#define MdecisionRT(net)		((net)->extension->decisionRT)
#define MdecisionResponse(net)		((net)->extension->decisionResponse)
#define MdecisionInitialBaseline(net)	((net)->extension->decisionInitialBaseline)
#define MdecisionBaselineIncrement(net)	((net)->extension->decisionBaselineIncrement)
#define MdecisionUpdateBaseline(net)	((net)->extension->decisionUpdateBaseline)
#define MdecisionNumAveraged(net)	((net)->extension->decisionNumAveraged)
#define MdecisionMaxTicks(net)		((net)->extension->decisionMaxTicks)
#define MdecisionBaseline(net)		((net)->extension->decisionBaseline)

#define MuseStoppingCriteria(net)   ((net)->extension->useStoppingCriteria)
#define MrunPrimingInRelaxationTrace(net) ((net)->extension->runPrimingInRelaxationTrace)

#define Mrefixate(net)		((net)->extension->refixate)
#define MaveFixations(net)	((net)->extension->aveFixations)
#define MaveSlicesCorrect(net)	((net)->extension->aveSlicesCorrect)
#define MaveExamplesCorrect(net) ((net)->extension->aveExamplesCorrect)
#define MoutputsIncorrect(net)	((net)->extension->outputsIncorrect)
#define MaveTime(net)		((net)->extension->aveTime)

#define MmaxTicks(net)		((net)->extension->maxTicks)
#define MnumSlices(net)		((net)->extension->numSlices)
#define MgraceTimePerSlice(net) ((net)->extension->graceTimePerSlice)
#define MminTimeCorrect(net)	((net)->extension->minTimeCorrect)
#define MminTimePerSlice(net)	((net)->extension->minTimePerSlice)
#define MmaxTimePerSlice(net)	((net)->extension->maxTimePerSlice)
#define MassocMinProb(net)	((net)->extension->assocMinProb)

#define MgraceTimeFirstSlice(net) ((net)->extension->graceTimeFirstSlice)
#define MminTimeFirstSlice(net)	((net)->extension->minTimeFirstSlice)
#define MmaxTimeFirstSlice(net)	((net)->extension->maxTimeFirstSlice)
#define MtimeFirstSliceMin(net) ((net)->extension->timeFirstSliceMin)
#define MtimeFirstSliceMax(net) ((net)->extension->timeFirstSliceMax)
#define MtimeFirstSliceTau(net) ((net)->extension->timeFirstSliceTau)

#define MgraceTimeLastSlice(net) ((net)->extension->graceTimeLastSlice)
#define MminTimeLastSlice(net)	((net)->extension->minTimeLastSlice)
#define MmaxTimeLastSlice(net)	((net)->extension->maxTimeLastSlice)
#define MtimeLastSliceMin(net)	((net)->extension->timeLastSliceMin)
#define MtimeLastSliceMax(net)	((net)->extension->timeLastSliceMax)
#define MtimeLastSliceTau(net)	((net)->extension->timeLastSliceTau)

#define MtimeWindow(net)	((net)->extension->timeWindow)
#define MbpWindow(net)		((net)->extension->bpWindow)
#define Mdt(net)		((net)->extension->dt)
#define MdtCorrection(net)	((net)->extension->dtCorrection)
#define Mcdt(net)		((net)->extension->cdt)
#define MrequireSliceCorrect(net) ((net)->extension->requireSliceCorrect)

#define MuseMinimize(net)	((net)->extension->useMinimize)

#define MnoWeightCost(group)		((group)->extension->noWeightCost)
#define MminBias(group)			((group)->extension->minBias)
#define MmaxBias(group)			((group)->extension->maxBias)
#define MminIncomingWeight(group)	((group)->extension->minIncomingWeight)
#define MmaxIncomingWeight(group)	((group)->extension->maxIncomingWeight)
#define MminOutgoingWeight(group)	((group)->extension->minOutgoingWeight)
#define MmaxOutgoingWeight(group)	((group)->extension->maxOutgoingWeight)
#define MgroupXPosition(group)		((group)->extension->groupXPosition)
#define MgroupYPosition(group)		((group)->extension->groupYPosition)
#define MunitCost(group)		((group)->extension->unitCost)
#define MunitPrior(group)		((group)->extension->unitPrior)
#define MmeanOutput(group)		((group)->extension->meanOutput)
#define MuseOutputChange(group)		((group)->extension->useOutputChange)

#define Mstatus(unit_or_link)	((unit_or_link)->extension->status)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)
#define Mlength(link)		((link)->extension->length)

#define MequateIncomingWeights(unit)	((unit)->extension->equateIncomingWeights)
#define MequateOutgoingWeights(unit)	((unit)->extension->equateOutgoingWeights)

#define Mi_slice(unit)		((unit)->clampingHistory)
#define Mx(unit)		((unit)->totalInput)
#define Moldx(unit)		((unit)->extension->old_x)
#define Mext(unit)		((unit)->extension->ext_x)
#define My(unit)		((unit)->output)
#define My_slice(unit)		((unit)->outputHistory)
#define My_t(unit)		((unit)->extension->y_t)
#define Msigx_t(unit)		((unit)->extension->sigx_t)
#define Md(unit)		((unit)->target)
#define Md_slice(unit)		((unit)->targetHistory)
#define Md_t(unit)		((unit)->extension->d_t)
#define MdEdx(unit)		((unit)->inputDeriv)
#define MdEdy(unit)		((unit)->outputDeriv)
#define McdEdy(unit)		((unit)->extension->cdEdy)
#define MerrorFactor(unit)	((unit)->extension->errorFactor)

#define NULL_POS	-1

#define DECISION_YES	1
#define DECISION_NO	0
#define DECISION_NORESP	-1

#define DECISION_UPDATE(net,input,output,other)	\
   (Mdt(net) * MdecisionDtFactor(net) *			\
    (input						\
     + output*(MdecisionSelfExcitation(net)-1.0)	\
     - other*MdecisionMutualInhibition(net)		\
     + gaussianNoise(MdecisionNoise(net))))


/*****************************************************************************
 * PERMUTED ITERATION
 * 
 * for (i=0; i<max; i++) {
 *   local_variable_defs ;
 *   PERMUTE(i,max) ;
 *   commands ;
 *   PERMUTE_RESTORE(i,max) ;
 * }
 *****************************************************************************/

/* must be immediately following local variable definitions inside for-loop */
#define PERMUTE(var,max)							\
  int permVar=var, permIdx, *permIdxArr ;					\
  if (var==0) {									\
    permIdxArr = (int *)calloc(max, sizeof(int)) ;				\
    for (permIdx=0; permIdx<max; permIdx++) permIdxArr[permIdx]=permIdx ; }	\
  permIdx = randomInteger(max-var) ;						\
  var     = permIdxArr[permIdx] ;						\
  for ( ; permIdx < max ; permIdx++) permIdxArr[permIdx] = permIdxArr[permIdx+1] ;

/* must be at end of PERMUTE loop */
#define PERMUTE_RESTORE(var,max)			\
  var = permVar ;					\
  if (var==max-1) free(permIdxArr) ;


/* must be immediately following local variable definitions inside for-loop */
#define PERMUTE_PAIR(innerVar,innerMax,outerVar,outerMax)			\
  int permInnerVar = innerVar, permOuterVar = outerVar, *permIdxArr ;		\
  int permPair, permPairMaxIdx, permIdxMax ;					\
  int permIdx = innerVar + outerVar*innerMax ;					\
  if (permIdx==0) {								\
    permIdxMax = innerMax*outerMax ;						\
    permIdxArr = (int *)calloc(permIdxMax, sizeof(int)) ;			\
    for (permIdx=0; permIdx<permIdxMax; permIdx++) permIdxArr[permIdx]=permIdx ;\
    permIdx = 0 ; }								\
  permPairMaxIdx = permIdxMax-permIdx ;						\
  permIdx  = randomInteger(permPairMaxIdx) ;					\
  permPair = permIdxArr[permIdx] ;						\
  innerVar = permPair%innerMax ;						\
  outerVar = permPair/innerMax ;						\
  for ( ; permIdx < permIdxMax-1 ; permIdx++) permIdxArr[permIdx] = permIdxArr[permIdx+1] ;


/* must be at end of PERMUTE loop */
#define PERMUTE_RESTORE_PAIR(innerVar,innerMax,outerVar,outerMax)		\
  innerVar = permInnerVar ;							\
  outerVar = permOuterVar ;							\
  if (innerVar+outerVar*innerMax==innerMax*outerMax-1) free(permIdxArr) ;	

/**********************************************************************/
#define DELTA				(0.000001)
#define SIMPLE_CROSS_ENTROPY(y, d)   (-(d)*log(DELTA+y) - (1.0-(d))*log(DELTA+1.0-(y)))
#define CROSS_ENTROPY_ZERO_TARGET(y) ((y) == 0.0 ? 0.0 : -log(DELTA+1.0 - (y)))
#define CROSS_ENTROPY_ONE_TARGET(y)  ((y) == 1.0 ? 0.0 : -log(DELTA+y))
#define CE(y, d)		\
  (((d) == 0.0) ? CROSS_ENTROPY_ZERO_TARGET(y) : 	\
   (((d) == 1.0) ? CROSS_ENTROPY_ONE_TARGET(y) : SIMPLE_CROSS_ENTROPY(y, d)))
#define CE_DERIV(y, d)		\
  ((d) == 0.0 ?   1.0/(1.0-(y)) : 		\
   ((d) == 1.0 ? -1.0/(y)      : ((y)-(d))/((y)*(1.0-(y)))))
#define DIV(y, d)		((d) == 0.0 ? 0.0 : -(d)*log(DELTA+y))
#define DIV_DERIV(y, d)		((d) == 0.0 ? 0.0 : -(d)/(y))
/**********************************************************************/


/* types of clamping */
#define	SOFT	((Mask)(1<<0))	/* use external input */
#define	ONESHOT	((Mask)(1<<1))	/* only present unit states once */

extern struct TRACE	afterExample ; 		/* BIND */
extern struct TRACE	forwardEachTime ;	/* BIND */
extern struct TRACE    	inRelaxation ;		/* BIND */
extern struct TRACE    	primingInRelaxation ;	/* BIND */


/* compile-barak.c */
extern	void	netCompile		ARGS((Net net)) ;
extern	void	netCompileTicks		ARGS((Net net)) ;
extern	void	getCompiledWeights	ARGS((Net net)) ;
extern	void	putCompiledDerivs	ARGS((Net net)) ;

/* lesion-barak.c */
extern void	lesionUnit		ARGS((Unit unit)) ;
extern void	unlesionUnit		ARGS((Unit unit)) ;
extern Boolean	isLesionedUnit		ARGS((Unit unit)) ;
extern void	lesionLink		ARGS((Link link)) ;
extern void	unlesionLink		ARGS((Link link)) ;
extern Boolean	isLesionedLink		ARGS((Link link)) ;

extern void	netIterate			ARGS((Net net, int training)) ;

extern char	featLMSEphoneme ;	/* SM.c */

#endif				/* BARAK_H */
