#include <commands.h>
#include <xerion/utilsDCP.h>
#include "barak.h"

/**********************************************************************
 * 				Compilation
 *********************************************************************/
static void	unitCompile(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numLinks	= MnumCompiledLinks(net) ;

  if (Mstatus(unit)&LESIONED) return ;

  for (idx = 0 ; idx < unit->numIncoming ; ++idx) {
    Link	link  = unit->incomingLink[idx] ;
    if (! Mstatus(link)&LESIONED) {
      MpostInputs(net)[numLinks]	= &(link->postUnit->totalInput) ;
      MpostInputDerivs(net)[numLinks]	= &(link->postUnit->inputDeriv) ;
      MpreOutputs(net)[numLinks]	= &(link->preUnit->output) ;
      MpreOutputDerivs(net)[numLinks]	= &(link->preUnit->extension->cdEdy) ;
      Mweights(net)[numLinks]		= &(link->weight) ;
      Mderivs(net)[numLinks] 		= &(link->deriv) ;
      numLinks++ ;
      if (link->type&BIDIRECTIONAL) {
	MpostInputs(net)[numLinks]	= &(link->preUnit->totalInput) ;
	MpostInputDerivs(net)[numLinks]	= &(link->preUnit->inputDeriv) ;
	MpreOutputs(net)[numLinks]	= &(link->postUnit->output) ;
	MpreOutputDerivs(net)[numLinks]	= &(link->postUnit->extension->cdEdy) ;
	Mweights(net)[numLinks] 	= &(link->weight) ;
	Mderivs(net)[numLinks]		= &(link->deriv) ;
	numLinks++ ;
      }
    }
  }
  MnumCompiledLinks(net) = numLinks ;
}
/**********************************************************************/
static void	countCompiledLinks(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx, sum = 0 ;
  if (Mstatus(unit)&LESIONED) return ;
  for (idx = 0 ; idx < unit->numIncoming ; ++idx) {
    Link	link  = unit->incomingLink[idx] ;
    if (!Mstatus(link)&LESIONED) {
      sum++ ;
      if (link->type&BIDIRECTIONAL) sum++ ;
    }
  }
  MnumCompiledLinks(unit->net) += sum ;
}
/**********************************************************************/
static void	unitCompileTicks(unit, ticks)
  Unit		unit ;
  void		*ticks  ;
{
  int	oldTicks = MmaxTicks(unit->net) ;
  extendList((void *)&My_t(unit),	sizeof(Real), oldTicks, ticks);
  extendList((void *)&Msigx_t(unit),	sizeof(Real), oldTicks, ticks);
  if (unit->group->type & OUTPUT)
    extendList((void *)&Md_t(unit),	sizeof(Real), oldTicks, ticks);
}
/***********************************************************************/
void	netCompileTicks(net)
  Net		net ;
{
  Real	time = (MtimeWindow(net)>0 ? MtimeWindow(net) :
		MnumSlices(net) * MmaxTimePerSlice(net)) ;
  int	maxTicks ;

  if (Mdt(net) == 0.0) 
    IErrorAbort("dt = 0") ;
  else
    maxTicks = 1 + (time / Mdt(net)) ;

  if (maxTicks > MmaxTicks(net)) {
    netForAllUnits(net, ALL, (UnitDataProc)unitCompileTicks, (void *)maxTicks) ;
    MmaxTicks(net) = maxTicks ;
  }
}
/***********************************************************************/
void	netCompile(net)
  Net		net ;
{
  static int	maxLinks = 0;
  int		numLinks = 0;

  MnumCompiledLinks(net) = 0;
  netForAllUnits(net, ALL, (UnitDataProc)countCompiledLinks, NULL) ;
  numLinks = MnumCompiledLinks(net);

  if (numLinks > maxLinks) {
    extendList((void *)&MpostInputs(net),      sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&MpostInputDerivs(net), sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&MpreOutputs(net),      sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&MpreOutputDerivs(net), sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&Mweights(net),	sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&Mderivs(net),	sizeof(Real*), maxLinks, numLinks);

    extendList((void *)&Mwts(net), sizeof(Real), maxLinks, numLinks);
    extendList((void *)&Mdvs(net), sizeof(Real), maxLinks, numLinks);

    maxLinks = numLinks ;
  }
  MnumCompiledLinks(net) = 0;
  netForAllUnits(net, ALL, (UnitDataProc)unitCompile, NULL) ;
  netCompileTicks(net) ;
  Mcompiled(net) = TRUE ;
}

/**********************************************************************/
static void	linkSumWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  *(Real *)data += link->weight ;
}
/**********************************************************************/
static void	linkSetWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = *(Real *)data ;
}
/**********************************************************************/
static void	linkMinWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = MAX(link->weight, *(Real *)data) ;
}
/**********************************************************************/
static void	linkMaxWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = MIN(link->weight, *(Real *)data) ;
}
/**********************************************************************/
static void	unitConstrainWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  /* CONSTRAIN BIAS */
  if (unit->numIncoming > 0) {
    if (MmaxBias(unit->group)<MAXFLOAT)
      unit->incomingLink[0]->weight = 
	MIN(unit->incomingLink[0]->weight,MmaxBias(unit->group)) ;
    if (MminBias(unit->group)>-MAXFLOAT)
      unit->incomingLink[0]->weight = 
	MAX(unit->incomingLink[0]->weight,MminBias(unit->group)) ;
  }

  /* INCOMING WEIGHTS */
  if (unit->numIncoming > 1) {
    if (MequateIncomingWeights(unit)) {
      Real	weight = 0.0 ;
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSumWt, (void *)&weight) ;
      weight /= (Real)(unit->numIncoming-1) ;
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSetWt, (void *)&weight) ;
    }
    if (MmaxIncomingWeight(unit->group)<MAXFLOAT)
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkMaxWt, (void *)&MmaxIncomingWeight(unit->group)) ;
    if (MminIncomingWeight(unit->group)>-MAXFLOAT)
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkMinWt, (void *)&MminIncomingWeight(unit->group)) ;
  }

  /* OUTGOING WEIGHTS */
  if (unit->numOutgoing > 0) {
    if (MequateOutgoingWeights(unit)) {
      Real	weight = 0.0 ;
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSumWt, (void *)&weight) ;
      weight /= (Real)(unit->numOutgoing) ;
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSetWt, (void *)&weight) ;
    }
    if (MmaxOutgoingWeight(unit->group)<MAXFLOAT)
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkMaxWt,(void *)&MmaxOutgoingWeight(unit->group)) ;
    if (MminOutgoingWeight(unit->group)>-MAXFLOAT)
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkMinWt,(void *)&MminOutgoingWeight(unit->group)) ;
  }
}
/**********************************************************************/
void	getCompiledWeights(net)
  Net net;
{
  int	idx ;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*wt, *dv, **weight, **deriv ;

  netForAllUnits (net, ALL, (UnitDataProc)unitConstrainWeights, NULL) ;
  for (idx=0, wt=Mwts(net), weight=Mweights(net), dv=Mdvs(net), deriv=Mderivs(net); 
       idx < numCompiledLinks ; 
       idx++) {
    (*wt++) = *(*weight++) ;
    (*dv++) = *(*deriv++) = 0.0 ;
  }
}
/**********************************************************************/
void	putCompiledDerivs(net)
  Net net;
{
  int 	idx;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*dv, **deriv ;

  for (idx = 0, dv=Mdvs(net), deriv=Mderivs(net);
       idx < numCompiledLinks ; 
       idx++) 
    *(*deriv++) += (*dv++) ;
}
/***********************************************************************/
int	command_compile(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Compiles network");
    IHelp
      (IHelpArgs,
       "Compiles network.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netCompile(currentNet) ;

  return 1 ;
}
/**********************************************************************/
