#include <commands.h>
#include <xerion/example.h>
#include <xerion/utilsDCP.h>
#include "barak.h"

#define PRINT_PERFORMANCE(net,tol)		\
  fprintf(dout, "%6.4f %d %d %f %f %f ",	\
	  MaveTime(net),			\
	  numDiffFromTarget(currentNet,0.5),	\
	  numDiffFromTarget(currentNet,tol),	\
	  aveOutputStress(currentNet),		\
	  aveOutputActivity(currentNet,MoutputActivityThresh(currentNet)), \
	  outputProximity(currentNet)) ;	\
  fflush(dout) ;

static	int	_unitIdx ;
/**********************************************************************/
static	void	overlayInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  int s, numSlices = unit->net->timeSlices ;

  for (s = 0 ; s < numSlices ; s++) {
    Real value = ((Real **)data)[s][_unitIdx] ;
    if (isNaN(Mi_slice(unit)[s]) && !isNaN(value)) 
      Mi_slice(unit)[s] = value ;
  }
  _unitIdx++ ;
}
/**********************************************************************/
static	void	overlayTarget(unit, data)
  Unit	unit ;
  void	*data ;
{
  int s, numSlices = unit->net->timeSlices ;

  for (s = 0 ; s < numSlices ; s++) {
    Real value = ((Real **)data)[s][_unitIdx] ;
    if (isNaN(Md_slice(unit)[s]) && !isNaN(value)) 
      Md_slice(unit)[s] = value ;
  }
  _unitIdx++ ;
}
/**********************************************************************/
static void	netOverlayInputs(net, v)
  Net	net ;
  Real	*v ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, INPUT, (UnitDataProc)overlayInput, (void *)v) ;
}
/**********************************************************************/
static void	netOverlayTargets(net, v)
  Net	net ;
  Real	*v ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, OUTPUT, (UnitDataProc)overlayTarget, (void *)v) ;
}
/**********************************************************************/
static void	unitCount(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(int *)data = *(int *)data + 1 ;
}
/**********************************************************************/
static void	setInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  int s ;
  for (s = 0 ; s < MnumSlices(unit->net) ; s++)
    Mi_slice(unit)[s] = *(Real *)data ;
}
/**********************************************************************/
static Proc	runNet (net)
  Net	net ;
{
  MexampleFrequency(net)	= 1.0 ;
  MaveTime(net)	 	   	= 0 ;
  MaveSlicesCorrect(net)	= 0 ;
  MaveExamplesCorrect(net)	= 0 ;

  netIterate(net, FALSE) ;

  Mslice(net) = MnumSlices(net) - 1 ;
  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    net->error = crossEntropyError(net) ;
    break ;
  case SQUARED_ERROR:
    net->error = sumSquaredError(net) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  IDoTrace(&afterExample) ;
}
/**********************************************************************/
int command_testExamples (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		target, exampleType=TRAINING, numUnits, r, reps = 1 ;
  int		numExamples, firstExample = 0 ;
  Real		tol = 0.2, initial = -1.0, isoa = -1.0 ;
  Real		*states=NULL, *inputs=NULL ;
  Boolean	useStoppingCriteria = MuseStoppingCriteria(currentNet) ;
  Boolean	runPrimingInRelaxationTrace = MrunPrimingInRelaxationTrace(currentNet) ;
  int		initializeUnits = MinitializeUnits(currentNet) ;
  Real		inputNoise      = MinputNoise(currentNet) ;
  Real		maxTimePerSlice = MmaxTimePerSlice(currentNet) ;
  ExampleSet	exampleSet ;
  String	type=NULL ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	   &type, 	 "type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-PfirstExample <int>%i%?]", &firstExample,"first example to use") ;
  Args(argd, "[-Preps <int>%i%?]", 	   &reps,	 "number of replications") ;
  Args(argd, "[-Pinitial <real>%r%?]",     &initial,     "initial input to run before target") ;
  Args(argd, "[-Pisoa <real>%r%?]",  &isoa,
       "stimulus onset asyncrony from initial input (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ptol <real>%r%?]",  &tol,
       "tolerance for binary comparison with targets (default 0.2)") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("test network on examples") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  numExamples = exampleSet->numExamples ;

  if (!Mcompiled(currentNet)) netCompile(currentNet) ;
  getCompiledWeights(currentNet) ;

  if (initial >= 0.0) {
    numUnits = netCountUnits(currentNet) ;
    states = (Real*)calloc(numUnits, sizeof(Real)) ;
    if (MtimeAverageInputs(currentNet)) {
      inputs = (Real*)calloc(numUnits, sizeof(Real)) ;
    }
    MinputNoise(currentNet)      = 0.0 ;
    MinitializeUnits(currentNet) = TRUE ;
    MrunPrimingInRelaxationTrace(currentNet) = FALSE ;
    netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&initial) ;
    if (isoa >= 0.0) {
      MmaxTimePerSlice(currentNet) = isoa ;
      MuseStoppingCriteria(currentNet) = FALSE ;
    } else { 
      MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
      MuseStoppingCriteria(currentNet) = TRUE ;
    }
    runNet(currentNet) ;
    netSaveOutputs(currentNet, states) ;
    if (MtimeAverageInputs(currentNet)) 
      netSaveTotalInputs(currentNet, inputs) ;
    MinitializeUnits(currentNet) = FALSE ;
  } else 
    MinitializeUnits(currentNet) = (MinitializeUnitsProb(currentNet) == 1.0) ;
  MinputNoise(currentNet)	   = inputNoise ;
  MmaxTimePerSlice(currentNet)     = maxTimePerSlice ;
  MuseStoppingCriteria(currentNet) = useStoppingCriteria ;
  MrunPrimingInRelaxationTrace(currentNet) = runPrimingInRelaxationTrace ;

  for (r = 0; r < reps; r++) {
    for (target = firstExample ; target < numExamples ; target++) {
      fprintf(dout, "%d %s ", target, exampleSet->example[target].name) ;
      if (initial >= 0.0) {
	netRestoreOutputs(currentNet, states) ;
	if (MtimeAverageInputs(currentNet)) 
	  netRestoreTotalInputs(currentNet, inputs) ;
      }
      MgetNumber(exampleSet, target) ;
      /* afterExample trace must output performance measure(s) */
      runNet(currentNet) ;
      PRINT_PERFORMANCE(currentNet,tol) ; fprintf(dout,"\n") ;
    }
  }
  MinitializeUnits(currentNet) = initializeUnits ;
  return 1 ;
}
/**********************************************************************/
static void runPriming(net, prime, target, primeExampleSet, targetExampleSet, states, inputs, overlay, tol)
  Net net ;
  int prime, target, overlay ;
  ExampleSet	primeExampleSet, targetExampleSet ;
  Real **states,**inputs, tol ;
  {
    netRestoreOutputs(net, states[prime]) ;
    if (MtimeAverageInputs(net)) netRestoreTotalInputs(net, inputs[prime]) ;
      MgetNumber(targetExampleSet, target) ;
      if (overlay) {
	ExampleRec *primeEx = &(primeExampleSet->example[prime]) ;
	netOverlayInputs (net, primeEx->inputVector) ;
	netOverlayTargets(net, primeEx->targetVector) ;
      }
      /* afterExample trace must output performance measure(s) */
      runNet(net) ;
      PRINT_PERFORMANCE(net,tol) ;
  }

/**********************************************************************/
int command_priming (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		numUnits ;
  int		prime, numPrimeExamples, primeExampleType=TRAINING ;
  int		target, numTargetExamples, targetExampleType=TRAINING ;
  int		overlay=FALSE, numInputs, numTargets, practice=FALSE ;
  int		adjustMinTime=FALSE ;
  ExampleSet	primeExampleSet, targetExampleSet, foilExampleSet ;
  String	primeType=NULL, targetType=NULL, foilType=NULL ;
  int		iPrime, iTarget, foilNumPairs, *foilPairs=NULL ;
  Real		**states=NULL, **inputs=NULL, tol=0.2, initial=-1.0, middle=0.0 ;
  Real		*istates, *iinputs ;
  Real		*targetRT ;
  Boolean	useStoppingCriteria = MuseStoppingCriteria(currentNet) ;
  Boolean	runPrimingInRelaxationTrace = MrunPrimingInRelaxationTrace(currentNet) ;
  int		initializeUnits = MinitializeUnits(currentNet) ;
  Real		inputNoise      = MinputNoise(currentNet) ;
  Real		minTimePerSlice = MminTimePerSlice(currentNet) ;
  Real		maxTimePerSlice = MmaxTimePerSlice(currentNet) ;
  int		decisionUpdateBaseline = MdecisionUpdateBaseline(currentNet) ;
  Real		soa = -1.0, isoa = -1.0, duration = -1.0 ;
  String	pairsfile = NULL ;
  char		pairsType[64] ;
  int		pairsPrime, pairsTarget ;
  FILE		*pairsfp ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-PprimeType <type>%s%?]", &primeType,
       "type of primes example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-PtargetType <type>%s%?]", &targetType,
       "type of targets example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-PfoilType <type>%s%?]", &foilType,
       "type of foil example set (for LD) (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Ppractice %P%?]", &practice, 
       "run practice trials to estimate decision baseline (only if foilType is set)") ;
  Args(argd, "[-Pinitial <real>%r%?]", &initial,
       "initial input to run before prime") ;
  Args(argd, "[-Pisoa <real>%r%?]", &isoa,
       "stimulus onset asyncrony of initial input (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Psoa <real>%r%?]", &soa,
       "stimulus onset asyncrony (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Pduration <real>%r%?]",  &duration,
       "prime duration (defaults to soa) (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Pmiddle <real>%r%?]", &middle,
       "input between prime and target if duration < soa (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ptol <real>%r%?]", &tol,
       "tolerance for binary comparison with targets (default 0.2)") ;
  Args(argd, "[-Poverlay %P%?]", &overlay, 
       "overlay target on prime (overwriting NANs in inputs and targets)") ;
  Args(argd, "[-PadjustMinTime %P%?]", &adjustMinTime, 
       "adjust minTimePerSlice to be no less than RT to target in isolation") ;
  Args(argd, "[-Ppairsfile <filename>%s%?]", &pairsfile,
       "file listing prime-target pairs");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("test priming among examples") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "NOTE: inputNoise only applies to targets." ,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  if (primeType) primeExampleType = findMask(primeType) ;
  primeExampleSet = netGetExampleSet(currentNet, primeExampleType) ;
  if (primeExampleSet == NULL) 
    IErrorAbort("No prime example set specified for net \"%s\".", currentNet->name) ;
  numPrimeExamples = primeExampleSet->numExamples ;

  if (targetType) targetExampleType = findMask(targetType) ;
  targetExampleSet = netGetExampleSet(currentNet, targetExampleType) ;
  if (targetExampleSet == NULL) 
    IErrorAbort("No target example set specified for net \"%s\".", currentNet->name) ;
  numTargetExamples = targetExampleSet->numExamples ;

  if (foilType) {
    foilExampleSet = netGetExampleSet(currentNet, findMask(foilType)) ;
    if (foilExampleSet == NULL) 
      IErrorAbort("No foil example set specified for net \"%s\".", currentNet->name) ;
    if (foilExampleSet->numExamples != numTargetExamples)
      IErrorAbort("Number of targets (%d) and foils (%d) must be equal", numTargetExamples, foilExampleSet->numExamples) ;
    MdecisionUpdateBaseline(currentNet) = FALSE ;
  }

  fprintf(dout, "numPrimeExamples = %d\n", numPrimeExamples) ;

  numUnits = netCountUnits(currentNet) ;
  states = (Real**)calloc(numPrimeExamples, sizeof(Real*)) ;
  for (prime = 0 ; prime < numPrimeExamples ; prime++)
    states[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  if (MtimeAverageInputs(currentNet)) {
    inputs = (Real**)calloc(numPrimeExamples, sizeof(Real*)) ;
    for (prime = 0 ; prime < numPrimeExamples ; prime++)
      inputs[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  }

  if (!Mcompiled(currentNet)) netCompile(currentNet) ;
  getCompiledWeights(currentNet) ;

  /* run and save "neutral" states  (with no prime and no noise) */
  MinputNoise(currentNet)      = 0.0 ;
  MrunPrimingInRelaxationTrace(currentNet) = FALSE ;

  /******************************************************************************
   * run and save "neutral" states  (with no prime and no noise)
   ******************************************************************************/
  if (initial >= 0.0) {
    istates = (Real*)calloc(numUnits, sizeof(Real)) ;
    if (MtimeAverageInputs(currentNet))
      iinputs = (Real*)calloc(numUnits, sizeof(Real)) ;
    MinitializeUnits(currentNet) = TRUE ;
    netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&initial) ;
    if (isoa >= 0.0) {
      MmaxTimePerSlice(currentNet) = isoa ;
      MuseStoppingCriteria(currentNet) = FALSE ;
    } else { 
      MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
      MuseStoppingCriteria(currentNet) = TRUE ;
    }
    runNet(currentNet) ;
    netSaveOutputs(currentNet, istates) ;
    if (MtimeAverageInputs(currentNet)) 
      netSaveTotalInputs(currentNet, iinputs) ;
    MinitializeUnits(currentNet) = FALSE ;
  }

  /******************************************************************************
   * run and save prime states
   ******************************************************************************/
  if (duration < 0.0) duration = soa ;
  if (pairsfile) {
    if (!(pairsfp=fopen(pairsfile, "r")))
      IErrorAbort("Unable to open pairsfile \"%s\".", pairsfile) ;
    for (prime = 0 ; prime < numPrimeExamples ; prime++) states[prime][0] = -1.0 ;
    prime = -1 ;
    while (fscanf(pairsfp, "%d %d %s", &pairsPrime, &pairsTarget, pairsType) != EOF) {

      fprintf(dout, "prime %d target %d\n", pairsPrime, pairsTarget) ;

      if (pairsPrime == prime) continue ;
      if (states[pairsPrime][0] != -1.0) continue ;
      fprintf(dout, "prime %d ", pairsPrime) ;
      MinitializeUnits(currentNet) = TRUE ;
      if (initial >= 0.0) {
	netRestoreOutputs(currentNet, istates) ;
	if (MtimeAverageInputs(currentNet)) 
	  netRestoreTotalInputs(currentNet, iinputs) ;
      }
      if (duration >= 0.0) {
	MmaxTimePerSlice(currentNet) = duration ;
	MuseStoppingCriteria(currentNet) = FALSE ;
      } else {	/* no soa specified - run to completion (however specified) */
	MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
	MuseStoppingCriteria(currentNet) = TRUE ;
      }      
      /* afterExample trace must output performance measure(s) */
      MgetNumber(primeExampleSet, pairsPrime) ;
      runNet(currentNet) ;
      PRINT_PERFORMANCE(currentNet,tol) ; 
      if (soa > duration) {
	netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&middle) ;
	MinitializeUnits(currentNet) = FALSE ;
	MmaxTimePerSlice(currentNet) = soa-duration ;
	runNet(currentNet) ;
	PRINT_PERFORMANCE(currentNet,tol) ;
      }
      fprintf(dout, "\n") ;
      netSaveOutputs(currentNet, states[pairsPrime]) ;
      if (MtimeAverageInputs(currentNet)) 
	netSaveTotalInputs(currentNet, inputs[pairsPrime]) ;
      prime = pairsPrime ;
    }
    fclose(pairsfp) ;
  } else {
    for (prime = 0 ; prime < numPrimeExamples ; prime++) {
      fprintf(dout, "prime %d ", prime) ;
      MinitializeUnits(currentNet) = TRUE ;
      if (initial >= 0.0) {
	netRestoreOutputs(currentNet, istates) ;
	if (MtimeAverageInputs(currentNet)) 
	  netRestoreTotalInputs(currentNet, iinputs) ;
      }
      if (duration >= 0.0) {
	MmaxTimePerSlice(currentNet) = duration ;
	MuseStoppingCriteria(currentNet) = FALSE ;
      } else {	/* no soa specified - run to completion (however specified) */
	MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
	MuseStoppingCriteria(currentNet) = TRUE ;
      }      
      /* afterExample trace must output performance measure(s) */
      MgetNumber(primeExampleSet, prime) ;
      runNet(currentNet) ;
      PRINT_PERFORMANCE(currentNet,tol) ; 
      if (soa > duration) {
	netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&middle) ;
	MinitializeUnits(currentNet) = FALSE ;
	MmaxTimePerSlice(currentNet) = soa-duration ;
	runNet(currentNet) ;
	PRINT_PERFORMANCE(currentNet,tol) ;
      }
      fprintf(dout, "\n") ;
      netSaveOutputs(currentNet, states[prime]) ;
      if (MtimeAverageInputs(currentNet)) 
	netSaveTotalInputs(currentNet, inputs[prime]) ;
    }
  }
  /******************************************************************************
   * run targets (and maybe foils) after every other word as prime
   ******************************************************************************/
  MinitializeUnits(currentNet)     		= FALSE ;
  MinputNoise(currentNet)	   		= inputNoise ;
  MmaxTimePerSlice(currentNet)     		= maxTimePerSlice ;
  MuseStoppingCriteria(currentNet)		= useStoppingCriteria ;
  MrunPrimingInRelaxationTrace(currentNet)	= runPrimingInRelaxationTrace ;
  if (foilType) {
    MdecisionUpdateBaseline(currentNet) = decisionUpdateBaseline ;
    if (practice)
      /* run targets and foils once with permuted primes */
      for (prime = 0, target = 0 ; prime < numPrimeExamples ; prime++, target=prime) {
	PERMUTE(prime,numPrimeExamples) ;
	fprintf(dout, "pr %d %d ", prime, target) ;
	runPriming(currentNet, prime, target, primeExampleSet, targetExampleSet, states, inputs, overlay, tol) ;
	if (!MdecisionUpdateBaseline(currentNet) && MdecisionResponse(currentNet) != DECISION_YES)
	  MdecisionInitialBaseline(currentNet) -= MdecisionBaselineIncrement(currentNet) ;
	runPriming(currentNet, prime, target, primeExampleSet, foilExampleSet,   states, inputs, overlay, tol) ;
	if (!MdecisionUpdateBaseline(currentNet) && MdecisionResponse(currentNet) != DECISION_NO)
	  MdecisionInitialBaseline(currentNet) += MdecisionBaselineIncrement(currentNet) ;
	fprintf(dout, "\n") ;
	PERMUTE_RESTORE(prime,numPrimeExamples) ;
      }
  }
  if (pairsfile) {
    pairsfp=fopen(pairsfile, "r") ;
    while (fscanf(pairsfp, "%d %d %s", &prime, &target, pairsType) != EOF) {
      MminTimePerSlice(currentNet) = minTimePerSlice ;
      if (adjustMinTime) {
	int initializeUnits = MinitializeUnits(currentNet) ;
	MinitializeUnits(currentNet) = TRUE ;
	MgetNumber(targetExampleSet, target) ;
	fprintf(dout, "minTimePerSlice = target RT = ") ;
	runNet(currentNet) ;
	MminTimePerSlice(currentNet) = MaveTime(currentNet) ;
	MinitializeUnits(currentNet) = initializeUnits ;
      }
      if (foilType) {
	PERMUTE_PAIR(prime,numPrimeExamples,target,numTargetExamples) ;
	fprintf(dout, "%d %d %s ", prime, target, pairsType) ;
	runPriming(currentNet, prime, target, primeExampleSet, targetExampleSet, states, inputs, overlay, tol) ;
	if (!MdecisionUpdateBaseline(currentNet) && MdecisionResponse(currentNet) != DECISION_YES)
	  MdecisionInitialBaseline(currentNet) -= MdecisionBaselineIncrement(currentNet) ;
	runPriming(currentNet, prime, target, primeExampleSet, foilExampleSet, states, inputs, overlay, tol) ;
	if (!MdecisionUpdateBaseline(currentNet) && MdecisionResponse(currentNet) != DECISION_NO)
	  MdecisionInitialBaseline(currentNet) += MdecisionBaselineIncrement(currentNet) ;
	PERMUTE_RESTORE_PAIR(prime,numPrimeExamples,target,numTargetExamples) ;
      } else {
	fprintf(dout, "%d %d %s ", prime, target, pairsType) ;
	runPriming(currentNet, prime, target, primeExampleSet, targetExampleSet, states, inputs, overlay, tol) ;
      }
      fprintf(dout,"\n") ;
    }
    if (pairsfile) fclose(pairsfp) ;

  } else {

    for (target = 0 ; target < numTargetExamples ; target++) {

      MminTimePerSlice(currentNet) = minTimePerSlice ;
      if (adjustMinTime) {
	int initializeUnits = MinitializeUnits(currentNet) ;
	MinitializeUnits(currentNet) = TRUE ;
	MgetNumber(targetExampleSet, target) ;
	fprintf(dout, "minTimePerSlice = target RT = ") ;
	runNet(currentNet) ;
	MminTimePerSlice(currentNet) = MaveTime(currentNet) ;
	MinitializeUnits(currentNet) = initializeUnits ;
      }

      for (prime = 0 ; prime < numPrimeExamples ; prime++) {
	if (foilType) {
	  PERMUTE_PAIR(prime,numPrimeExamples,target,numTargetExamples) ;
	  fprintf(dout, "%d %d ", prime, target) ;
	  runPriming(currentNet, prime, target, primeExampleSet, targetExampleSet, states, inputs, overlay, tol) ;
	  if (!MdecisionUpdateBaseline(currentNet) && MdecisionResponse(currentNet) != DECISION_YES)
	    MdecisionInitialBaseline(currentNet) -= MdecisionBaselineIncrement(currentNet) ;
	  runPriming(currentNet, prime, target, primeExampleSet, foilExampleSet, states, inputs, overlay, tol) ;
	  if (!MdecisionUpdateBaseline(currentNet) && MdecisionResponse(currentNet) != DECISION_NO)
	    MdecisionInitialBaseline(currentNet) += MdecisionBaselineIncrement(currentNet) ;
	  PERMUTE_RESTORE_PAIR(prime,numPrimeExamples,target,numTargetExamples) ;
	} else {
	  fprintf(dout, "%d %d ", prime, target) ;
	  runPriming(currentNet, prime, target, primeExampleSet, targetExampleSet, states, inputs, overlay, tol) ;
	}
	fprintf(dout,"\n") ;
      } /* prime */
    } /* target */
  }
  MinitializeUnits(currentNet) = initializeUnits ;
  for (prime = 0 ; prime < numPrimeExamples ; prime++) free(states[prime]) ;
  free(states) ;
  if (MtimeAverageInputs(currentNet)) {
    for (prime = 0 ; prime < numPrimeExamples ; prime++) free(inputs[prime]) ;
    free(inputs) ;
  }
  return 1 ;
}

/**********************************************************************/
int command_spanAssocPriming (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		prime, middle, target, numUnits ;
  int		numExamples, firstExample=0, exampleType=TRAINING ;
  ExampleSet	exampleSet ;
  String	type=NULL ;
  Real		**states=NULL, **inputs=NULL ;
  Real		initial = -1.0, isoa = -1.0, psoa = -1.0, msoa = -1.0, tol = 0.2 ;
  Boolean	useStoppingCriteria = MuseStoppingCriteria(currentNet) ;
  int		initializeUnits = MinitializeUnits(currentNet) ;
  Real		inputNoise      = MinputNoise(currentNet) ;
  Real		maxTimePerSlice = MmaxTimePerSlice(currentNet) ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", &type,
       "type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pinitial <real>%r%?]",  &initial,
       "initial input to run before prime") ;
  Args(argd, "[-Pisoa <real>%r%?]",  &isoa,
       "stimulus onset asyncrony of initial input (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ppsoa <real>%r%?]",  &psoa,
       "stimulus onset asyncrony of prime (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Pmsoa <real>%r%?]",  &msoa,
       "stimulus onset asyncrony of middle word (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ptol <real>%r%?]",  &tol,
       "tolerance for binary comparison with targets (default 0.2)") ;
  Args(argd, "[-PfirstExample <int>%i%?]", &firstExample,"first example to use") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("test priming spanning intermediate item") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "NOTE: inputNoise only applies to targets." ,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) 
    IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  numExamples = exampleSet->numExamples ;

  numUnits = netCountUnits(currentNet) ;
  states = (Real**)calloc(numExamples, sizeof(Real*)) ;
  for (prime = 0 ; prime < numExamples ; prime++)
    states[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  if (MtimeAverageInputs(currentNet)) {
    inputs = (Real**)calloc(numExamples, sizeof(Real*)) ;
    for (prime = 0 ; prime < numExamples ; prime++)
      inputs[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  }
  if (!Mcompiled(currentNet)) netCompile(currentNet) ;
  getCompiledWeights(currentNet) ;

  /* run and save "neutral" states  (with no prime and no noise) */
  MinputNoise(currentNet)      		= 0.0 ;
  for (prime = 0 ; prime < numExamples ; prime++) {
    fprintf(dout, "none %d ", prime) ;
    MinitializeUnits(currentNet) 		= TRUE ;
    if (initial >= 0.0) {
      MgetNumber(exampleSet, prime) ;
      netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&initial) ;
      if (isoa >= 0.0) {
	MmaxTimePerSlice(currentNet) = isoa ;
	MuseStoppingCriteria(currentNet) = FALSE ;
      } else { 
	MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
	MuseStoppingCriteria(currentNet) = TRUE ;
      }
      runNet(currentNet) ;
      MinitializeUnits(currentNet) = FALSE ;
    }
    if (psoa >= 0.0) {
      MmaxTimePerSlice(currentNet) = psoa ;
      MuseStoppingCriteria(currentNet) = FALSE ;
    } else {
      MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
      MuseStoppingCriteria(currentNet) = TRUE ;
    }      
    MgetNumber(exampleSet, prime) ;
    /* afterExample trace must output performance measure(s) */
    runNet(currentNet) ;
    PRINT_PERFORMANCE(currentNet,tol) ; fprintf(dout,"\n") ;
    netSaveOutputs(currentNet, states[prime]) ;
    if (MtimeAverageInputs(currentNet)) 
      netSaveTotalInputs(currentNet, inputs[prime]) ;
  }
  /* run association after every other word as middle word */
  MinitializeUnits(currentNet)     = FALSE ;
  for (prime = firstExample ; prime < numExamples ; prime++) {
    target = exampleSet->example[prime].extension->assoc[0] ;
    for (middle = 0 ; middle < numExamples ; middle++) {
      fprintf(dout, "%d %d %d ", prime, middle, target) ;
      /* restore prime */
      netRestoreOutputs(currentNet, states[prime]) ;
      if (MtimeAverageInputs(currentNet)) 
	netRestoreTotalInputs(currentNet, inputs[prime]) ;
      /* run middle word */
      MinputNoise(currentNet)	   	= 0.0 ;
      if (msoa >= 0.0) {
	MmaxTimePerSlice(currentNet)	 = msoa ;
	MuseStoppingCriteria(currentNet) = FALSE ;
      } else {
	MmaxTimePerSlice(currentNet) 	 = maxTimePerSlice ;
	MuseStoppingCriteria(currentNet) = TRUE ;
      }      
      MgetNumber(exampleSet, middle) ;
      runNet(currentNet) ;      /* afterExample trace must output performance measure(s) */
      PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
      /* run target */
      MinputNoise(currentNet)	   	= inputNoise ;
      MmaxTimePerSlice(currentNet) 	= maxTimePerSlice ;
      MuseStoppingCriteria(currentNet)	= useStoppingCriteria ;
      MgetNumber(exampleSet, target) ;
      runNet(currentNet) ;      /* afterExample trace must output performance measure(s) */
      PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
    }
  }
  MinitializeUnits(currentNet) = initializeUnits ;
  for (prime = 0 ; prime < numExamples ; prime++) free(states[prime]) ;
  free(states) ;
  if (MtimeAverageInputs(currentNet)) {
    for (prime = 0 ; prime < numExamples ; prime++) free(inputs[prime]) ;
    free(inputs) ;
  }
  fprintf(dout, "\n") ;
  return 1 ;
}
/**********************************************************************/
int command_spanAssocNeutral (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		prime, middle, target, numUnits, numExamples, firstExample=0, exampleType=TRAINING ;
  int		numCatMembers = 8, nmiddle, numMiddle = 1 ;
  ExampleSet	exampleSet ;
  String	type=NULL ;
  Real		**states=NULL, **inputs=NULL ;
   Real		initial = -1.0, isoa = -1.0, psoa = -1.0, msoa = -1.0, tol = 0.2 ;
 Boolean	useStoppingCriteria = MuseStoppingCriteria(currentNet) ;
  int		initializeUnits = MinitializeUnits(currentNet) ;
  Real		inputNoise      = MinputNoise(currentNet) ;
  Real		maxTimePerSlice = MmaxTimePerSlice(currentNet) ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", &type,
       "type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pinitial <real>%r%?]",  &initial,
       "initial input to run before prime") ;
  Args(argd, "[-Pisoa <real>%r%?]",  &isoa,
       "stimulus onset asyncrony of initial input (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ppsoa <real>%r%?]",  &psoa,
       "stimulus onset asyncrony of prime (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Pmsoa <real>%r%?]",  &msoa,
       "stimulus onset asyncrony of middle word (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ptol <real>%r%?]",  &tol,
       "tolerance for binary comparison with targets (default 0.2)") ;
  Args(argd, "[-PfirstExample <int>%i%?]", &firstExample,"first example to use") ;
  Args(argd, "[-PnumCatMembers <int>%i%?]",  &numCatMembers,
       "number of members in each category") ;
  Args(argd, "[-PnumMiddle <int>%i%?]",  &numMiddle,
       "number of random (unrelated) middle items") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("test neutral condition spanning intermediate item") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "NOTE: inputNoise only applies to targets." ,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) 
    IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  numExamples = exampleSet->numExamples ;

  numUnits = netCountUnits(currentNet) ;
  states = (Real**)calloc(numExamples, sizeof(Real*)) ;
  for (prime = 0 ; prime < numExamples ; prime++)
    states[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  if (MtimeAverageInputs(currentNet)) {
    inputs = (Real**)calloc(numExamples, sizeof(Real*)) ;
    for (prime = 0 ; prime < numExamples ; prime++)
      inputs[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  }
  if (!Mcompiled(currentNet)) netCompile(currentNet) ;
  getCompiledWeights(currentNet) ;

  /* run and save "neutral" states  (with no prime and no noise) */
  MinputNoise(currentNet)      		= 0.0 ;
  for (prime = 0 ; prime < numExamples ; prime++) {
    fprintf(dout, "none %d ", prime) ;
    MinitializeUnits(currentNet) = TRUE ;
    if (initial >= 0.0) {
      MgetNumber(exampleSet, prime) ;
      netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&initial) ;
      if (isoa >= 0.0) {
	MmaxTimePerSlice(currentNet) = isoa ;
	MuseStoppingCriteria(currentNet) = FALSE ;
      } else { 
	MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
	MuseStoppingCriteria(currentNet) = TRUE ;
      }
      runNet(currentNet) ;
      MinitializeUnits(currentNet) = FALSE ;
    }
    if (psoa >= 0.0) {
      MmaxTimePerSlice(currentNet) = psoa ;
      MuseStoppingCriteria(currentNet) = FALSE ;
    } else {
      MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
      MuseStoppingCriteria(currentNet) = TRUE ;
    }      
    MgetNumber(exampleSet, prime) ;
    /* afterExample trace must output performance measure(s) */
    runNet(currentNet) ;
    PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
    netSaveOutputs(currentNet, states[prime]) ;
    if (MtimeAverageInputs(currentNet)) 
      netSaveTotalInputs(currentNet, inputs[prime]) ;
  }
  /* run association after every other word as middle word */
  MinitializeUnits(currentNet)     = FALSE ;
  for (prime = firstExample ; prime < numExamples ; prime++) {
    for (target = 0 ; target < numExamples ; target++) {
      for (nmiddle = 0 ; nmiddle < numMiddle ; nmiddle++) {
	do {
	  middle = randomInteger(numExamples) ;
	} while ((middle/numCatMembers == target/numCatMembers) ||
		 (target == exampleSet->example[middle].extension->assoc[0])) ;
	fprintf(dout, "%d %d %d ", prime, middle, target) ;
	/* restore prime */
	netRestoreOutputs(currentNet, states[prime]) ;
	if (MtimeAverageInputs(currentNet)) 
	  netRestoreTotalInputs(currentNet, inputs[prime]) ;
	/* run middle word */
	MinputNoise(currentNet)	   	= 0.0 ;
	if (msoa >= 0.0) {
	  MmaxTimePerSlice(currentNet)	 = msoa ;
	  MuseStoppingCriteria(currentNet) = FALSE ;
	} else {
	  MmaxTimePerSlice(currentNet) 	 = maxTimePerSlice ;
	  MuseStoppingCriteria(currentNet) = TRUE ;
	}      
	MgetNumber(exampleSet, middle) ;
	runNet(currentNet) ;      /* afterExample trace must output performance measure(s) */
	PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
	/* run target */
	MinputNoise(currentNet)	   	= inputNoise ;
	MmaxTimePerSlice(currentNet) 	= maxTimePerSlice ;
	MuseStoppingCriteria(currentNet)= useStoppingCriteria ;
	MgetNumber(exampleSet, target) ;
	runNet(currentNet) ;      /* afterExample trace must output performance measure(s) */
	PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
      }
    }
  }
  MinitializeUnits(currentNet) = initializeUnits ;
  for (prime = 0 ; prime < numExamples ; prime++) free(states[prime]) ;
  free(states) ;
  if (MtimeAverageInputs(currentNet)) {
    for (prime = 0 ; prime < numExamples ; prime++) free(inputs[prime]) ;
    free(inputs) ;
  }
  fprintf(dout, "\n") ;
  return 1 ;
}
/**********************************************************************/
int command_mediatedPriming (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		prime, middle, target, numUnits, numExamples, firstExample=0, exampleType=TRAINING ;
  ExampleSet	exampleSet ;
  String	type=NULL ;
  Real		**states=NULL, **inputs=NULL ;
  Real		initial = -1.0, isoa = -1.0, soa = -1.0, tol = 0.2 ;
  Boolean	useStoppingCriteria = MuseStoppingCriteria(currentNet) ;
  int		initializeUnits = MinitializeUnits(currentNet) ;
  Real		inputNoise      = MinputNoise(currentNet) ;
  Real		maxTimePerSlice = MmaxTimePerSlice(currentNet) ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", &type,
       "type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pinitial <real>%r%?]",  &initial,
       "initial input to run before prime") ;
  Args(argd, "[-Pisoa <real>%r%?]",  &isoa,
       "stimulus onset asyncrony of initial input (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Psoa <real>%r%?]",  &soa,
       "stimulus onset asyncrony (ONLY WORKS FOR 1 SLICE)") ;
  Args(argd, "[-Ptol <real>%r%?]",  &tol,
       "tolerance for binary comparison with targets (default 0.2)") ;
  Args(argd, "[-PfirstExample <int>%i%?]", &firstExample,"first example to use") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("test priming among examples") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "NOTE: inputNoise only applies to targets." ,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) 
    IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  numExamples = exampleSet->numExamples ;

  numUnits = netCountUnits(currentNet) ;
  states = (Real**)calloc(numExamples, sizeof(Real*)) ;
  for (prime = 0 ; prime < numExamples ; prime++)
    states[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  if (MtimeAverageInputs(currentNet)) {
    inputs = (Real**)calloc(numExamples, sizeof(Real*)) ;
    for (prime = 0 ; prime < numExamples ; prime++)
      inputs[prime] = (Real*)calloc(numUnits, sizeof(Real)) ;
  }
  if (!Mcompiled(currentNet)) netCompile(currentNet) ;
  getCompiledWeights(currentNet) ;

  /* run and save "neutral" states  (with no prime and no noise) */
  MinputNoise(currentNet)      = 0.0 ;
  for (prime = 0 ; prime < numExamples ; prime++) {
    fprintf(dout, "none %d ", prime) ;
    MinitializeUnits(currentNet) = TRUE ;
    if (initial >= 0.0) {
      MgetNumber(exampleSet, prime) ;
      netForAllUnits(currentNet, INPUT, (UnitDataProc)setInput, (void *)&initial) ;
      if (isoa >= 0.0) {
	MmaxTimePerSlice(currentNet) = isoa ;
	MuseStoppingCriteria(currentNet) = FALSE ;
      } else { 
	MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
	MuseStoppingCriteria(currentNet) = TRUE ;
      }
      runNet(currentNet) ;
      MinitializeUnits(currentNet) = FALSE ;
    }
    if (soa >= 0.0) {
      MmaxTimePerSlice(currentNet) = soa ;
      MuseStoppingCriteria(currentNet) = FALSE ;
    } else {
      MmaxTimePerSlice(currentNet) = maxTimePerSlice ;
      MuseStoppingCriteria(currentNet) = TRUE ;
    }      
    MgetNumber(exampleSet, prime) ;
    /* afterExample trace must output performance measure(s) */
    runNet(currentNet) ;
    PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
    netSaveOutputs(currentNet, states[prime]) ;
    if (MtimeAverageInputs(currentNet)) 
      netSaveTotalInputs(currentNet, inputs[prime]) ;
  }
  /* run after every other word as prime */
  MinitializeUnits(currentNet)     = FALSE ;
  MinputNoise(currentNet)	   = inputNoise ;
  MmaxTimePerSlice(currentNet)     = maxTimePerSlice ;
  MuseStoppingCriteria(currentNet) = useStoppingCriteria ;
  for (prime = firstExample ; prime < numExamples ; prime++) {
    middle = exampleSet->example[prime].extension->assoc[0] ;
    target = exampleSet->example[middle].extension->assoc[0] ;
    fprintf(dout, "%d %d ", prime, target) ;
    netRestoreOutputs(currentNet, states[prime]) ;
    if (MtimeAverageInputs(currentNet)) 
      netRestoreTotalInputs(currentNet, inputs[prime]) ;
    MgetNumber(exampleSet, target) ;
    /* afterExample trace must output performance measure(s) */
    runNet(currentNet) ;
    PRINT_PERFORMANCE(currentNet, tol) ; fprintf(dout,"\n") ;
  }
  MinitializeUnits(currentNet) = initializeUnits ;
  for (prime = 0 ; prime < numExamples ; prime++) free(states[prime]) ;
  free(states) ;
  if (MtimeAverageInputs(currentNet)) {
    for (prime = 0 ; prime < numExamples ; prime++) free(inputs[prime]) ;
    free(inputs) ;
  }
  fprintf(dout, "\n") ;
  return 1 ;
}
/**********************************************************************/
