
/**********************************************************************
 * $Id: main.c,v 1.2 92/11/30 11:33:10 drew Exp $
 **********************************************************************/

/**********************************************************************
 *   Copyright 1990,1991,1992,1993 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#ifndef linux
#include <nan.h>
#endif
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "chl.h"

struct TRACE    inRelaxation ;
struct TRACE    afterExample ;

static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;
static Proc	updateNetActivity      ARGS((Net	net)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initGroup    		ARGS((Group	 group)) ;
static void	deinitGroup    		ARGS((Group	 group)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExample	    	ARGS((Example    example)) ;
static void	deinitExample		ARGS((Example    example)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;

/**********************************************************************
 *				     Main
 **********************************************************************/
int main(argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateGroupHook(initGroup) ;
  setDestroyGroupHook(deinitGroup) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleHook(initExample) ;
  setDestroyExampleHook(deinitExample) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}

/**********************************************************************
 *				Initialization
 **********************************************************************/
static void	initNet(net)
  Net	net ;
{
  /* These MUST be set for the simulator to work properly */
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  /* non-zero defaults */
  MtempMin(net)		= 1.0 ;
  MtempMax(net)		= 1.0 ;
  MtempDecay(net)	= 1.0 ;
  Mdt(net)		= 0.2 ;
  MmaxUpdates(net)	= 100 ;
  MinitialOutput(net)	= 0.2 ;
}

static void	deinitNet(net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}

/***********************************************************************/
static void	initGroup(group)
  Group	group ;
{
  group->extension = (GroupExtension)calloc(1, sizeof(GroupExtensionRec)) ;
}

static void	deinitGroup(group)
  Group	group ;
{
  if (group->extension != NULL)
    free(group->extension) ;
}

/**********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
  Mlesioned(unit) = FALSE ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}

/**********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;
  Mlesioned(link) = FALSE ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}

/***********************************************************************/
static void	initExample(example)
  Example  example ;
{
  example->extension = (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;
}

static void	deinitExample(example)
  Example  example ;
{
  if (example->extension != NULL)
    free(example->extension) ;
}

/**********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleSetExtension)calloc(1, sizeof(ExampleSetExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}

/*********************************************************************
 * 				Link Length
 *********************************************************************/
void		setLinkLength(link, length)
  Link		link ;
  Real		length ;
{
  /* DO NOTHING */
  /* Mlength(link) = length ; */

}

Real		getLinkLength(link)
  Link		link ;
{
  /* DO NOTHING */
  /* return Mlength(link) ; */
  return 1.0 ;
}

int		unitXPosition(unit)
  Unit	unit ;
{
  return MxPosition(unit) ;
}

int		unitYPosition(unit)
  Unit	unit ;
{
  return MyPosition(unit) ;
}


/*********************************************************************
 * 				Lesioning
 *********************************************************************/
void		lesionUnit(unit)
  Unit		unit ;
{
  unit->output = 0.0 ;
  Mlesioned(unit) = TRUE ;
}
void		unlesionUnit(unit)
  Unit		unit ;
{
  Mlesioned(unit) = FALSE ;
}
Boolean		isLesionedUnit(unit)
  Unit		unit ;
{
  return Mlesioned(unit) ;
}
void		lesionLink(link)
  Link		link ;
{
    link->deriv = 0.0 ;
    MlesionedWeight(link) = link->weight ;
    link->weight = 0.0 ;
    Mlesioned(link) = TRUE ;
}
void		unlesionLink(link)
  Link		link ;
{
  if (Mlesioned(link)) link->weight = MlesionedWeight(link) ;
  Mlesioned(link) = FALSE ;
}
Boolean		isLesionedLink(link)
  Link		link ;
{
  return Mlesioned(link) ;
}

/**********************************************************************
 *				   Utilities
 **********************************************************************/
static void	unitZeroLinkDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Link	*linkPtr  = unit->incomingLink ;
  Link	*finalPtr = linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link	link = *(linkPtr++) ;
    link->deriv = 0.0 ;
  }
}

/**********************************************************************/
static void	unitClampTarget(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output = unit->target ;
}

/**********************************************************************/
static void	unitInitOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output = MinitialOutput(unit->net) ;
}

/**********************************************************************/
static void	unitUpdateOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output = MnewOutput(unit) ;
}

/**********************************************************************/
static void	unitDecayWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  Net	net		= unit->net ;
  Real	weightCost	= MweightCost(net) ;
  Link	*linkPtr	= unit->incomingLink ;
  Link	*finalPtr	= linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link link   = *(linkPtr++) ;
    Real wtCost = link->weight * weightCost ;
    net->cost   += wtCost * link->weight ;
    link->deriv += 2.0 * wtCost ;
  }
}

/**********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  if (MbatchSize(net) < exampleSet->numExamples) {
    /* select random example according to it's probability */
    do {
      MexampleNum(net) = randomInteger(exampleSet->numExamples);
      MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
    } while (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net))) ;
  } else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      return FALSE ;       /* skip this example */
  }
  MgetNumber(exampleSet, MexampleNum(net)) ;
  return TRUE ;
}

/**********************************************************************
 *				 Unit updating
 **********************************************************************/
static void	unitComputeOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net		= unit->net ;
  Real	totalInput	= 0.0 ;
  Real	sigIn, outputChange ;
  Link	*linkPtr	= unit->incomingLink ;
  Link	*finalPtr	= linkPtr + unit->numIncoming ;
  
  while (linkPtr != finalPtr) {
    Link link   = *(linkPtr++) ;
    totalInput += link->preUnit->output * link->weight ;
  }
  totalInput	       /= Mtemperature(net) ;
  unit->totalInput	= totalInput ;
  sigIn			= sigmoid(totalInput) ;
  outputChange		= sigIn - unit->output ;
  MmaxOutputChange(net) = MAX(MmaxOutputChange(net),outputChange) ;
  /* synchronous updating - will be copied to output at end of update */
  MnewOutput(unit)	= unit->output + outputChange * Mdt(net) ;
}

/**********************************************************************/
static void	unitUpdateDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net		= unit->net ;
  Link	*linkPtr	= unit->incomingLink ;
  Link	*finalPtr	= linkPtr + unit->numIncoming ;
  
  /* signs are reversed to do gradient descent */
  if ((int)data)
    /* positive phase */
    while (linkPtr != finalPtr) {
      Link link    = *(linkPtr++) ;
      link->deriv -= link->preUnit->output * link->postUnit->output ;
    }
  else
    /* negative phase */
    while (linkPtr != finalPtr) {
      Link link    = *(linkPtr++) ;
      link->deriv += link->preUnit->output * link->postUnit->output ;
    }
}

/**********************************************************************
 *				Error functions
 **********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error ;
  Real	output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
  error  = INFO(output, target) ; 
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}

/**********************************************************************/
static void	unitIncrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}

/**********************************************************************/
Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  groupForAllUnits(group, (UnitDataProc)unitIncrementError, &error) ;
  return error ;
}

/**********************************************************************
 *				   Annealing
 **********************************************************************/
static void	netAnneal(net, positivePhase)
  Net		net ;
  Boolean	positivePhase ;
{
  int updateMask = (positivePhase ? ~(INPUT | OUTPUT | BIAS) : ~(INPUT | BIAS)) ;

  netForAllUnits(net, updateMask, (UnitDataProc)unitInitOutput, NULL) ;
  if (positivePhase) 
    netForAllUnits(net, OUTPUT, (UnitDataProc)unitClampTarget, NULL) ;

  MnumUpdates(net) = 0 ;
  Mtemperature(net) = MtempMax(net) ;
  do {
    MmaxOutputChange(net) = 0.0 ;

    /* synchronous updating */
    netForAllUnits(net, updateMask, (UnitDataProc)unitComputeOutput, NULL) ;
    netForAllUnits(net, updateMask, (UnitDataProc)unitUpdateOutput, NULL) ;

    if (!Mrunning(net)) {
      int idx, delayCount = MdelayCount(net), tmp = 0 ;
      IDoTrace(&inRelaxation) ;
      /* this nonsense is just to slow the display and fool the compiler */
      for (idx = 0, tmp = net->currentTime ; idx < delayCount*10000 ; idx++) net->currentTime++ ;
      net->currentTime = tmp ;
    }
    MnumUpdates(net)++ ;

    if (Mdebug(net))
      fprintf(dout, "%d: %f %f\n", 
	      MnumUpdates(net), Mtemperature(net), MmaxOutputChange(net)) ;

    Mtemperature(net) = MtempMin(net) + 
      (Mtemperature(net)-MtempMin(net))*MtempDecay(net) ;
  } while (MnumUpdates(net) < MmaxUpdates(net) && 
	   MmaxOutputChange(net) > MoutputChangeTolerance(net)) ;

  if (Mdebug(net)) {
    if (MnumUpdates(net) >= MmaxUpdates(net)) fprintf(dout,"Exceeded maxUpdates\n") ;
    if (MmaxOutputChange(net) <= MoutputChangeTolerance(net)) fprintf(dout,"Reached outputChangeTolerance\n") ;
  }

  if (!positivePhase)
    netForAllUnits(net, OUTPUT, (UnitDataProc)unitIncrementError, (void *)&net->error) ;
}

/**********************************************************************
 *				 Entry points
 **********************************************************************/
static Proc	updateNetActivity (net)
  Net	net ;
{
  Mrunning(net) = FALSE ;
  net->error               	= 0.0 ;
  MexampleFrequency(net)	= 1.0 ;

  netAnneal(net, FALSE) ;

  IDoTrace(&afterExample) ;
}

/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int	example ;

  Mrunning(net) = TRUE ;
  net->error = 0.0 ;
  net->cost  = 0.0 ;
  initEpoch(net, MalphaOnset(net)) ;
  netForAllUnits(net, ALL, (UnitDataProc)unitZeroLinkDerivs, net) ;
  for (example = 0 ; example < MbatchSize(net) ; ++example) {
    if (!getExample(net, exampleSet, example)) continue ;
    netAnneal(net, TRUE) ;
    netForAllUnits(net, ALL, (UnitDataProc)unitUpdateDerivs, (void *)TRUE) ;
    /* do negative phase last so final displayed states are correct */
    netAnneal(net, FALSE) ;
    netForAllUnits(net, ALL, (UnitDataProc)unitUpdateDerivs, (void *)FALSE) ;
  }
  if (example <= 0) IErrorAbort("calculateNetErrorDeriv: no examples processed") ;
  netForAllUnits(net, ALL, (UnitDataProc)unitDecayWeights, NULL) ;
  Mrunning(net) = FALSE ;
}

/**********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int	example ;

  Mrunning(net) = TRUE ;
  net->error = 0.0 ;
  for (example = 0 ; example < MbatchSize(net) ; ++example) {
    if (!getExample(net, exampleSet, example)) continue ;
    netAnneal(net, FALSE) ;
  }
  if (example <= 0) IErrorAbort("calculateNetError: no examples processed") ;
  Mrunning(net) = FALSE ;
}
/**********************************************************************/
