
/**********************************************************************
 * $Id: bp.h,v 1.7 91/09/26 11:05:16 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#ifndef ELMAN_H
#define ELMAN_H

/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  Boolean	running ;
  int	delayCount ;		/* netParam: */
  Real	zeroErrorRadius ;	/* netParam: */
  Real	correctTolerance ;	/* netParam: */
  Real	weightCost ;		/* netParam: */
  Real	standardWeight ;	/* netParam: */
  Real	initialUnitOutput ;	/* netParam: */
  Real	inputNoise ;		/* netParam: */
  Real	inputStrength ;		/* netParam: */
  Real  relaxStepSize ;         /* netParam: */
  int	maxRelaxSteps ;         /* netParam: */
  int	alphaOnset ;		/* netParam: */
  int	refixate ;		/* netParam: */
  Real	aveFixations ;
  Real	avePhonemesCorrect ;	
  Real	aveWordsCorrect ;	
  Real	aveRelaxSteps ;	
  int	teacherForcing ;	/* netParam: */
  int	numSlices ;
  int	exampleNum ;
  Real	exampleFrequency ;
  int	sampleExamples ;	/* netParam: */
  Boolean	compiled ;
} NetExtensionRec ;

typedef struct GroupExtensionRec {	/* BIND */
  int	foo ;
} GroupExtensionRec ;

typedef struct UnitExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		oldOutput ;
  int		numCompiledIncoming ;
  Real*		*preOutputs ;		
  Real*		*preOutputDerivs ;	
  Real		*weights ;		
  Real		*derivs ;		
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		lesionedWeight ;
} LinkExtensionRec ;

typedef struct ExampleExtensionRec {	/* BIND */
  int	dummy ;
} ExampleExtensionRec ;

typedef struct ExampleSetExtensionRec {	/* BIND */
  void	*data ;
} ExampleSetExtensionRec ;

#define MbatchSize(net)		((net)->batchSize)
#define	Mtime(net)		((net)->currentTime)

#define MweightCost(net)	((net)->extension->weightCost)

#define Mdebug(net)		((net)->extension->debug)
#define Mrunning(net)		((net)->extension->running)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define McorrectTolerance(net)	((net)->extension->correctTolerance)
#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MinitialUnitOutput(net)	((net)->extension->initialUnitOutput)
#define MinputNoise(net)	((net)->extension->inputNoise)
#define MinputStrength(net)	((net)->extension->inputStrength)
#define MrelaxStepSize(net)	((net)->extension->relaxStepSize)
#define MmaxRelaxSteps(net)	((net)->extension->maxRelaxSteps)
#define Mrefixate(net)		((net)->extension->refixate)
#define MaveFixations(net)	((net)->extension->aveFixations)
#define MavePhonemesCorrect(net) ((net)->extension->avePhonemesCorrect)
#define MaveWordsCorrect(net)	((net)->extension->aveWordsCorrect)
#define MaveRelaxSteps(net)	((net)->extension->aveRelaxSteps)
#define MteacherForcing(net)	((net)->extension->teacherForcing)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MnumSlices(net)		((net)->extension->numSlices)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MsampleExamples(net) 	((net)->extension->sampleExamples)
#define Mcompiled(net)		((net)->extension->compiled)

#define Mlesioned(unit_or_link)	((unit_or_link)->extension->lesioned)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)

#define MoldOutput(unit)	((unit)->extension->oldOutput)
#define MnumCompiledIncoming(unit) ((unit)->extension->numCompiledIncoming)
#define MpreOutputs(unit)	((unit)->extension->preOutputs)
#define MpreOutputDerivs(unit)	((unit)->extension->preOutputDerivs)
#define Mweights(unit)		((unit)->extension->weights)
#define Mderivs(unit)		((unit)->extension->derivs)

extern struct TRACE	afterExample ; 		/* BIND */
extern struct TRACE	forwardEachTime ;	/* BIND */
extern struct TRACE    	inRelaxation ;		/* BIND */

/* compile-elman.c */
extern	void	netCompile		ARGS((Net net)) ;
extern	void	getCompiledWeights	ARGS((Unit unit, void* data)) ;
extern	void	putCompiledDerivs	ARGS((Unit unit, void* data)) ;

/* lesion-elman.c */
extern void	lesionUnit		ARGS((Unit unit)) ;
extern void	unlesionUnit		ARGS((Unit unit)) ;
extern Boolean	isLesionedUnit		ARGS((Unit unit)) ;
extern void	lesionLink		ARGS((Link link)) ;
extern void	unlesionLink		ARGS((Link link)) ;
extern Boolean	isLesionedLink		ARGS((Link link)) ;

#endif				/* ELMAN_H */
