#include <commands.h>
#include <simulator.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "ffbp.h"
#include "ffbp-fm.h"
#include "ffbp-artic.h"

/* for articulatory to acoustic mapping stuff */
static String 	articUnitNames2[NUM_ARTIC_DFS*2] = {
    "Low",         "High", 
    "Front",       "Back", 
    "PlaceFront",  "PlaceBack", 
    "Closed",      "Opened", 
    "NasalClosed", "NasalOpened", 
    "Unvoiced",    "Voiced",
  };

static String 	articUnitNames1[NUM_ARTIC_DFS*1] = {
    "Height",
    "Backness",
    "Place",
    "Constriction",
    "Nasal",
    "Voicing"
  };

static String 	articUnitNames8[NUM_ARTIC_DFS*8] = {
    "Height1",    "Height2",    "Height3",    "Height4",    "Height5",    "Height6",    "Height7",    "Height8",    
    "Backness1",  "Backness2",  "Backness3",  "Backness4",  "Backness5",  "Backness6",  "Backness7",  "Backness8",  
    "Bilabial1",  "LabioDental2", "Dental3",  "Alveolar4",  "PalatoAlv6", "Palatal7",   "Velar8",     "Glottal9",
    "Opened1",    "Opened2",    "Opened3",    "Opened4",    "Opened5",    "Opened6",    "Opened7",    "Opened8",   
    "NasalOpen1", "NasalOpen2", "NasalOpen3", "NasalOpen4", "NasalOpen5", "NasalOpen6", "NasalOpen7", "NasalOpen8",
    "Voicing1",   "Voicing2",   "Voicing3",   "Voicing4",   "Voicing5",   "Voicing6",   "Voicing7",   "Voicing8" 
  };

static String 	articUnitNames12[NUM_ARTIC_DFS*12] = {
    "<<", "<", "Height1",    "Height2",    "Height3",    "Height4",    "Height5",    "Height6",    "Height7",    "Height8",     ">", ">>",
    "<<", "<", "Backness1",  "Backness2",  "Backness3",  "Backness4",  "Backness5",  "Backness6",  "Backness7",  "Backness8",   ">", ">>",
    "<<", "<", "Bilabial1",  "LabioDental2", "Dental3",  "Alveolar4",  "PalatoAlv6", "Palatal7",   "Velar8",     "Glottal9",    ">", ">>",
    "<<", "<", "Opened1",    "Opened2",    "Opened3",    "Opened4",    "Opened5",    "Opened6",    "Opened7",    "Opened8",     ">", ">>",
    "<<", "<", "NasalOpen1", "NasalOpen2", "NasalOpen3", "NasalOpen4", "NasalOpen5", "NasalOpen6", "NasalOpen7", "NasalOpen8",  ">", ">>",
    "<<", "<", "Voicing1",   "Voicing2",   "Voicing3",   "Voicing4",   "Voicing5",   "Voicing6",   "Voicing7",   "Voicing8",    ">", ">>"
  };
 
enum articDFs {
    HEIGHT, BACK, PLACE, OPEN, NASAL, VOICE };

static String articDFnames[NUM_ARTIC_DFS] = {
    "HEIGHT", "BACK", "PLACE", "OPEN", "NASAL", "VOICE" };
    
static String	acousticUnitNames3[NUM_ACOUSTIC_DFS*(2+1)] = {
    "F1low",   "F1high",  "F1norm",  
    "F1down",  "F1up",    "F1DVnorm", 
    "F2low",   "F2high",  "F2norm",
    "F2down",  "F2up",    "F2DVnorm",
    "F3low",   "F3high",  "F3norm",
    "F3down",  "F3up",    "F3DVnorm",
    "FNlow",   "FNhigh",  "FNnorm", 
    "FBlow",   "FBhigh",  "FBnorm",
    "Quiet",   "Loud",    "Lnorm"
  };
  
static String	acousticUnitNames13[NUM_ACOUSTIC_DFS*(12+1)] = {
    "<<", "<", "F1value1",    "F1value2",    "F1value3",    "F1value4",    "F1value5",    "F1value6",    "F1value7",    "F1value8",   ">", ">>", "F1norm",  
    "<<", "<", "F1valueDV1",  "F1valueDV2",  "F1valueDV3",  "F1valueDV4",  "F1valueDV5",  "F1valueDV6",  "F1valueDV7",  "F1valueDV8", ">", ">>", "F1DVnorm", 
    "<<", "<", "F2value1",    "F2value2",    "F2value3",    "F2value4",    "F2value5",    "F2value6",    "F2value7",    "F2value8",   ">", ">>", "F2norm",
    "<<", "<", "F2valueDV1",  "F2valueDV2",  "F2valueDV3",  "F2valueDV4",  "F2valueDV5",  "F2valueDV6",  "F2valueDV7",  "F2valueDV8", ">", ">>", "F2DVnorm",
    "<<", "<", "F3value1",    "F3value2",    "F3value3",    "F3value4",    "F3value5",    "F3value6",    "F3value7",    "F3value8",   ">", ">>", "F3norm",
    "<<", "<", "F3valueDV1",  "F3valueDV2",  "F3valueDV3",  "F3valueDV4",  "F3valueDV5",  "F3valueDV6",  "F3valueDV7",  "F3valueDV8", ">", ">>", "F3DVnorm",
    "<<", "<", "Fnoise1",     "Fnoise2",     "Fnoise3",     "Fnoise4",     "Fnoise5",     "Fnoise6",     "Fnoise7",     "Fnoise8",    ">", ">>", "FNnorm", 
    "<<", "<", "Fburst1",     "Fburst2",     "Fburst3",     "Fburst4",     "Fburst5",     "Fburst6",     "Fburst7",     "Fburst8",    ">", ">>", "FBnorm",
    "<<", "<", "Loudness1",   "Loudness2",   "Loudness3",   "Loudness4",   "Loudness5",   "Loudness6",   "Loudness7",   "Loudness8",  ">", ">>", "Lnorm"
  };
 
enum acousticDFs {
    F1, F1DV, F2, F2DV, F3, F3DV, FNOISE, FBURST, LOUD };

static String acousticDFnames[NUM_ACOUSTIC_DFS] = {
    "F1", "F1DV", "F2", "F2DV", "F3", "F3DV", "FNOISE", "FBURST", "LOUD" };

static void	freeDouble( ptr, size )
  Real	**ptr;
  int	size;
{
  int	i;
  
  for( i = 0; i < size; i++ )
    free( ptr[i] );
  free( ptr );
}

static Real**	makeDouble( idx1, idx2 )
  int idx1;
  int idx2;
{
  int 	i;
  Real 	**ptr;
  
  ptr = (Real **)malloc( idx1 * sizeof( Real * ) );
  if( !ptr ) IErrorAbort( "bad malloc for size %d", idx1 * sizeof( Real * ) );
  for( i = 0; i < idx1; i++ ) {
    ptr[i] = (Real *)malloc( idx2 * sizeof( Real ) );
    if ( !(ptr[i]) ) IErrorAbort( "bad malloc for size %d", idx2 * sizeof( Real ) );
  }
  return ptr;
}

static Real	myExp( num )
  Real num ;
{  
  Real expNum;
  
  if( num < -20.0 ) return 0.0;

  expNum = exp( num );
  
  if( expNum < 0 )
    IErrorAbort( "myExp failed: %f\n", expNum );
    
  return expNum;   
}

static Real	myGaussian (mean, sigma, offset, scale)
  Real	mean;
  Real	sigma;
  Real	offset;
  Real	scale;
{
  Real	gaussValue;
  
  gaussValue = myExp( (mean - offset) * (mean - offset) * -1.0 / sigma ) * scale;
  return gaussValue;
}

static Real	gaussianMean( unitBank, numUnits )
  Unit 	*unitBank;
  int	numUnits;
{
  Real 	meanValue = 0.0, normCheck = 0.0, gaussOffset;
  int 	k, gaussBuff;
  Net 	net;
  Group group;

  gaussOffset = (Real)(numUnits-1) / 2.0;
  net = unitBank[0]->net;
  group = unitBank[0]->group;
  gaussBuff = MgaussianBuffer( unitBank[0]->group );
  
  for( k = 0; k < numUnits; k++ ) {
    meanValue += (numUnits > 2 ? ((Real)k-gaussOffset) * unitBank[k]->output : 
		 (numUnits > 1 ? unitBank[k]->output * Mrange11((Real)k) : 
		  Mrange11( CLIP( unitBank[k]->output, 0.0, 1.0 ) )));
    normCheck += unitBank[k]->output;
  }
  if( (normCheck < 0.999 || normCheck > 1.001) && Mnormalize(group) ) {
    fprintf(dout,"Gaussian bank %s is not normalized: %f!  Setting to zero...\n", unitBank[gaussBuff]->name, normCheck );
    return 0;
  } else {
    meanValue = (numUnits > 2 ? CLIP( meanValue / ((Real)(numUnits - (gaussBuff * 2) - 1) / 2.0), -1.0, 1.0 ) : meanValue);
    return ( meanValue );
  }
}

static void	loadGaussianBank( gaussUnits, numUnits, meanValue )
  Unit	*gaussUnits;
  int	numUnits;
  Real	meanValue;
{
  int	k;
  Real	gaussOffset, targetValue, sumGauss = 0.0, gateValue;
  Net	net;
  Group group;
  
  net = gaussUnits[0]->net;
  group = gaussUnits[0]->group;
  gaussOffset = (Real)(numUnits-1) / 2.0;

  gateValue = (MgatingUnit(group) ? 1.0 - gaussUnits[numUnits]->target : 1.0);
 
  for( k = 0; k < numUnits; k++ ) {
    gaussUnits[k]->target = (numUnits > 2 ? myGaussian( meanValue, MgaussianSigma(net), ((Real)k - gaussOffset), MgaussianScale(net) )
      : ((Real)k - (1.0 - Mrange01(meanValue))) );
    sumGauss += gaussUnits[k]->target;
  }      
    
  for( k = 0; k < numUnits; k++ ) {
    gaussUnits[k]->target = (gaussUnits[k]->target / sumGauss) * gateValue;
    if( !MinRange01( gaussUnits[k]->target ) )
      IErrorAbort( "normalization of gaussian %s %d out of range: %f\n", gaussUnits[k]->name, k, gaussUnits[k]->target );
  }   
}

extern int	copiedFromLayer( net, groupNum )
  Net net;
  int groupNum;
{
  int	i;

  for( i = 1; i <= net->numGroups; i++ )
    if( McopyFromLayer(net->group[i]) == groupNum )
      break;

  if( i > net->numGroups ) return NULL;
    else return i;
}

extern Proc	netOutcomeID (net)
  Net	net ;
{
  int i ;
  Group actionGp, outcomeGp ;
  
  if (!MactionLayer(net) || !MoutcomeLayer(net))
    IErrorAbort("action or outcome layer not specified!") ;

  actionGp  = net->group[MactionLayer(net)] ;
  outcomeGp = net->group[MoutcomeLayer(net)] ;

  if (actionGp->numUnits != outcomeGp->numUnits)
    IErrorAbort("action layer (%s %d) and outcome layer (%s %d) layer have different numbers of units (%d vs. %d)",
		actionGp->name, MactionLayer(net), outcomeGp->name, MoutcomeLayer(net), actionGp->numUnits, outcomeGp->numUnits) ;

  for (i = 0 ; i < actionGp->numUnits ; i++) 
    outcomeGp->unit[i]->target = actionGp->unit[i]->output ;
}

extern Proc	netOutcomeGaussID (net)
  Net	net ;
{
  int 		i ;
  Group 	actionGp, outcomeGp ;
  Real		inputMean;
  
  if (!MactionLayer(net) || !MoutcomeLayer(net))
    IErrorAbort("action or outcome layer not specified!") ;

  actionGp  = net->group[MactionLayer(net)] ;
  outcomeGp = net->group[MoutcomeLayer(net)] ;

  inputMean = gaussianMean( actionGp->unit, actionGp->numUnits ) * ((Real)(outcomeGp->numUnits - (MgaussianBuffer(outcomeGp) * 2) - 1) / 2.0);
  inputMean = CLIP( inputMean, ((Real)(outcomeGp->numUnits - 1) / -2.0), ((Real)(outcomeGp->numUnits - 1) / 2.0) );
  loadGaussianBank( outcomeGp->unit, outcomeGp->numUnits, inputMean );
}

extern void	distort4bit (net)
  Net	net ;
{
  int	idx, jdx, i, toggle,
  	num4bitUnits;
  	  
  Group outcomeGp;
  
  Real	sumBits, currUnit, preUnit, sinVal;

  /*  This now distorts then blends 4bit slot reps */
   
  if (!MoutcomeLayer(net))
    IErrorAbort("outcome layer not specified!") ;

  outcomeGp = net->group[MoutcomeLayer(net)] ;
  
  if( (num4bitUnits = outcomeGp->numUnits / 4) == 0 )
    IErrorAbort( "outcome layer (%s %d) has less than 4 units!", outcomeGp->name, outcomeGp->numUnits );
    
  if( (outcomeGp->numUnits % 4) > 0 )
    IErrorAbort( "outcome layer number of units (%s %d) is not divisable by 4!", outcomeGp->name, outcomeGp->numUnits );

  /* distort the input, an abstract rep of articulation to acoustics */
  for( idx=0; idx < num4bitUnits*4; idx+=4 ) {
    sumBits = 0;
    for( jdx=0; jdx < 4; jdx++ ) 
      sumBits += outcomeGp->unit[idx+jdx]->output;
    for( jdx=0; jdx < 4; jdx++ ) {
      sinVal = (1 + sin( (outcomeGp->unit[idx+jdx]->output * sumBits) )) / 2;
      outcomeGp->unit[idx+jdx]->output = CLIP( sin( (sinVal * 3.1415927) ),0,1 );
    }
  }

  /* blend the 4-bit units, right to left since blending is perserverative */  
  for( idx = (num4bitUnits-1)*4; idx > 0; idx-=4 ) {
    for( jdx=0; jdx < 4; jdx++ ) {
      i = idx+jdx;
      currUnit = outcomeGp->unit[i]->output;
      preUnit  = outcomeGp->unit[i-4]->output;
      outcomeGp->unit[i]->output = preUnit*.2 + currUnit*.8 ;
    }
  } 
}

extern Proc	netOutcomeDistort4bit (net)
  Net	net ;
{
  Group outcomeGp, actionGp;
  int	idx;
  Real	temp;
    
  /*  This an artificial translation function that takes a N slot-based, 4-bit
      action rep and creates a N slot-based, 4-bit goal.  The function shifts
      each on bit rightwards with rap-around.  The shift
      is applied after the previous 4-bit slot is blended 15% with the current slot.
  */
  
  if (!MactionLayer(net) || !MoutcomeLayer(net))
    IErrorAbort("action or outcome layer not specified!") ;

  actionGp  = net->group[MactionLayer(net)] ;
  outcomeGp = net->group[MoutcomeLayer(net)] ;

  if( actionGp->numUnits != outcomeGp->numUnits )
    IErrorAbort("action layer (%s %d) and outcome layer (%s %d) layer have different numbers of units (%d vs. %d)",
		actionGp->name, MactionLayer(net), outcomeGp->name, MoutcomeLayer(net), actionGp->numUnits, outcomeGp->numUnits) ;
    
  /* copy the action group to the outcome */
  for( idx = 0; idx < actionGp->numUnits; idx++ )
    outcomeGp->unit[idx]->target = actionGp->unit[idx]->output;
    
  /* distort4bit works on output, so temporarily swap output and targets... an ugly kluge */
  for( idx = 0; idx < outcomeGp->numUnits; idx++ ) {
    temp = outcomeGp->unit[idx]->target;
    outcomeGp->unit[idx]->target = outcomeGp->unit[idx]->output;
    outcomeGp->unit[idx]->output = temp;
  }
  
  /* blend and shift the outcome layer */
  distort4bit(net) ;

  /* swap back */
  for( idx = 0; idx < outcomeGp->numUnits; idx++ ) {
    temp = outcomeGp->unit[idx]->target;
    outcomeGp->unit[idx]->target = outcomeGp->unit[idx]->output;
    outcomeGp->unit[idx]->output = temp;
  }
}

/***********************************************************************/
int	command_setAcousticNames(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  Group		actionGp, outcomeGp ;
  int		j;
  String	*acousticNames, *articNames;
  
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("For articulatory->acoustic mapping, set the unit names for interpretability");
    IHelp
      (IHelpArgs,
       "For articulatory->acoustic mapping, set the unit names for interpretability",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  if (MactionLayer(currentNet)) {
    actionGp  = currentNet->group[MactionLayer(currentNet)] ;
    if( actionGp->numUnits != NUM_ARTIC_DFS * (MgaussianSize(actionGp)  + MgatingUnit(actionGp)) )
      IErrorAbort( "action group (%s) has illegal # of units (%d)!", actionGp->name, actionGp->numUnits );
    /* set unit name pointers to appropriate names in String arrays */
    switch( MgaussianSize(actionGp) + MgatingUnit(actionGp) ) {
      case 1:  articNames = articUnitNames1;  break;
      case 2:  articNames = articUnitNames2;  break;
      case 8:  articNames = articUnitNames8;  break;
      case 12: articNames = articUnitNames12; break;
      default: IErrorAbort( "names for artic gauss size %d are not available!", 
        MgaussianSize(actionGp) + MgatingUnit(actionGp) ); break;
    }
    for( j = 0; j < actionGp->numUnits; j++ )
      actionGp->unit[j]->name = articNames[j];
  }

  if (MoutcomeLayer(currentNet)) {
    outcomeGp = currentNet->group[MoutcomeLayer(currentNet)] ;
    if( outcomeGp->numUnits != NUM_ACOUSTIC_DFS * (MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp)) )
      IErrorAbort( "outcome group (%s) has illegal # of units (%d)!", outcomeGp->name, outcomeGp->numUnits );
    switch( MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp) ) {
      case 3:  acousticNames = acousticUnitNames3;  break;
      case 13: acousticNames = acousticUnitNames13; break;
      default: IErrorAbort( "names for acoustic gauss size %d are not available!", 
        MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp) ); break;
    }
    for( j = 0; j < outcomeGp->numUnits; j++ ) 
      outcomeGp->unit[j]->name = acousticNames[j];
  }
  return 1 ;
}

static void	mapArticToAcoustics( articDF, pastDF, acousticDF, net )
  Real	*articDF;
  Real	*pastDF;
  Real	*acousticDF;
  Net	net;
{
    Real 	placeWeightNow, placeWeightPast, Flocus, heightDV, backDV, openDV, xsqr;
    int 	j;
            
/*    DB( for(j=0;j<NUM_ARTIC_DFS;j++) fprintf(dout,"%s %f\t", articDFnames[j], articDF[j]); fprintf(dout,"\n"); );
*/
    for( j = 0; j < NUM_ARTIC_DFS; j++ ) 
      if( !MinRange11( articDF[j] ) )
        IErrorAbort( "articDF %s has value out of range %f\n", articDFnames[j], articDF[j] );
    
    /* first calculate the DVs based on diff of current and previous states */
    heightDV = (articDF[HEIGHT] - pastDF[HEIGHT]) / 2.0;
    backDV   = (articDF[BACK] - pastDF[BACK]) / 2.0;
    openDV   = (articDF[OPEN] - pastDF[OPEN]) / 2.0;

    /* the first pass at formant mapping is for vowels only */
    
    acousticDF[F1] = articDF[HEIGHT] * -1.0;
    
    acousticDF[F1DV] = heightDV * -1.0;
    
    acousticDF[F2] = articDF[BACK] * -1.0;
    
    acousticDF[F2DV] = backDV * -1.0;

    acousticDF[F3] = Mrange01( articDF[HEIGHT] ) * (articDF[BACK] * -1.0);
    
    acousticDF[F3DV] = ((Mrange01( articDF[HEIGHT] ) * (articDF[BACK] * -1.0)) - 
                        (Mrange01( pastDF[HEIGHT] ) * (pastDF[BACK] * -1.0))) / 2.0;

    /* Fnoise is like a rough approximation of formant freq of the noise, but includes other factors
       such as dispersement and loudness.  Assumes that there is a set of acoustic cues for frication
       that distinguishes articulatory place of frication (a big simplification).  Also, Fburst is same
       as Fnoise, but the gating on Fburst can be different */
          
    acousticDF[FNOISE]  = articDF[PLACE] * -1.0;
      
    acousticDF[FBURST]  = acousticDF[FNOISE];
      
    /* loudness is complicated.  First start with vowel-type loudness */
    acousticDF[LOUD] = Mbow1( (Mrange01( articDF[VOICE] ) * MAX( Mrange01( articDF[OPEN] ),
      Mrange01( articDF[NASAL] ) * 0.66 ) ), 3 );

    /* then add frication loudness */
    xsqr = ((articDF[OPEN] + 0.66) * 7.0) * ((articDF[OPEN] + 0.66) * 7.0);
    acousticDF[LOUD] = MAX( acousticDF[LOUD], myExp( -1.0 * xsqr ) * 0.66 );
    
    /* and finally add burst loudness */
    acousticDF[LOUD] = MAX( acousticDF[LOUD], (sigmoid( (openDV - 0.6) * 20.0 ) * 0.75) * Mrange01( pastDF[NASAL]*-1.0 ) );
    acousticDF[LOUD] = Mrange11( acousticDF[LOUD] );
          
    /* now recalc formants for consonants, and determine formants based on
       the weighting from amount of constriction (less = vowel, more = consonant */
       
    placeWeightNow = sigmoid( (articDF[OPEN] - 0.3) * 20.0 );
    placeWeightPast = sigmoid( (pastDF[OPEN] - 0.3) * 20.0 );
    acousticDF[F1] = (placeWeightNow * acousticDF[F1]) + ((1.0 - placeWeightNow) *
      Mrange11( (Mrange01( acousticDF[F1] ) * Mrange01( (articDF[NASAL]*-1.0) )) )); 
      
    acousticDF[F1DV] = (placeWeightNow * acousticDF[F1DV]) + ((1.0 - placeWeightNow) * openDV);
            
    Flocus = (acousticDF[F2] - pastDF[PLACE]) / 2.0 ;
    acousticDF[F2DV] = (placeWeightPast * acousticDF[F2DV]) + ((1.0 - placeWeightPast) *
      (Flocus - sigmoid( (Flocus - 0.5) * 20.0 ) / 2.0));
            
    Flocus = (acousticDF[F3] - Mrange11( pastDF[PLACE] * pastDF[PLACE] )) / 2.0;
    acousticDF[F3DV] = (placeWeightPast * acousticDF[F3DV]) + ((1.0 - placeWeightPast) *
      (Flocus - sigmoid( (Flocus - 0.5) * 20.0 ) / 2.0));
      
/*    DB( for(j=0;j<NUM_ACOUSTIC_DFS;j++) fprintf(dout,"%s %f\t", acousticDFnames[j], acousticDF[j]); fprintf(dout,"\n"););
*/
    for( j = 0; j < NUM_ACOUSTIC_DFS; j++ )
      if( !MinRange11( acousticDF[j] ) )
        IErrorAbort( "acousticDF %s has value out of range %f\n", acousticDFnames[j], acousticDF[j] );        
}      

static void	loadAcousticGroup( articDF, pastDF, acousticDF, unitArray )
  Real*  articDF;
  Real*  pastDF;
  Real*  acousticDF;
  Unit*  unitArray;
{
  int 	j, gaussianSize, bankSize;
  Real	scale, meanValue, gaussOffset, period, burst, openDV, xsqr;
  Group group;
  
  group = unitArray[0]->group;
  gaussianSize = MgaussianSize( group );
  bankSize     = gaussianSize + MgatingUnit( group );
  
  gaussOffset = (Real)(gaussianSize-1) / 2.0;
  openDV  = (articDF[OPEN] - pastDF[OPEN]) / 2.0;

  period = Mbow1( (Mrange01( articDF[VOICE] ) * MAX( Mrange01( articDF[OPEN] ),
           Mrange01( articDF[NASAL] ) * 0.66 )), 3 );
  burst  = sigmoid( (openDV - 0.6) * 20.0 ) * Mrange01( pastDF[NASAL]*-1.0 );

    for( j = 0; j < NUM_ACOUSTIC_DFS; j++ ) {
    
      meanValue = (gaussianSize > 2 ? acousticDF[j] * ((Real)(gaussianSize - (MgaussianBuffer(group) * 2) - 1) / 2.0)
        : acousticDF[j]);
      meanValue = (gaussianSize > 2 ? CLIP( meanValue, -gaussOffset, gaussOffset ) : meanValue);

      scale = 0.0; xsqr = 0.0;
      if( j <= F3DV ) 
        scale = 1.0 - period;
      if( j <= F1DV )
	scale = 1.0 - MAX( period*burst*0.66, period*(1.0-burst) );
      if( j == FNOISE ) {
	xsqr = ((articDF[OPEN] + 0.66) * 7.0) * ((articDF[OPEN] + 0.66) * 7.0);
        scale = 1.0 - myExp( -1.0 * xsqr );
      }
      if( j == FBURST )
        scale = 1.0 - burst;
      if( !MinRange01( scale ) )
        IErrorAbort( "gaussian scale %s is out of range: %f %f\n", unitArray[(j+1)*bankSize-1]->name, scale, xsqr );
      if( MgatingUnit( group ) )
      	unitArray[(j+1)*bankSize-1]->target = scale;

      loadGaussianBank( &(unitArray[j*bankSize]), gaussianSize, meanValue );
      
/*      DB( fprintf( dout, "%s-scale: %f\t", acousticDFnames[j], scale ); ); */
    } 
/*    DB( fprintf( dout, "\n" ); ); */
}
  
extern Proc	netOutcomeAcoustics (net)
  Net	net ;
{
  int 	j, idx, copyNum;
  	
  Group actionGp, outcomeGp, copyGp ;
  
  Real	articDF[ NUM_ARTIC_DFS ],
        pastDF[ NUM_ARTIC_DFS ],
  	acousticDF[ NUM_ACOUSTIC_DFS ];

  
  if (!MactionLayer(net) || !MoutcomeLayer(net))
    IErrorAbort("action or outcome layer not specified!") ;

  actionGp  = net->group[MactionLayer(net)] ;
  outcomeGp = net->group[MoutcomeLayer(net)] ;
  if( !(copyNum = copiedFromLayer( net, MactionLayer(net) )) )
    IErrorAbort( "Cannot find context group for action/articulatory layer (%d)", MactionLayer(net) );
  copyGp = net->group[copyNum];

  if( actionGp->numUnits != NUM_ARTIC_DFS * (MgaussianSize(actionGp) + MgatingUnit(actionGp)) )
    IErrorAbort( "action group (%s %d) has illegal # of units (%d)!", 
      actionGp->name, actionGp->numUnits, NUM_ARTIC_DFS * (MgaussianSize(actionGp) + MgatingUnit(actionGp)) );
  if( outcomeGp->numUnits != NUM_ACOUSTIC_DFS * (MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp)) )
    IErrorAbort( "outcome group (%s %d) has illegal # of units (%d)!", 
      outcomeGp->name, outcomeGp->numUnits, NUM_ACOUSTIC_DFS * (MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp)) );
    
  for( j = 0; j < NUM_ARTIC_DFS; j++ ) {
    idx = j * (MgaussianSize(actionGp) + MgatingUnit(actionGp));
    articDF[j] = gaussianMean( &(actionGp->unit[idx]), MgaussianSize(actionGp) );
    pastDF[j]  = gaussianMean( &(copyGp->unit[idx]), MgaussianSize(actionGp) );
  }
   
  mapArticToAcoustics( articDF, pastDF, acousticDF, net );
            
  /* now convert to gaussians on the output */
  loadAcousticGroup( articDF, pastDF, acousticDF, outcomeGp->unit );
  
  if( net->type&RECURRENT ) 
    for( j = 0; j < outcomeGp->numUnits; j++ )
      outcomeGp->unit[j]->targetHistory[net->currentTime] = outcomeGp->unit[j]->target;
}

static Real        pseudoGaussRandom( minVal, maxVal )
  Real	minVal;
  Real	maxVal;
{
  Real	sumRandom = 0;
  int	i;

  /* use central limit to create gaussian prob dist, arbitrary # of samples for now */
  /* probably a better way to do this */
  for( i = 0; i < 3; i++ )
    sumRandom += uniformNoise( minVal, maxVal );
  return( sumRandom / 3.0 );
}

extern void	makeRandomExample (net, example)
  Net		net;
  Example	example;
{
  int	idx, i, j, numArt, artNum, numSlices, normSize, bankSize, minRandom, gatingUnit;
  
  Real	meanValue, gaussOffset, *normArray, *articDF, 
  	normTotal, maxInput, sampleValue;
  Group		articGp;  
  Boolean	newValue;

  numSlices = (net->type&RECURRENT ? net->timeSlices : 1 );
  if( MoutcomeCode(net) == 2 ) {		/* 2 is netOutcomeAcoustics */
    numArt = NUM_ARTIC_DFS;
    normSize = (MactionLayer(net) ? MgaussianSize(net->group[MactionLayer(net)]) : example->numInputs );
  } else {
    numArt = 1;
    normSize = example->numInputs;
  }
  if( MoutcomeProc(net) == netOutcomeGaussID && example->numInputs != example->numTargets )
    IErrorAbort( "example must have same # of inputs and targets for netOutcomeID!" );

  gatingUnit = (MactionLayer(net) ? MgatingUnit( net->group[MactionLayer(net)] ) : 0 );
  bankSize = normSize + gatingUnit;
  /* NEED NOT APPLY ONLY FOR ARTIC->ACOUS MAPPING */
  /*
  if( example->numInputs % (numArt * bankSize) > 0 )
    IErrorAbort( "input example has illegal # of units for art->acoustic mapping!" );
  */
  gaussOffset = (normSize - 1.0) / 2.0;
  normArray   = (Real *)malloc( (bankSize + 1) * sizeof( Real ) );  
  articDF     = (Real *)malloc( numArt * sizeof( Real ) );
  for( i = 0; i < numArt; i++ ) {
    articDF[i] = uniformNoise( 0.0, 1.0 );
    if( i == NASAL )
      articDF[i] = CLIP( sigmoid( Mrange11( articDF[i] ) * 7.0 ), 0.0, 1.0 );
  }

  if( MrandomExampleType(net) == 4 ) {
    if( !MactionLayer(net) )
      IErrorAbort( "Must have an artic group for artic randomExampleType %d", MrandomExampleType(net) );
    articGp = net->group[MactionLayer(net)];
    /* NEED NOT APPLY ONLY FOR ARTIC->ACOUS MAPPING */
    /*
    if( articGp->numUnits != example->numInputs )
      IErrorAbort( "articulatory group must have same number of units as example inputs for randomExampleType %d", MrandomExampleType(net) );
    */
  }

  for( i = 0; i < numSlices; i++ ) {
  
   switch( MrandomExampleType(net) ) {
  
     case 2: 			/* normalized gaussian input */
            
       for( idx = 0; idx < example->numInputs; idx += bankSize ) {
         normArray[normSize] = (gatingUnit ? uniformNoise( 0.0, 1.0 ) : 1.0);
         meanValue = uniformNoise( -1.0, 1.0 ) * (normSize > 2 ? gaussOffset : 1.0);
         for( j = 0, normTotal = 0.0; j < normSize; j++ ) {
           normArray[j] = (normSize > 2 ? myGaussian( meanValue, MgaussianSigma(net), ((Real)j - gaussOffset), MgaussianScale(net) )
             : (j == 0 ? (1.0 - Mrange01( meanValue )) : Mrange01( meanValue )));
           normTotal += normArray[j];
         }
         for( j = 0; j < normSize; j++ ) 
           example->inputVector[i][idx+j] = (normSize > 2 ? (normArray[j] / normTotal) : normArray[j]) * normArray[normSize]; 
         if( gatingUnit ) example->inputVector[i][idx+normSize] = 1.0 - normArray[normSize];
       }
       break;

     case 3:			/* normalized random net input */

       for( j = 0, maxInput = -MAXREAL; j < normSize; j++ ) {
         normArray[j] = uniformNoise( -MrandomMagnitude(net), MrandomMagnitude(net) );
         if( normArray[j] > maxInput ) maxInput = normArray[j];
       }
       for( j = 0, normTotal = 0.0; j < normSize; j++ ) {
         normArray[j] = exp( (normArray[j] + (log(MAXREAL/normSize) - maxInput - 2.0)) );
         normTotal += normArray[j];
       }
       for( j = 0; j < normSize; j++ ) 
         example->inputVector[i][idx+j] = (normArray[j] / normTotal) * normArray[normSize]; 
       if( gatingUnit ) example->inputVector[i][idx+normSize] = 1.0 - normArray[normSize];
       break;
    
    case 4:                             /* biased articulatory sampling (eg, oscillation of OPEN df */

      for( idx = 0, artNum = 0; idx < articGp->numUnits; idx += bankSize, artNum++ ) {
        normArray[normSize] = (gatingUnit ? uniformNoise( 0.0, 1.0 ) : 1.0);
	if( artNum == OPEN ) {
	  meanValue =  Mbow1( uniformNoise( 0.0, 1.0 ), 3 );
	  meanValue *= (articDF[artNum] > 0 ? -1.0 : 1.0);
	  meanValue *= (uniformNoise( 0.0, 1.0 ) < exp( -1.0 * (articDF[artNum]*articDF[artNum]) ) ? -1.0 : 1.0);
	} else {
	  meanValue = articDF[artNum] + pseudoGaussRandom( -0.5, 0.5 );
	  meanValue = CLIP( meanValue, -1.0, 1.0 );
	}
	sampleValue = (normSize > 1 ? uniformNoise( CLIP( (meanValue * -1.0), 0.0, 1.0 ), (1.0 - CLIP( meanValue, 0.0, 1.0 )) )
		       : Mrange01( meanValue ));
        for( j = 0, normTotal = 0.0; j < normSize; j++ ) 
	  if( Mnormalize(articGp) ) {
	    normArray[j] = (normSize > 2 ? myGaussian( meanValue, MgaussianSigma(net), ((Real)j - gaussOffset), MgaussianScale(net) )
	      : (j == 0 ? (1.0 - Mrange01( meanValue )) : Mrange01( meanValue )));
            normTotal += normArray[j];
	  }
	  else 
	    normArray[j] = (j == 0 ? sampleValue : meanValue + sampleValue);
        
        if( MoutcomeProc(net) == netOutcomeGaussID && example->numInputs != example->numTargets )
          IErrorAbort( "example must have same # of inputs and targets for netOutcomeID!" );
        for( j = 0; j < normSize; j++ ) {
	  articGp->unit[idx+j]->outputHistory[i] = (normArray[j] / (Mnormalize(articGp) ? normTotal : 1.0)) * normArray[normSize]; 
	  articGp->unit[idx+j]->clampingHistory[i] = articGp->unit[idx+j]->outputHistory[i]; 
	}
/*      if( gatingUnit ) example->inputVector[i][idx+normSize] = 1.0 - normArray[normSize];  */
	articDF[artNum] = meanValue;
      }
      break;

    case 1: default:			/* uniform random values across input */
    
      if( fabs( MrandomMagnitude(net) ) > 1 )
        IErrorAbort( "randomMagnitude must be within +/-1 for uniform random input!" );
      minRandom = (MrandomMagnitude(net) < 0 ? MrandomMagnitude(net) : 0);
      newValue = (i == 0 ? 1 : (uniformNoise( 0.0, 1.0 ) <= MprobNewInput(net) ? 1 : 0));
      for( j = 0; j < example->numInputs; j++ ) 
        example->inputVector[i][j] = (newValue ? uniformNoise( minRandom, fabs( MrandomMagnitude(net) ) )
          : example->inputVector[i-1][j]);
      break;
   }
  }
  free( normArray );
  free( articDF );
}

extern void	forwardArticNormalized( net, group )
  Net	net;
  Group group;
{
  int 	i, bankSize;
           
  bankSize = (MgaussianSize(group) > 0 ? MgaussianSize(group) + MgatingUnit(group) : group->numUnits); 

  if( group->numUnits % bankSize > 0 )
    IErrorAbort( "%s layer has odd # of units (%d) for art->acoustic normalization!",
      group->name, group->numUnits );
      
  for( i = 0; i < group->numUnits; i += bankSize ) 
    forwardNormalized( &(group->unit[i]), bankSize );
}

static String possiblePhonemes = "@a^e6oIiUujwhrlmnNfvsSzZTDCJpbtdkg_" ;
static String doublePhonemes =   "CJpbtdkg";

static void   computeArticState (event, part2, articTarget, articFrame)
  char     event;
  Boolean  part2;
  Real     *articTarget;
  Real     *articFrame;
{
  Real  scale1, scale2;

  /* first do vowel height (scale1) and backness (scale2) */

  scale1 = scale2 = 0;
  switch( event ) {
    case 'a': scale1 = 1;  scale2 = 9;  break;	/* "cot" */
    case '@': scale1 = 2;  scale2 = 6;  break;	/* "cat" */
    case '^': scale1 = 3;  scale2 = 10; break;	/* "up" */
    case 'e': scale1 = 5;  scale2 = 4;  break;	/* "get" */
    case 'o': scale1 = 5;  scale2 = 14; break;	/* "boat", but just the first part of the dipthong */
    case '6': scale1 = 6;  scale2 = 8;  break;	/* schwa */
    case 'U': scale1 = 7;  scale2 = 11; break;	/* "foot" */
    case 'I': scale1 = 8;  scale2 = 3;  break;	/* "hit" */
    case 'u': scale1 = 10; scale2 = 13; break;	/* "boot" */
    case 'i': scale1 = 11; scale2 = 1;  break;	/* "feet" */
  }
  
  /* convert using maxima on scales, allow for full coarticulation if not a vowel */

  articTarget[HEIGHT] = 0.0;     articFrame[HEIGHT]    = 1.0;
  articTarget[BACK]   = 0.0;     articFrame[BACK]      = 1.0;
  
  if( scale1 || scale2 ) {
    articTarget[HEIGHT]    = (scale1 - 1.0) / (10.0 / 2.0) - 1.0;  articFrame[HEIGHT]    = 0.05;
    articTarget[BACK]      = (scale2 - 1.0) / (13.0 / 2.0) - 1.0;  articFrame[BACK]      = 0.05;
  }

  /* next is place of articulation */

  scale1 = 0; scale2 = 0.05;
  switch( event ) {
    case 'p': case 'b': case 'm': case 'w'                    : scale1 = 1; break;
    case 'f': case 'v'                                        : scale1 = 2; break;
    case 'T': case 'D'                                        : scale1 = 3; break;
    case 'n': case 't': case 'd': case 's': case 'z': case 'l': scale1 = 4; break;
    case 'S': case 'Z': case 'C': case 'J'                    : scale1 = 5; break;
    case 'j'			                              : scale1 = 6; break;
    case 'k': case 'g': case 'r': case 'N'	              : scale1 = 7; break;
    case 'h'			                              : scale1 = 8; break;
  }
  
  if( !scale1 ) {
    articTarget[PLACE] = 0.0; articFrame[PLACE] = 1.0;
  }
  else {
    if( strchr( doublePhonemes, event ) && !part2 ) 
      scale2 = 0.2;
    articTarget[PLACE] = (scale1 - 1.0) / (7.0 / 2.0) - 1.0;  articFrame[PLACE] = scale2;
  }
  
  /* next is amount of constriction.  Assume its a vowel and overwrite if consonant */
  /* part2 of plosives must be a release rather than closed */

  scale1 = 15; scale2 = 0.4;
  switch( event ) {
    case 'p': case 't': case 'k': case 'b': case 'd': case 'g': 
    case 'm': case 'n': case 'N':                               scale1 = 1; scale2 = 0.1; break;
    case 'C': case 'J':					        scale1 = 2; scale2 = 0.1; break;
    case 'f': case 'v': case 'T': case 'D': case 's': case 'z': 
    case 'S': case 'Z': case 'h':				scale1 = 3; scale2 = 0.05; break;
    case 'w': case 'j':     	                                scale1 = 9; scale2 = 0.1;  break;
    case 'l': case 'r':    	                                scale1 = 7; scale2 = 0.1;  break;
    case '_':                                                   scale1 = 8; scale2 = 0.2;  break;
  }
  if( part2 )
    scale1 = 13;

  articTarget[OPEN] = (scale1 - 1.0) / (14.0 / 2.0) - 1.0;
  articFrame[OPEN]  = scale2;
  
  /* next is nasal. must allow for nasalization of vowels. Also allow for some prevoicing of voiced plosives */
  scale1 = 0.5; scale2 = 1.0;
  switch( event ) {
    case 'p': case 't': case 'k': case 'C': 	 		scale1 = -1.0;  scale2 = 0.05; break;
    case 'b': case 'd': case 'g': case 'J':			scale1 = -0.7;  scale2 = 0.1; break;	
    case 'f': case 'v': case 's': case 'z': case 'Z':
    case 'S': case 'T': case 'D': case 'h':			scale1 = -0.75; scale2 = 0.1; break;
    case 'm': case 'n': case 'N': case '_':			scale1 = 1.0;   scale2 = 0.1; break;
  }
  if( part2 ) 
    scale2 = 0.2; 
  
  articTarget[NASAL] = scale1;
  articFrame[NASAL]  = scale2;
  
  /* last is voicing.  Assume a vowel and overwrite. Voicing for plosives only occurs in second event */
  scale1 = 8; scale2 = 0.2;
  switch( event ) {
    case 'p': case 't': case 'k': case 'f': case 'T': 
    case 's': case 'S': case 'h': case 'C':		scale1 = 1; scale2 = 0.1; break;
    case 'b': case 'd': case 'g': case 'J':		scale1 = 3; scale2 = 0.2; break;
    case 'v': case 'z': case 'Z': case 'D':		scale1 = 6; scale2 = 0.3; break;
    case '_':						scale1 = 1; scale2 = 0.05; break;
  }
  if( part2 ) {
    if( strchr( "bdgJ", event ) )  
      scale1 = 5;
    else
      scale1 = 2;
    scale2 = 0.2;
  }
  articTarget[VOICE] = (scale1 - 1.0) / (7.0 / 2.0) - 1.0;
  articFrame[VOICE]  = scale2;
}

static Boolean     equalArticStates( artic1, artic2 )
  Real *artic1;
  Real *artic2;
{
  int i;

  for( i = 0; i < NUM_ARTIC_DFS; i++ )
    if( fabs( artic1[i] - artic2[i] ) > 0.1 )
      return FALSE;
  
  return TRUE;
}

#define MAX_EVENTS	50

static	Real	articTarget[MAX_EVENTS][NUM_ARTIC_DFS];
static	Real	articFrame[MAX_EVENTS][NUM_ARTIC_DFS];
static	String	eventString[MAX_EVENTS] ;

extern	int	computeAcousticGoals (net, example)
  Net		net;
  Example	example;
{
  int		i, j, k, idx, maxEvents,	/* maxEvents for current net, not absolute capacity */
                numEvents, numFuses, groupIdx;  
  Group		outcomeGp, *group;
  
  Real		acousticTarget[NUM_ACOUSTIC_DFS], 
  		movement, minTarget, maxTarget, initTarget, coartAmount;
  
  if( !MoutcomeLayer(net) )
    IErrorAbort( "Outcome layer must be specified for computing acoustic goal!" );
  outcomeGp = net->group[MoutcomeLayer(net)];
  
  if( outcomeGp->numUnits != NUM_ACOUSTIC_DFS * (MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp)) )
    IErrorAbort( "outcome layer (%d) does not have correct # of units (%d) for computing acoustic goal!", 
      outcomeGp->numUnits, NUM_ACOUSTIC_DFS * (MgaussianSize(outcomeGp) + MgatingUnit(outcomeGp))) ;
   
  /* first insert any extra slots for phonemes that have > 1 artic/acoustic event 
     Note: we are assuming a 1-to-1 mapping from articulatory to acoustic events, but
     not from phonemes to artic or acoustic events.  */
  
  maxEvents = (net->type&RECURRENT ? net->timeSlices : 1);

  /* No padding with silence, must be added to phoneme string if wanted */
  for( i = 1, j = 0, numFuses = 0; i < strlen( example->name ) - 1; i++, j++ ) {
    /* DCP: trap on unknown phoneme */
    if (strchr(possiblePhonemes,example->name[i])==NULL)
      IErrorAbort("unknown phoneme \"%c\" in word \"%s\"", example->name[i], example->name) ;
    eventString[j] = example->name[i];
    if( strchr(doublePhonemes,example->name[i]) ) {
      eventString[++j] = '#';
      /* DCP: use example->name instead of eventString */
      if( strchr( "ptkbdg", example->name[i] ) && strchr( "ptkbdg", example->name[i+1] ) )
	numFuses++;
    }
  }
  numEvents = j;
  if( numEvents-numFuses > maxEvents )
    IErrorAbort( "# of artic/acoustic events (%d-%d) greater than net max (%d) for \"%s\"", 
		numEvents, numFuses, maxEvents, example->name );

  /* compute articulatory target frame for each phoneme */
  for( i = 0; i < numEvents; i++ ) {
    if( eventString[i] == '#' )
      computeArticState( eventString[i-1], TRUE, articTarget[i], articFrame[i] );
    else
      computeArticState( eventString[i], FALSE, articTarget[i], articFrame[i] );
  }

  /* adjust targets (within confines of frames) by forward and backward coarticulation.
     adjustment is done serially left-to-right, and adjustments are weighted linearly in time 
     coart of DVs is done with the difference of their static states FORWARD in time */
     
  for( i = 0; i < numEvents; i++ ) {
    for( j = 0; j < NUM_ARTIC_DFS; j++ ) {
      minTarget = CLIP( articTarget[i][j] - articFrame[i][j], -1.0, 1.0 );  
      maxTarget = CLIP( articTarget[i][j] + articFrame[i][j], -1.0, 1.0 );  
      initTarget = pseudoGaussRandom( minTarget, maxTarget );
      for( k = -1; k < numEvents; k++ ) {
        coartAmount = 1.0 / MAX( (k > i ? Manticipate(net) : Mperseverate(net)), .00001 );
        coartAmount = 1.0 / ((Real)abs(k-i) * coartAmount + 1.0);
        movement = (k < 0 ? initTarget : articTarget[k][j]) - initTarget;
        articTarget[i][j] = CLIP( (articTarget[i][j] + movement * coartAmount), minTarget, maxTarget );
      }
    }  
  } 

  /* DCP: determine outcomeLayer's displacement in targetVector (idx) if necessary */
  /* DO NOT STORE IN EXAMPLE
  if( outcomeGp->type&OUTPUT ) 
    for( idx=0, groupIdx=0, group=net->group; groupIdx < MoutcomeLayer(net) ; groupIdx++, group++ ) 
      if( (*group)->type&OUTPUT ) idx += (*group)->numUnits ;
  */

  if ( MgoalLayer(net) && net->group[MgoalLayer(net)]->numUnits != outcomeGp->numUnits)
    IErrorAbort("goal layer \"%s\" (%d) and outcome layer \"%s\" (%d) must be same size",
		net->group[MgoalLayer(net)]->name, net->group[MgoalLayer(net)]->numUnits,
		outcomeGp->name, outcomeGp->numUnits) ;

  for( i = 0, k = 0; i < numEvents; i++ ) {
    /* don't count 1st middle event of adjacent plosives */
    if( eventString[i] == '#' && strchr( "ptkbdg", eventString[i-1] ) && strchr( "ptkbdg", eventString[i+1] ) )
      continue;

    mapArticToAcoustics( articTarget[i], (i ? articTarget[i-1] : articTarget[i]), acousticTarget, net );
    
    loadAcousticGroup( articTarget[i], (i ? articTarget[i-1] : articTarget[i]), acousticTarget, outcomeGp->unit );

    for( j = 0; j < outcomeGp->numUnits; j++ ) {
      Real target = outcomeGp->unit[j]->target;
      if( net->type&RECURRENT ) {
        outcomeGp->unit[j]->outputHistory[k]   = target ;
        outcomeGp->unit[j]->clampingHistory[k] = target ;
        outcomeGp->unit[j]->targetHistory[k]   = target ;
	if( MgoalLayer(net) ) {
          net->group[MgoalLayer(net)]->unit[j]->outputHistory[k]   = target ;
          net->group[MgoalLayer(net)]->unit[j]->clampingHistory[k] = target ;
          net->group[MgoalLayer(net)]->unit[j]->targetHistory[k]   = target ;
	}
      }
      /* DCP: set output for first slice even for recurrent networks */
      /*
      if (k = 0) {
	outcomeGp->unit[j]->output = target ;
	if( MgoalLayer(net) )
	  net->group[MgoalLayer(net)]->unit[j]->output = target ;
      }
      */
      /* DCP: added displacement idx in case of multiple output layers */
      /* DO NOT STORE IN EXAMPLE
      if( outcomeGp->type&OUTPUT ) 
        example->targetVector[k][j+idx] = outcomeGp->unit[j]->target;
      */
    }
    k++;
  }
  return( numEvents-numFuses );
}

extern void	computeGaussianIDgoal (net, example)
  Net 		net;
  Example	example;
{
  int 	i;
  Group	outcomeGp, inputGp;
  Real	inputMean;
  
  if( !MoutcomeLayer(net) )
    IErrorAbort( "Outcome layer must be specified for computing gaussian ID goal!" );
  outcomeGp   = net->group[MoutcomeLayer(net)];

  /* NOTE!  only works when there is only one input group */
  for (i = 0, inputGp=net->group[i] ; i < net->numGroups ; inputGp=net->group[++i])
    if (inputGp->type&INPUT) break;
  if( i >= net->numGroups )
    IErrorAbort( "no input groups found for computeGaussianIDgoal!" );

  inputMean = gaussianMean( inputGp->unit, inputGp->numUnits ) * ((Real)(outcomeGp->numUnits - (MgaussianBuffer(outcomeGp) * 2) - 1) / 2.0);
  inputMean = CLIP( inputMean, ((Real)(outcomeGp->numUnits - 1) / -2.0), ((Real)(outcomeGp->numUnits - 1) / 2.0) );
  loadGaussianBank( outcomeGp->unit, outcomeGp->numUnits, inputMean );

  if( outcomeGp->numUnits != example->numTargets )
    IErrorAbort( "# of units in outcome and target must be same for computeGaussianIDgoal!" );    
  for( i = 0; i < outcomeGp->numUnits; i++ ) {
    outcomeGp->unit[i]->output = outcomeGp->unit[i]->target;
    example->targetVector[0][i] = outcomeGp->unit[i]->target; 
  }
  
}
