
/**********************************************************************
 * $Id: bp.h,v 1.7 91/09/26 11:05:16 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#ifndef FFBP_H
#define FFBP_H

#define MIN_UNIT_OUTPUT 0.000001
#define LOG(x)	log((double)(x))

#define SIMPLE_DIVERGE(output, target) 	(LOG((target)/(output))*(target))             
   
#define DIVERGE(output, target)  ((target) == 0.0 ? 0.0 : 			\
  (target) == 1.0 ? (-target)*LOG(MAX(output,MIN_UNIT_OUTPUT)) : SIMPLE_DIVERGE(output, target))

#define DIVERGE_DERIV(output, target)              \
  ((output) - (target))				/* deriv of the diverge error wrt to the net_in */


/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  Boolean	running ;
  int	delayCount ;		/* netParam: */
  int	alphaOnset ;		/* netParam: */
  Real	unitNoise ;		/* netParam: */
  Real	inputNoise ;		/* netParam: */
  Real	inputScaleFactor ;	/* netParam: */
  int	errorFunction ;		/* netParam: 0=CE 1=SSE */
  Boolean bpToOutputs ;		/* netParam: */
  Real	nonExtremalTarget ;	/* netParam: */
  Real	zeroErrorRadius ;	/* netParam: */
  int	outputsIncorrect ;

  /* Elman networks */
  int	maxSlice ;
  Boolean	normError ;	/* netParam: */
  Real	initialContextOutput ;	/* netParam: */
  Real	correctTolerance ;	/* netParam: */
  Real	pExampleCorrect ;
  Real	pSliceCorrect ;

  /* Refixation networks */
  Boolean	refixate ;
  Real		aveFixations ;
  String	inputGroupName ;
  String	attentionGroupName ;
  String	doneGroupName ;

  /* StJohn/McClelland-style networks */
  int	bridgeLayer ;
  int	predictOutput ;		/* netParam: */
  Boolean abortIfIncorrect ;	/* netParam: */

    
  /* cost term on unit activations */
  Real	unitCost ;		/* netParam: weighting on penalty */
  Real	unitPrior ;		/* netParam: */
  Boolean lengthScaleDerivs ;	/* netParam: */

  Real	targetExternalInput ;	/* netParam: g */
  Real	targExtInConstant ;	/* netParam: k where extIn = g*t/(t+k) */
  Real	targExtInHi ;
  Real	targExtInLo ;
  int	freqFunction ;		/* 0=raw 1=sqrt 2=log */

  /* adaptive weight decay/elimination */
  Real	standardWeight ;	/* netParam: */
  Real	weightCostIncrement ;	/* netParam: */
  Real	weightCostDecrement ;	/* netParam: */
  Real	lastError ;
  Real	aveError ;
  Real	aveErrorFactor ;	/* netParam: */
  Real	desiredError ;		/* netParam: */

  int	exampleNum ;
  Real	exampleFrequency ;
  Boolean	sampleExamples ;	/* netParam: */
  Boolean	compiled ;
  
  /* Forward models */
  int		goalLayer ;	/* netParam: */
  int		actionLayer ;	/* netParam: */
  int		outcomeLayer ;	/* netParam: */
  NetProc	outcomeProc ;
  int		outcomeCode ;   /* netParam: */	
  Real		predictionError ;
  int		randomExampleType;	/* netParam: */
  Real		randomMagnitude;	/* netParam: */
  Real		gaussianScale;	/* netParam: */
  Real		gaussianSigma;	/* netParam: */
  Real		anticipate;	/* netParam: */
  Real		perseverate;	/* netParam: */
  Real		probNewInput;	/* netParam: */
  Boolean	phonologyFM ;	/* netParam: */
  Boolean	adaptComprehension ;	/* netParam: */
  Real		comprehensionError ;
  Real		imitationError ;
  Real		intentionError ;
  Boolean	trainFM ;	/* netParam: */
  Boolean	trainComp ;	/* netParam: */
  Boolean	onlyTrainImitate ;	/* netParam: */
  Boolean	stopAtAcoustics ;	/* netParam: */
  Boolean	babble ;	/* netParam: */
  Boolean	onlyBabble ;	/* netParam: */

} NetExtensionRec ;

typedef struct GroupExtensionRec {	/* BIND */
  int	copyFromLayer ;
  int	copyTargetsFromLayer ;
  int	noUnitCost ;
  int	noWeightCost ;
  Real	minBias ;
  Real	maxBias ;
  Real	minIncomingWeight ;
  Real	maxIncomingWeight ;
  Real	minOutgoingWeight ;
  Real	maxOutgoingWeight ;
  
  /* normalization and artic to acoustic mapping */
  Boolean 	alreadyNormed ;
  Boolean	normalize ;	
  int		gaussianSize;	
  int		gaussianBuffer;
  Boolean	gatingUnit;
  
} GroupExtensionRec ;

typedef struct UnitExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Boolean	equateIncomingWeights ;
  Boolean	equateOutgoingWeights ;
  int		numCompiledIncoming ;
  Real		saved ;
  Real*		*preOutputs ;		
  Real*		*preOutputDerivs ;	
  Real		*weights ;		
  Real		*derivs ;		
  Link		*compiledLinks ;
  Real		*outputSlice;
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		lesionedWeight ;
  Real		length ;
} LinkExtensionRec ;

typedef struct ExampleSetExtensionRec {	/* BIND */
  void	*data ;
} ExampleSetExtensionRec ;

#define MbatchSize(net)		((net)->batchSize)

#define Mdebug(net)		((net)->extension->debug)
#define Mrunning(net)		((net)->extension->running)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MerrorFunction(net)	((net)->extension->errorFunction)
#define MbpToOutputs(net)	((net)->extension->bpToOutputs)
#define MnonExtremalTarget(net)	((net)->extension->nonExtremalTarget)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define MoutputsIncorrect(net)	((net)->extension->outputsIncorrect)
#define MunitNoise(net)		((net)->extension->unitNoise)
#define MinputNoise(net)	((net)->extension->inputNoise)
#define MinputScaleFactor(net)	((net)->extension->inputScaleFactor)
#define MmaxSlice(net)		((net)->extension->maxSlice)
#define MnormError(net)		((net)->extension->normError)
#define MinitialContextOutput(net) ((net)->extension->initialContextOutput)
#define McorrectTolerance(net)	((net)->extension->correctTolerance)
#define MpExampleCorrect(net)	((net)->extension->pExampleCorrect)
#define MpSliceCorrect(net)	((net)->extension->pSliceCorrect)
#define Mrefixate(net)		((net)->extension->refixate)
#define MaveFixations(net)	((net)->extension->aveFixations)
#define MinputGroupName(net)	((net)->extension->inputGroupName)
#define MattentionGroupName(net) ((net)->extension->attentionGroupName)
#define MdoneGroupName(net)	((net)->extension->doneGroupName)
#define MbridgeLayer(net)	((net)->extension->bridgeLayer)
#define MpredictOutput(net)	((net)->extension->predictOutput)
#define MabortIfIncorrect(net)	((net)->extension->abortIfIncorrect)

#define MunitCost(net) 		((net)->extension->unitCost)
#define MunitPrior(net)		((net)->extension->unitPrior)
#define MlengthScaleDerivs(net)	((net)->extension->lengthScaleDerivs)
#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MweightCostIncrement(net) ((net)->extension->weightCostIncrement)
#define MweightCostDecrement(net) ((net)->extension->weightCostDecrement)
#define MlastError(net)		((net)->extension->lastError)
#define MaveError(net)		((net)->extension->aveError)
#define MaveErrorFactor(net)	((net)->extension->aveErrorFactor)
#define MdesiredError(net)	((net)->extension->desiredError)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MsampleExamples(net) 	((net)->extension->sampleExamples)
#define Mcompiled(net)		((net)->extension->compiled)

#define MphonologyFM(net)	((net)->extension->phonologyFM)
#define MadaptComprehension(net) ((net)->extension->adaptComprehension)
#define McomprehensionError(net) ((net)->extension->comprehensionError)
#define MimitationError(net)	((net)->extension->imitationError)
#define MintentionError(net)	((net)->extension->intentionError)
#define MtrainFM(net)		((net)->extension->trainFM)
#define MtrainComp(net)		((net)->extension->trainComp)
#define MonlyTrainImitate(net)	((net)->extension->onlyTrainImitate)
#define MstopAtAcoustics(net)	((net)->extension->stopAtAcoustics)
#define Mbabble(net)		((net)->extension->babble)
#define MonlyBabble(net)		((net)->extension->onlyBabble)

#define MtargetExternalInput(net) ((net)->extension->targetExternalInput)
#define MtargExtInConstant(net)	((net)->extension->targExtInConstant)
#define MtargExtInHi(net)	((net)->extension->targExtInHi)
#define MtargExtInLo(net)	((net)->extension->targExtInLo)
#define MfreqFunction(net)	((net)->extension->freqFunction)

#define McopyFromLayer(group)		((group)->extension->copyFromLayer)
#define McopyTargetsFromLayer(group)	((group)->extension->copyTargetsFromLayer)
#define MnoUnitCost(group)		((group)->extension->noUnitCost)
#define MnoWeightCost(group)		((group)->extension->noWeightCost)
#define MminBias(group)			((group)->extension->minBias)
#define MmaxBias(group)			((group)->extension->maxBias)
#define MminIncomingWeight(group)	((group)->extension->minIncomingWeight)
#define MmaxIncomingWeight(group)	((group)->extension->maxIncomingWeight)
#define MminOutgoingWeight(group)	((group)->extension->minOutgoingWeight)
#define MmaxOutgoingWeight(group)	((group)->extension->maxOutgoingWeight)
#define MalreadyNormed(group)		((group)->extension->alreadyNormed)
#define Mnormalize(group)		((group)->extension->normalize)

#define Msaved(unit)		((unit)->extension->saved)
#define MnumCompiledIncoming(unit) ((unit)->extension->numCompiledIncoming)
#define MpreOutputs(unit)	((unit)->extension->preOutputs)
#define MpreOutputs(unit)	((unit)->extension->preOutputs)
#define MpreOutputDerivs(unit)	((unit)->extension->preOutputDerivs)
#define Mweights(unit)		((unit)->extension->weights)
#define Mderivs(unit)		((unit)->extension->derivs)
#define McompiledLinks(unit)	((unit)->extension->compiledLinks)
#define MequateIncomingWeights(unit)	((unit)->extension->equateIncomingWeights)
#define MequateOutgoingWeights(unit)	((unit)->extension->equateOutgoingWeights)

#define Mlesioned(unit_or_link)	((unit_or_link)->extension->lesioned)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)
#define Mlength(link)		((link)->extension->length)

#define DB(x)	if (Mdebug(currentNet)) { x }

extern	struct TRACE	afterExample ; 		/* BIND */
extern	struct TRACE	forwardEachTime ;	/* BIND */

void	forward				ARGS((Unit unit, void *data)) ;
void	backward			ARGS((Unit unit, void *data)) ;
void	backwardNoWtDvs			ARGS((Unit unit, void *data)) ;
Real	sumTotalInput			ARGS((Unit unit)) ;

void	scaleInput			ARGS((Unit unit, void *data)) ;
void	addInputNoise			ARGS((Unit unit, void *data)) ;
void	forwardTargExtInput		ARGS((Unit unit, void *data)) ;
void	restoreOutputFromSlice		ARGS((Unit unit, void *slice));

void	incrementError			ARGS((Unit unit, void *data)) ;
void	zeroOutputDeriv			ARGS((Unit unit, void *data)) ;
void	injectError			ARGS((Unit unit, void *data)) ;
void	injectUnitCost			ARGS((Unit unit, void *data)) ;
void	netDoUnits			ARGS((Net net, int first, int last, Mask mask, Boolean test, UnitDataProc proc, void *data)) ;

void	PMSPtargExtInRange		ARGS((Net net)) ;

static void	backwardRecomputeOutputError    ARGS((Unit unit, void *data)) ;
extern void	copyOutputsToOutputs 		ARGS((Net net, int fromGroupIdx, int toGroupIdx)) ;
extern void	saveOutputToSlice		ARGS((Unit unit, void *slice)) ;
extern void	netBackwardFromContext		ARGS((Net net, int last, int first)) ;

static void	netForward	ARGS((Net 	 net)) ;
static void	netBackward	ARGS((Net	 net)) ;


#endif				/* FFBP_H */


