
/**********************************************************************
 * $Id: bp.h,v 1.7 91/09/26 11:05:16 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#ifndef FFBP_H
#define FFBP_H

/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  Boolean	running ;
  int	alphaOnset ;		/* netParam: */
  Real	unitNoise ;		/* netParam: */
  Real	inputNoise ;		/* netParam: */
  Real	inputScaleFactor ;	/* netParam: */
  Boolean symmetricSigmoid ;		/* netParam: */
  int	errorFunction ;		/* netParam: 0=CE 1=SSE */
  Boolean bpToOutputs ;		/* netParam: */
  Real	nonExtremalOneTarget ;	/* netParam: */
  Real	nonExtremalZeroTarget ;	/* netParam: */
  Real	zeroErrorRadius ;	/* netParam: */
  int	outputsIncorrect ;

  /* cascading */
  Real	dt ;			/* netParam: */
  int	delayCount ;		/* netParam: */
  Boolean initializeUnits ;	/* netParam: */
  Real	initialUnitOutput ;	/* netParam: */
  Real	outputChange ;	
  Real	outputChangeTolerance ;	/* netParam: */
  Boolean outputChangeUseAve ;		/* netParam: */
  Boolean useStoppingCriteria ;		/* netParam: */
  Real	aveTime ;
  Real	minTime ;		/* netParam: */
  Real	maxTime ;		/* netParam: */

  /* Elman networks */
  int	maxSlice ;
  Boolean normError ;		/* netParam: */
  Boolean initializeContext ;	/* netParam: */
  int	lastExample ;
  Real	initialContextOutput ;	/* netParam: */
  Real	correctTolerance ;	/* netParam: */
  Boolean	correctWTA ;	/* netParam: */
  Real	pExampleCorrect ;
  Real	pSliceCorrect ;

  /* Refixation networks */
  Boolean	refixate ;
  Real		aveFixations ;
  String	inputGroupName ;
  String	attentionGroupName ;
  String	doneGroupName ;

  /* StJohn/McClelland-style networks */
  int	bridgeLayer ;
  int	predictOutput ;		/* netParam: */
  Boolean abortIfIncorrect ;	/* netParam: */

  /* cost term on unit activations */
  Real	unitCost ;		/* netParam: weighting on penalty */
  Real	unitPrior ;		/* netParam: */
  Boolean lengthScaleDerivs ;	/* netParam: */

  Real	targetExternalInput ;	/* netParam: g */
  Real	targExtInConstant ;	/* netParam: k where extIn = g*t/(t+k) */
  Real	targExtInHi ;
  Real	targExtInLo ;
  int	freqFunction ;		/* 0=raw 1=sqrt 2=log */

  /* adaptive weight decay/elimination */
  Real	standardWeight ;	/* netParam: */
  Real	weightCostIncrement ;	/* netParam: */
  Real	weightCostDecrement ;	/* netParam: */
  Real	lastError ;
  Real	aveError ;
  Real	aveErrorFactor ;	/* netParam: */
  Real	desiredError ;		/* netParam: */

  int	exampleNum ;
  Real	exampleFrequency ;
  Boolean	sampleExamples ;	/* netParam: */
  Real	assocProb ;		/* netParam: */
  Boolean	marcus ;
  Real	marcusRepeatProb ;	/* netParam: */
  Real	marcusSilenceProb ;	/* netParam: */

  Boolean	incTraining ;	/* netParam: */
  Boolean	incTrainingSMnsyl ;	/* netParam: */
  Real	incTrainingThreshold ;	/* netParam: */

  int	computeExample ;	/* netParam: */
  int	modifyExample ;	/* netParam: */
  Boolean	compiled ;
} NetExtensionRec ;

#define MARCUS_INPUT 1
#define MARCUS_CURRENT 4
#define MARCUS_NEXT 5

typedef struct GroupExtensionRec {	/* BIND */
  int	copyFromLayer ;
  int	copyWTA ;
  int	noUnitCost ;
  int	noWeightCost ;
  int	fixedBias ;
  Real	minBias ;
  Real	maxBias ;
  Real	minIncomingWeight ;
  Real	maxIncomingWeight ;
  Real	minOutgoingWeight ;
  Real	maxOutgoingWeight ;
  int	groupXPosition ;
  int	groupYPosition ;
  Real	inputNoise ;
  Real	meanOutput ;
} GroupExtensionRec ;

typedef struct UnitExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Boolean	equateIncomingWeights ;
  Boolean	equateOutgoingWeights ;
  int		n ;
  int		numCompiledIncoming ;
  Real*		*preOutputs ;		
  Real*		*preOutputDerivs ;	
  Real		*weights ;		
  Real		*derivs ;		
  Link		*compiledLinks ;
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		lesionedWeight ;
  Real		length ;
} LinkExtensionRec ;

#define NO_ASSOC	-1
#define MAX_ASSOCS	32

typedef struct ExampleExtensionRec {	/* BIND */
  int	numAssocs ;
  int	assoc[32] ;
  int	presentations ;
} ExampleExtensionRec ;

typedef struct ExampleSetExtensionRec {	/* BIND */
  void	*data ;
} ExampleSetExtensionRec ;

#define MbatchSize(net)		((net)->batchSize)

#define Mdebug(net)		((net)->extension->debug)
#define Mrunning(net)		((net)->extension->running)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MsymmetricSigmoid(net)	((net)->extension->symmetricSigmoid)
#define MerrorFunction(net)	((net)->extension->errorFunction)
#define MbpToOutputs(net)	((net)->extension->bpToOutputs)
#define MnonExtremalOneTarget(net)  ((net)->extension->nonExtremalOneTarget)
#define MnonExtremalZeroTarget(net) ((net)->extension->nonExtremalZeroTarget)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define MoutputsIncorrect(net)	((net)->extension->outputsIncorrect)

#define Mdt(net)		((net)->extension->dt)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MinitializeUnits(net)	((net)->extension->initializeUnits)
#define MlastExample(net)	((net)->extension->lastExample)
#define MinitialUnitOutput(net)	((net)->extension->initialUnitOutput)
#define MoutputChange(net)	((net)->extension->outputChange)
#define MoutputChangeTolerance(net) ((net)->extension->outputChangeTolerance)
#define MoutputChangeUseAve(net)    ((net)->extension->outputChangeUseAve)
#define MaveTime(net)		((net)->extension->aveTime)
#define MminTime(net)		((net)->extension->minTime)
#define MmaxTime(net)		((net)->extension->maxTime)

#define MunitNoise(net)		((net)->extension->unitNoise)
#define MinputNoise(net)	((net)->extension->inputNoise)
#define MinputScaleFactor(net)	((net)->extension->inputScaleFactor)
#define MmaxSlice(net)		((net)->extension->maxSlice)
#define MnormError(net)		((net)->extension->normError)
#define MinitializeContext(net) ((net)->extension->initializeContext)
#define MinitialContextOutput(net) ((net)->extension->initialContextOutput)
#define McorrectTolerance(net)	((net)->extension->correctTolerance)
#define McorrectWTA(net)	((net)->extension->correctWTA)
#define MpExampleCorrect(net)	((net)->extension->pExampleCorrect)
#define MpSliceCorrect(net)	((net)->extension->pSliceCorrect)
#define Mrefixate(net)		((net)->extension->refixate)
#define MaveFixations(net)	((net)->extension->aveFixations)
#define MinputGroupName(net)	((net)->extension->inputGroupName)
#define MattentionGroupName(net) ((net)->extension->attentionGroupName)
#define MdoneGroupName(net)	((net)->extension->doneGroupName)
#define MbridgeLayer(net)	((net)->extension->bridgeLayer)
#define MpredictOutput(net)	((net)->extension->predictOutput)
#define MabortIfIncorrect(net)	((net)->extension->abortIfIncorrect)
#define MunitCost(net) 		((net)->extension->unitCost)
#define MunitPrior(net)		((net)->extension->unitPrior)
#define MlengthScaleDerivs(net)	((net)->extension->lengthScaleDerivs)
#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MweightCostIncrement(net) ((net)->extension->weightCostIncrement)
#define MweightCostDecrement(net) ((net)->extension->weightCostDecrement)
#define MlastError(net)		((net)->extension->lastError)
#define MaveError(net)		((net)->extension->aveError)
#define MaveErrorFactor(net)	((net)->extension->aveErrorFactor)
#define MdesiredError(net)	((net)->extension->desiredError)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MsampleExamples(net) 	((net)->extension->sampleExamples)
#define MassocProb(net) 	((net)->extension->assocProb)
#define Mmarcus(net)		((net)->extension->marcus)
#define MmarcusRepeatProb(net)	((net)->extension->marcusRepeatProb)
#define MmarcusSilenceProb(net)	((net)->extension->marcusSilenceProb)
#define MincTraining(net)	((net)->extension->incTraining)
#define MincTrainingSMnsyl(net) ((net)->extension->incTrainingSMnsyl)
#define MincTrainingThreshold(net) ((net)->extension->incTrainingThreshold)
#define McomputeExample(net)	((net)->extension->computeExample)
#define MmodifyExample(net)	((net)->extension->modifyExample)
#define Mcompiled(net)		((net)->extension->compiled)

#define MtargetExternalInput(net) ((net)->extension->targetExternalInput)
#define MtargExtInConstant(net)	((net)->extension->targExtInConstant)
#define MtargExtInHi(net)	((net)->extension->targExtInHi)
#define MtargExtInLo(net)	((net)->extension->targExtInLo)
#define MfreqFunction(net)	((net)->extension->freqFunction)

#define McopyFromLayer(group)	((group)->extension->copyFromLayer)
#define McopyWTA(group)		((group)->extension->copyWTA)
#define MnoUnitCost(group)	((group)->extension->noUnitCost)
#define MnoWeightCost(group)	((group)->extension->noWeightCost)
#define MfixedBias(group)	((group)->extension->fixedBias)
#define MminBias(group)		((group)->extension->minBias)
#define MmaxBias(group)		((group)->extension->maxBias)
#define MminIncomingWeight(group) ((group)->extension->minIncomingWeight)
#define MmaxIncomingWeight(group) ((group)->extension->maxIncomingWeight)
#define MminOutgoingWeight(group) ((group)->extension->minOutgoingWeight)
#define MmaxOutgoingWeight(group) ((group)->extension->maxOutgoingWeight)
#define MgroupXPosition(group)	((group)->extension->groupXPosition)
#define MgroupYPosition(group)	((group)->extension->groupYPosition)
#define MmeanOutput(group)	((group)->extension->meanOutput)

#define MnumCompiledIncoming(unit) ((unit)->extension->numCompiledIncoming)
#define MpreOutputs(unit)	((unit)->extension->preOutputs)
#define MpreOutputDerivs(unit)	((unit)->extension->preOutputDerivs)
#define Mweights(unit)		((unit)->extension->weights)
#define Mderivs(unit)		((unit)->extension->derivs)
#define McompiledLinks(unit)	((unit)->extension->compiledLinks)
#define MequateIncomingWeights(unit)	((unit)->extension->equateIncomingWeights)
#define MequateOutgoingWeights(unit)	((unit)->extension->equateOutgoingWeights)
#define Mn(unit)		((unit)->extension->n)

#define Mlesioned(unit_or_link)	((unit_or_link)->extension->lesioned)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)
#define Mlength(link)		((link)->extension->length)

/* just added these for priming-ffbp.c */
#define Mi_slice(unit)		((unit)->clampingHistory)
#define Mx(unit)		((unit)->totalInput)
#define Moldx(unit)		((unit)->extension->old_x)
#define Mext(unit)		((unit)->extension->ext_x)
#define My(unit)		((unit)->output)
#define My_slice(unit)		((unit)->outputHistory)
#define My_t(unit)		((unit)->extension->y_t)
#define Msigx_t(unit)		((unit)->extension->sigx_t)
#define Md(unit)		((unit)->target)
#define Md_slice(unit)		((unit)->targetHistory)
#define Md_t(unit)		((unit)->extension->d_t)
#define MdEdx(unit)		((unit)->inputDeriv)
#define MdEdy(unit)		((unit)->outputDeriv)
#define McdEdy(unit)		((unit)->extension->cdEdy)
#define MerrorFactor(unit)	((unit)->extension->errorFactor)
#define	Mslice(net)		((net)->currentTime)
#define MuseStoppingCriteria(net)   ((net)->extension->useStoppingCriteria)

/* OLD
#define CASCADE(net,time)						\
  for (time = 0.0, MoutputChange(net) = 0.0 ;				\
       time < MmaxTime(net) &&						\
       (MoutputChange(net) > MoutputChangeTolerance(net) ||		\
	time < MminTime(net)) &&					\
       (MoutputChange(net)=0.0) == 0.0 ;				\
       time += Mdt(net))
OLD */

/* update of outputChange inside test is to allow initialization after comparison */
#define CASCADE(net,time)						\
  for (time = 0.0, MoutputChange(net) = 0.0, _numOutputs = 0.0 ;	\
       time < MmaxTime(net) &&						\
       (time < MminTime(net)       ||					\
	!MuseStoppingCriteria(net) ||					\
	(MoutputChangeUseAve(net) 					\
	 ? (_numOutputs == 0.0 ? TRUE :					\
	    MoutputChange(net)/_numOutputs > MoutputChangeTolerance(net)) \
	  : MoutputChange(net) > MoutputChangeTolerance(net))) &&	\
	(MoutputChange(net)=0.0)==0.0 && (_numOutputs=0.0)==0.0 ;	\
       time += Mdt(net))

/* loop is nonsense just to slow the display and fool the compiler */ 
#define DISPLAY_CASCADE(net)						\
  { int	idx, delayCount = MdelayCount(net), tmp = 0 ;			\
    IDoTrace(&inCascade) ;						\
    for (idx = 0, tmp = net->currentTime ; idx < delayCount*10000 ; idx++) \
      net->currentTime++ ;						\
    net->currentTime = tmp ; }

#define SYM2ASYM(y) (0.5*(y)+0.5)
#define ASYM2SYM(y) (2.0*(y)-1.0)

extern	struct TRACE	afterExample ; 		/* BIND */
extern	struct TRACE	forwardEachTime ;	/* BIND */
extern struct TRACE    	inCascade ;		/* BIND */

#endif				/* FFBP_H */
